@@ -77,6 +77,14 @@ CK_RV do_SignVerifyUpdateRSA(struct GENERATED_TEST_SUITE_INFO *tsuite)
                        (unsigned int) tsuite->mech.mechanism);
         goto testcase_cleanup;
     }
+    if (!mech_supported(slot_id, tsuite->keygen_mech.mechanism)) {
+        testsuite_skip(tsuite->tvcount,
+                       "Slot %u doesn't support %s (0x%x)",
+                       (unsigned int) slot_id,
+                       mech_to_str(tsuite->keygen_mech.mechanism),
+                       (unsigned int) tsuite->keygen_mech.mechanism);
+        goto testcase_cleanup;
+    }
 
     // iterate over test vectors
     for (i = 0; i < tsuite->tvcount; i++) {
@@ -94,6 +102,12 @@ CK_RV do_SignVerifyUpdateRSA(struct GENERATED_TEST_SUITE_INFO *tsuite)
                        tsuite->name, i, s,
                        tsuite->tv[i].modbits, tsuite->tv[i].keylen);
 
+        if (tsuite->tv[i].modbits > 4096 && !rsa8k) {
+            testcase_skip("Tests with  modbits='%lu' are not enabled",
+                          tsuite->tv[i].modbits);
+            free(s);
+            continue;
+        }
         if (!keysize_supported(slot_id, tsuite->mech.mechanism,
                                tsuite->tv[i].modbits)) {
             testcase_skip("Token in slot %lu cannot be used with modbits='%lu'",
@@ -176,6 +190,7 @@ CK_RV do_SignVerifyUpdateRSA(struct GENERATED_TEST_SUITE_INFO *tsuite)
 
         // generate key pair
         rc = generate_RSA_PKCS_KeyPair_cached(session,
+                                              tsuite->keygen_mech.mechanism,
                                               tsuite->tv[i].modbits,
                                               tsuite->tv[i].publ_exp,
                                               tsuite->tv[i].publ_exp_len,
@@ -358,6 +373,14 @@ CK_RV do_SignVerifyUpdate_RSAPSS(struct GENERATED_TEST_SUITE_INFO * tsuite)
                        (unsigned int) tsuite->mech.mechanism);
         goto testcase_cleanup;
     }
+    if (!mech_supported(slot_id, tsuite->keygen_mech.mechanism)) {
+        testsuite_skip(tsuite->tvcount,
+                       "Slot %u doesn't support %s (0x%x)",
+                       (unsigned int) slot_id,
+                       mech_to_str(tsuite->keygen_mech.mechanism),
+                       (unsigned int) tsuite->keygen_mech.mechanism);
+        goto testcase_cleanup;
+    }
     // iterate over test vectors
     for (i = 0; i < tsuite->tvcount; i++) {
 
@@ -374,6 +397,12 @@ CK_RV do_SignVerifyUpdate_RSAPSS(struct GENERATED_TEST_SUITE_INFO * tsuite)
                        tsuite->name, i, s,
                        tsuite->tv[i].modbits, tsuite->tv[i].keylen);
 
+        if (tsuite->tv[i].modbits > 4096 && !rsa8k) {
+            testcase_skip("Tests with  modbits='%lu' are not enabled",
+                          tsuite->tv[i].modbits);
+            free(s);
+            continue;
+        }
         if (!keysize_supported(slot_id, tsuite->mech.mechanism,
                                tsuite->tv[i].modbits)) {
             testcase_skip("Token in slot %lu cannot be used with modbits='%lu'",
@@ -423,6 +452,7 @@ CK_RV do_SignVerifyUpdate_RSAPSS(struct GENERATED_TEST_SUITE_INFO * tsuite)
 
         // generate key pair
         rc = generate_RSA_PKCS_KeyPair_cached(session,
+                                              tsuite->keygen_mech.mechanism,
                                               tsuite->tv[i].modbits,
                                               tsuite->tv[i].publ_exp,
                                               tsuite->tv[i].publ_exp_len,
@@ -629,6 +659,12 @@ CK_RV do_VerifyUpdateRSA(struct PUBLISHED_TEST_SUITE_INFO * tsuite)
 
         testcase_begin("%s Verify with test vector %u.", tsuite->name, i);
 
+        if (tsuite->tv[i].mod_len * 8 > 4096 && !rsa8k) {
+            testcase_skip("Tests with  modbits='%lu' are not enabled",
+                          tsuite->tv[i].mod_len * 8);
+            free(s);
+            continue;
+        }
         if (!keysize_supported(slot_id, tsuite->mech.mechanism,
                                tsuite->tv[i].mod_len * 8)) {
             testcase_skip("Token in slot %lu cannot be used with modbits='%lu'",
@@ -716,6 +752,10 @@ CK_RV do_VerifyUpdateRSA(struct PUBLISHED_TEST_SUITE_INFO * tsuite)
                 testcase_skip("RSA key import is not allowed by policy");
                 continue;
             }
+            if (rc == CKR_KEY_SIZE_RANGE) {
+                testcase_skip("RSA key size is not in supported range");
+                continue;
+            }
 
             testcase_error("create_RSAPublicKey(), rc=%s", p11_get_ckr(rc));
             goto error;
@@ -834,6 +874,12 @@ CK_RV do_SignUpdateRSA(struct PUBLISHED_TEST_SUITE_INFO * tsuite)
         }
         testcase_begin("%s Sign with test vector %u.", tsuite->name, i);
 
+        if (tsuite->tv[i].mod_len * 8 > 4096 && !rsa8k) {
+            testcase_skip("Tests with  modbits='%lu' are not enabled",
+                          tsuite->tv[i].mod_len * 8);
+            free(s);
+            continue;
+        }
         if (!keysize_supported(slot_id, tsuite->mech.mechanism,
                                tsuite->tv[i].mod_len * 8)) {
             testcase_skip("Token in slot %lu cannot be used with modbits='%lu'",
@@ -957,6 +1003,10 @@ CK_RV do_SignUpdateRSA(struct PUBLISHED_TEST_SUITE_INFO * tsuite)
                 testcase_skip("RSA key import is not allowed by policy");
                 continue;
             }
+            if (rc == CKR_KEY_SIZE_RANGE) {
+                testcase_skip("RSA key size is not in supported range");
+                continue;
+            }
 
             testcase_error("create_RSAPrivateKey(), rc=%s", p11_get_ckr(rc));
             goto error;
