@@ -28,6 +28,7 @@
 #include "mech_to_str.h"
 #include "ec_curves.h"
 #include "common.c"
+#include <openssl/rsa.h>
 
 CK_RSA_PKCS_OAEP_PARAMS oaep_params_sha1 = {
         .hashAlg = CKM_SHA_1,
@@ -127,6 +128,14 @@ struct wrapping_mech_info wrapping_tests[] = {
         .rsa_publ_exp_len = 3,
         .rsa_publ_exp = {0x01, 0x00, 0x01},
     },
+    {
+        .name = "Wrap/Unwrap with RSA 8192 PKCS",
+        .wrapping_mech = { CKM_RSA_PKCS, 0, 0 },
+        .wrapping_key_gen_mech = { CKM_RSA_PKCS_KEY_PAIR_GEN, 0, 0 },
+        .rsa_modbits = 8192,
+        .rsa_publ_exp_len = 3,
+        .rsa_publ_exp = {0x01, 0x00, 0x01},
+    },
     {
         .name = "Wrap/Unwrap with RSA 1024 PKCS OAEP (SHA1)",
         .wrapping_mech = { CKM_RSA_PKCS_OAEP, &oaep_params_sha1,
@@ -339,6 +348,16 @@ struct wrapping_mech_info wrapping_tests[] = {
         .rsa_publ_exp_len = 3,
         .rsa_publ_exp = {0x01, 0x00, 0x01},
     },
+    {
+        .name = "Wrap/Unwrap with RSA 8192 AES KEY WRAP (AES 256, OAEP SHA-1)",
+        .wrapping_mech = { CKM_RSA_AES_KEY_WRAP,
+                           &rsa_aeskw_aes_256_oaep_sha1_params,
+                           sizeof(CK_RSA_AES_KEY_WRAP_PARAMS) },
+        .wrapping_key_gen_mech = { CKM_RSA_PKCS_KEY_PAIR_GEN, 0, 0 },
+        .rsa_modbits = 8192,
+        .rsa_publ_exp_len = 3,
+        .rsa_publ_exp = {0x01, 0x00, 0x01},
+    },
     {
         .name = "Wrap/Unwrap with ECDH AES KEY WRAP (AES 256, P-256)",
         .wrapping_mech = { CKM_ECDH_AES_KEY_WRAP,
@@ -463,6 +482,12 @@ struct wrapped_mech_info wrapped_key_tests[] = {
         .operation_mech = { CKM_SHA_1_HMAC, 0, 0 },
         .sym_keylen = 32,
     },
+    {
+        .name = "key type SHA-1 HMAC",
+        .wrapped_key_gen_mech = { CKM_SHA_1_KEY_GEN, 0, 0 },
+        .operation_mech = { CKM_SHA_1_HMAC, 0, 0 },
+        .sym_keylen = 32,
+    },
     {
         .name = "key type RSA PKCS 512",
         .wrapped_key_gen_mech = { CKM_RSA_PKCS_KEY_PAIR_GEN, 0, 0 },
@@ -506,6 +531,60 @@ struct wrapped_mech_info wrapped_key_tests[] = {
         .operation_mech = { CKM_IBM_DILITHIUM, 0, 0 },
         .pqc_keyform = CK_IBM_DILITHIUM_KEYFORM_ROUND3_65,
     },
+    {
+        .name = "key type IBM ML-DSA 44",
+        .wrapped_key_gen_mech = { CKM_IBM_ML_DSA_KEY_PAIR_GEN, 0, 0 },
+        .operation_mech = { CKM_IBM_ML_DSA, 0, 0 },
+        .pqc_keyform = CKP_IBM_ML_DSA_44,
+    },
+    {
+        .name = "key type IBM ML-DSA 65",
+        .wrapped_key_gen_mech = { CKM_IBM_ML_DSA_KEY_PAIR_GEN, 0, 0 },
+        .operation_mech = { CKM_IBM_ML_DSA, 0, 0 },
+        .pqc_keyform = CKP_IBM_ML_DSA_65,
+    },
+    {
+        .name = "key type IBM ML-DSA 87",
+        .wrapped_key_gen_mech = { CKM_IBM_ML_DSA_KEY_PAIR_GEN, 0, 0 },
+        .operation_mech = { CKM_IBM_ML_DSA, 0, 0 },
+        .pqc_keyform = CKP_IBM_ML_DSA_87,
+    },
+    {
+        .name = "key type IBM ML-KEM 512 (using CKM_IBM_ML_KEM)",
+        .wrapped_key_gen_mech = { CKM_IBM_ML_KEM_KEY_PAIR_GEN, 0, 0 },
+        .operation_mech = { CKM_IBM_ML_KEM, 0, 0 },
+        .pqc_keyform = CKP_IBM_ML_KEM_512,
+    },
+    {
+        .name = "key type IBM ML-KEM 768 (using CKM_IBM_ML_KEM)",
+        .wrapped_key_gen_mech = { CKM_IBM_ML_KEM_KEY_PAIR_GEN, 0, 0 },
+        .operation_mech = { CKM_IBM_ML_KEM, 0, 0 },
+        .pqc_keyform = CKP_IBM_ML_KEM_768,
+    },
+    {
+        .name = "key type IBM ML-KEM 1024 (using CKM_IBM_ML_KEM)",
+        .wrapped_key_gen_mech = { CKM_IBM_ML_KEM_KEY_PAIR_GEN, 0, 0 },
+        .operation_mech = { CKM_IBM_ML_KEM, 0, 0 },
+        .pqc_keyform = CKP_IBM_ML_KEM_1024,
+    },
+    {
+        .name = "key type IBM ML-KEM 512 (using CKM_IBM_ML_KEM_WITH_ECDH)",
+        .wrapped_key_gen_mech = { CKM_IBM_ML_KEM_KEY_PAIR_GEN, 0, 0 },
+        .operation_mech = { CKM_IBM_ML_KEM_WITH_ECDH, 0, 0 },
+        .pqc_keyform = CKP_IBM_ML_KEM_512,
+    },
+    {
+        .name = "key type IBM ML-KEM 768 (using CKM_IBM_ML_KEM_WITH_ECDH)",
+        .wrapped_key_gen_mech = { CKM_IBM_ML_KEM_KEY_PAIR_GEN, 0, 0 },
+        .operation_mech = { CKM_IBM_ML_KEM_WITH_ECDH, 0, 0 },
+        .pqc_keyform = CKP_IBM_ML_KEM_768,
+    },
+    {
+        .name = "key type IBM ML-KEM 1024 (using CKM_IBM_ML_KEM_WITH_ECDH)",
+        .wrapped_key_gen_mech = { CKM_IBM_ML_KEM_KEY_PAIR_GEN, 0, 0 },
+        .operation_mech = { CKM_IBM_ML_KEM_WITH_ECDH, 0, 0 },
+        .pqc_keyform = CKP_IBM_ML_KEM_1024,
+    },
 };
 
 #define NUM_WRAPPED_KEY_TESTS sizeof(wrapped_key_tests) / \
@@ -533,14 +612,49 @@ CK_RV do_perform_operation(CK_MECHANISM *mech,
     CK_RV rc;
     CK_MECHANISM_INFO mech_info;
     CK_BBOOL encr = FALSE, decr = FALSE, sign = FALSE, verify = FALSE;
+    CK_BBOOL derive = FALSE;
     CK_ULONG input_size, cipher_size, output_size;
     CK_BYTE input_data[32] = { 0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,
                                0x08,0x09,0x0a,0x0b,0x0c,0x0d,0x0e,0x0f,
                                0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,
                                0x18,0x19,0x1a,0x1b,0x1c,0x1d,0x1e,0x1f};
-    CK_BYTE cipher_data[512];
+    CK_BYTE cipher_data[OPENSSL_RSA_MAX_MODULUS_BITS / 8];
     CK_BYTE output_data[8192];
     CK_OBJECT_HANDLE encr_key, decr_key, sign_key, verify_key;
+    CK_IBM_ML_KEM_PARAMS ml_kem_params;
+    CK_IBM_ML_KEM_WITH_ECDH_PARAMS ml_kem_ecdh_params;
+    CK_OBJECT_CLASS class = CKO_SECRET_KEY;
+    CK_KEY_TYPE key_type = CKK_AES;
+    CK_BBOOL true = CK_TRUE;
+    CK_BBOOL false = CK_FALSE;
+    CK_ULONG secret_key_len = 32;
+    CK_ATTRIBUTE  derive_tmpl[] = {
+        {CKA_CLASS, &class, sizeof(class)},
+        {CKA_KEY_TYPE, &key_type, sizeof(key_type)},
+        {CKA_SENSITIVE, &false, sizeof(false)},
+        {CKA_VALUE_LEN, &secret_key_len, sizeof(secret_key_len)},
+        {CKA_SIGN, &true, sizeof(true)},
+        {CKA_VERIFY, &true, sizeof(true)},
+    };
+    CK_ULONG derive_tmpl_len = sizeof(derive_tmpl) / sizeof(CK_ATTRIBUTE);
+    CK_OBJECT_HANDLE key1 = CK_INVALID_HANDLE;
+    CK_OBJECT_HANDLE key2 = CK_INVALID_HANDLE;
+    CK_MECHANISM kem_mech;
+    CK_MECHANISM aes_ebc_mech = { CKM_AES_ECB, NULL, 0 };
+    CK_OBJECT_HANDLE ec_publ_key1 = CK_INVALID_HANDLE;
+    CK_OBJECT_HANDLE ec_priv_key1 = CK_INVALID_HANDLE;
+    CK_OBJECT_HANDLE ec_publ_key2 = CK_INVALID_HANDLE;
+    CK_OBJECT_HANDLE ec_priv_key2 = CK_INVALID_HANDLE;
+    CK_BYTE pubkey1_value[256];
+    CK_BYTE pubkey2_value[256];
+    CK_ATTRIBUTE  extr_tmpl1[] = {
+        {CKA_EC_POINT, pubkey1_value, sizeof(pubkey1_value)},
+    };
+    CK_ULONG extr_tmpl1_len = sizeof(extr_tmpl1)/sizeof(CK_ATTRIBUTE);
+    CK_ATTRIBUTE  extr_tmpl2[] = {
+        {CKA_EC_POINT, pubkey2_value, sizeof(pubkey2_value)},
+    };
+    CK_ULONG extr_tmpl2_len = sizeof(extr_tmpl2)/sizeof(CK_ATTRIBUTE);
 
     /* Check if Encrypt/Decrypt or Sign/Verify is supported */
     rc = funcs->C_GetMechanismInfo(slot1, mech->mechanism, &mech_info);
@@ -551,6 +665,7 @@ CK_RV do_perform_operation(CK_MECHANISM *mech,
     }
     encr = (mech_info.flags & CKF_ENCRYPT) != 0;
     sign = (mech_info.flags & CKF_SIGN) != 0;
+    derive = (mech_info.flags & CKF_DERIVE) != 0;
 
     rc = funcs->C_GetMechanismInfo(slot2, mech->mechanism, &mech_info);
     if (rc != CKR_OK) {
@@ -561,6 +676,7 @@ CK_RV do_perform_operation(CK_MECHANISM *mech,
 
     decr = (mech_info.flags & CKF_DECRYPT) != 0;
     verify = (mech_info.flags & CKF_VERIFY) != 0;
+    derive &= ((mech_info.flags & CKF_DERIVE) != 0);
 
     if (encr && decr) {
         /* Perform Encrypt/Decrypt operation */
@@ -640,6 +756,7 @@ CK_RV do_perform_operation(CK_MECHANISM *mech,
         case CKM_RSA_PKCS:
         case CKM_ECDSA:
         case CKM_IBM_DILITHIUM:
+        case CKM_IBM_ML_DSA:
             sign_key = priv_key2;
             verify_key = publ_key1;
             input_size = 20;
@@ -660,7 +777,8 @@ CK_RV do_perform_operation(CK_MECHANISM *mech,
         rc = funcs->C_Sign(sess2, input_data, input_size, output_data,
                            &output_size);
         if (rc != CKR_OK) {
-            if (mech->mechanism == CKM_IBM_DILITHIUM &&
+            if ((mech->mechanism == CKM_IBM_DILITHIUM ||
+                 mech->mechanism == CKM_IBM_ML_DSA) &&
                 rc == CKR_KEY_SIZE_RANGE) {
                 testcase_skip("IBM Dilithium key form not supported by slot %lu",
                               slot2);
@@ -684,8 +802,179 @@ CK_RV do_perform_operation(CK_MECHANISM *mech,
             testcase_error("C_Verify on slot %lu rc=%s", slot1, p11_get_ckr(rc));
             return rc;
         }
+    } else if (derive) {
+        /* Perform Derive/KEM operation */
+        switch (mech->mechanism) {
+        case CKM_IBM_ML_KEM:
+            kem_mech.mechanism = mech->mechanism;
+            kem_mech.pParameter = &ml_kem_params;
+            kem_mech.ulParameterLen = sizeof(ml_kem_params);
+
+            /* Perform encapsulation with public key */
+            memset(&ml_kem_params, 0, sizeof(ml_kem_params));
+            ml_kem_params.ulVersion = CK_IBM_ML_KEM_VERSION;
+            ml_kem_params.mode = CK_IBM_ML_KEM_ENCAPSULATE;
+            ml_kem_params.ulCipherLen = sizeof(output_data);
+            ml_kem_params.pCipher = output_data;
+            ml_kem_params.kdf = CKD_NULL;
+            ml_kem_params.pSharedData = NULL;;
+            ml_kem_params.ulSharedDataLen = 0;
+            ml_kem_params.bPrepend = CK_FALSE;
+            ml_kem_params.hSecret = CK_INVALID_HANDLE;
+
+            /* Encapsulation */
+            rc = funcs->C_DeriveKey(sess1, &kem_mech, publ_key1, derive_tmpl,
+                                    derive_tmpl_len, &key1);
+            if (rc != CKR_OK) {
+                testcase_error("C_DeriveKey (encapsulation) on slot %lu rc=%s",
+                               slot1, p11_get_ckr(rc));
+                return rc;
+            }
+            output_size = ml_kem_params.ulCipherLen;
+
+            /* Perform Decapsulation with private key */
+            memset(&ml_kem_params, 0, sizeof(ml_kem_params));
+            ml_kem_params.ulVersion = CK_IBM_ML_KEM_VERSION;
+            ml_kem_params.mode = CK_IBM_ML_KEM_DECAPSULATE;
+            ml_kem_params.ulCipherLen = output_size;
+            ml_kem_params.kdf = CKD_NULL;
+            ml_kem_params.pCipher = output_data;
+            ml_kem_params.pSharedData = NULL;;
+            ml_kem_params.ulSharedDataLen = 0;
+            ml_kem_params.bPrepend = CK_FALSE;
+            ml_kem_params.hSecret = CK_INVALID_HANDLE;
+
+            rc = funcs->C_DeriveKey(sess2, &kem_mech, priv_key2, derive_tmpl,
+                                    derive_tmpl_len, &key2);
+            if (rc != CKR_OK) {
+                testcase_error("C_DeriveKey (decapsulation) on slot %lu rc=%s",
+                               slot1, p11_get_ckr(rc));
+                funcs->C_DestroyObject(sess1, key1);
+                return rc;
+            }
+
+            rc = do_perform_operation(&aes_ebc_mech,
+                                      slot1, sess1, key1, CK_INVALID_HANDLE,
+                                      slot2, sess2, key2, CK_INVALID_HANDLE);
+            funcs->C_DestroyObject(sess2, key2);
+            if (rc != CKR_OK)
+                return rc;
+            break;
+        case CKM_IBM_ML_KEM_WITH_ECDH:
+            /* Generate EC key pairs on both slots */
+            rc = generate_EC_KeyPair(sess1, prime256v1, sizeof(prime256v1),
+                                     &ec_publ_key1, &ec_priv_key1, CK_FALSE);
+            if (rc != CKR_OK) {
+                testcase_error("generate_EC_KeyPair on slot %lu rc=%s",
+                               slot1, p11_get_ckr(rc));
+                return rc;
+            }
+
+            rc = funcs->C_GetAttributeValue(sess1, ec_publ_key1,
+                                            extr_tmpl1, extr_tmpl1_len);
+            if (rc != CKR_OK) {
+                testcase_error("C_GetAttributeValue: on slot %lu rc=%s",
+                               slot1, p11_get_ckr(rc));
+                funcs->C_DestroyObject(sess1, ec_priv_key1);
+                funcs->C_DestroyObject(sess1, ec_publ_key1);
+                return rc;
+            }
+
+            rc = generate_EC_KeyPair(sess2, prime256v1, sizeof(prime256v1),
+                                     &ec_publ_key2, &ec_priv_key2, CK_FALSE);
+            if (rc != CKR_OK) {
+                testcase_error("generate_EC_KeyPair on slot %lu rc=%s",
+                               slot2, p11_get_ckr(rc));
+                funcs->C_DestroyObject(sess1, ec_priv_key1);
+                funcs->C_DestroyObject(sess1, ec_publ_key1);
+                return rc;
+            }
+
+            rc = funcs->C_GetAttributeValue(sess2, ec_publ_key2,
+                                            extr_tmpl2, extr_tmpl2_len);
+            if (rc != CKR_OK) {
+                testcase_error("C_GetAttributeValue: on slot %lu rc=%s",
+                               slot2, p11_get_ckr(rc));
+                funcs->C_DestroyObject(sess1, ec_priv_key1);
+                funcs->C_DestroyObject(sess1, ec_publ_key1);
+                funcs->C_DestroyObject(sess2, ec_priv_key2);
+                funcs->C_DestroyObject(sess2, ec_publ_key2);
+                return rc;
+            }
+
+            kem_mech.mechanism = mech->mechanism;
+            kem_mech.pParameter = &ml_kem_ecdh_params;
+            kem_mech.ulParameterLen = sizeof(ml_kem_ecdh_params);
+
+            /* Perform encapsulation with public key */
+            memset(&ml_kem_ecdh_params, 0, sizeof(ml_kem_ecdh_params));
+            ml_kem_ecdh_params.mode = CK_IBM_ML_KEM_ENCAPSULATE;
+            ml_kem_ecdh_params.ulCipherLen = sizeof(output_data);
+            ml_kem_ecdh_params.pCipher = output_data;
+            ml_kem_ecdh_params.kdf = CKD_IBM_HYBRID_SHA256_KDF;
+            ml_kem_ecdh_params.pSharedData = NULL;
+            ml_kem_ecdh_params.ulSharedDataLen = 0;
+            ml_kem_ecdh_params.hECPrivateKey = ec_priv_key1;
+            ml_kem_ecdh_params.pPublicData = extr_tmpl2[0].pValue;
+            ml_kem_ecdh_params.ulPublicDataLen = extr_tmpl2[0].ulValueLen;
+
+            /* Encapsulation */
+            rc = funcs->C_DeriveKey(sess1, &kem_mech, publ_key1, derive_tmpl,
+                                    derive_tmpl_len, &key1);
+            if (rc != CKR_OK) {
+                testcase_error("C_DeriveKey (encapsulation) on slot %lu rc=%s",
+                               slot1, p11_get_ckr(rc));
+                funcs->C_DestroyObject(sess1, ec_priv_key1);
+                funcs->C_DestroyObject(sess1, ec_publ_key1);
+                funcs->C_DestroyObject(sess2, ec_priv_key2);
+                funcs->C_DestroyObject(sess2, ec_publ_key2);
+                return rc;
+            }
+            output_size = ml_kem_ecdh_params.ulCipherLen;
+
+            /* Perform Decapsulation with private key */
+            memset(&ml_kem_ecdh_params, 0, sizeof(ml_kem_ecdh_params));
+            ml_kem_ecdh_params.mode = CK_IBM_ML_KEM_DECAPSULATE;
+            ml_kem_ecdh_params.ulCipherLen = output_size;
+            ml_kem_ecdh_params.pCipher = output_data;
+            ml_kem_ecdh_params.kdf = CKD_IBM_HYBRID_SHA256_KDF;
+            ml_kem_ecdh_params.pSharedData = NULL;
+            ml_kem_ecdh_params.ulSharedDataLen = 0;
+            ml_kem_ecdh_params.hECPrivateKey = ec_priv_key2;
+            ml_kem_ecdh_params.pPublicData = extr_tmpl1[0].pValue;
+            ml_kem_ecdh_params.ulPublicDataLen = extr_tmpl1[0].ulValueLen;
+
+            rc = funcs->C_DeriveKey(sess2, &kem_mech, priv_key2, derive_tmpl,
+                                    derive_tmpl_len, &key2);
+            if (rc != CKR_OK) {
+                testcase_error("C_DeriveKey (decapsulation) on slot %lu rc=%s",
+                               slot1, p11_get_ckr(rc));
+                funcs->C_DestroyObject(sess1, key1);
+                funcs->C_DestroyObject(sess1, ec_priv_key1);
+                funcs->C_DestroyObject(sess1, ec_publ_key1);
+                funcs->C_DestroyObject(sess2, ec_priv_key2);
+                funcs->C_DestroyObject(sess2, ec_publ_key2);
+                return rc;
+            }
+
+            rc = do_perform_operation(&aes_ebc_mech,
+                                      slot1, sess1, key1, CK_INVALID_HANDLE,
+                                      slot2, sess2, key2, CK_INVALID_HANDLE);
+            funcs->C_DestroyObject(sess2, key2);
+            funcs->C_DestroyObject(sess1, ec_priv_key1);
+            funcs->C_DestroyObject(sess1, ec_publ_key1);
+            funcs->C_DestroyObject(sess2, ec_priv_key2);
+            funcs->C_DestroyObject(sess2, ec_publ_key2);
+            if (rc != CKR_OK)
+                return rc;
+            break;
+
+        default:
+            testcase_error("Operation not supported by testcase");
+            return CKR_FUNCTION_NOT_SUPPORTED;
+        }
     } else {
-        testcase_error("Neither Encrypt/Decrypt, nor Sign/Verify supported");
+        testcase_error("Neither Encrypt/Decrypt, Sign/Verify, nor derive supported");
         return CKR_FUNCTION_NOT_SUPPORTED;
     }
 
@@ -704,11 +993,13 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
     CK_ULONG wrapped_key_size = 0;
     CK_OBJECT_CLASS key_class;
     CK_KEY_TYPE key_type;
+    CK_BBOOL ck_true = CK_TRUE;
     CK_ULONG key_len, unwrapped_keylen;
     CK_ATTRIBUTE unwrap_tmpl[] = {
         {CKA_CLASS, &key_class, sizeof(CK_OBJECT_CLASS)},
         {CKA_KEY_TYPE, &key_type, sizeof(CK_KEY_TYPE)},
-        {CKA_VALUE_LEN, &key_len, sizeof(CK_ULONG)}
+        {CKA_VALUE_LEN, &key_len, sizeof(CK_ULONG)},
+        {0, NULL, 0}
     };
     CK_ULONG unwrap_tmpl_num;
     CK_ATTRIBUTE getattr_tmpl[] = {
@@ -767,7 +1058,8 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
     }
 
     if (wrap_mech->mechanism == CKM_AES_KEY_WRAP &&
-        tsuite->wrapped_key_gen_mech.mechanism == CKM_GENERIC_SECRET_KEY_GEN &&
+        (tsuite->wrapped_key_gen_mech.mechanism == CKM_GENERIC_SECRET_KEY_GEN ||
+         tsuite->wrapped_key_gen_mech.mechanism == CKM_SHA_1_KEY_GEN) &&
         tsuite->sym_keylen < 16) {
         testcase_skip("Mechanism CKM_AES_KEY_WRAP can not wrap generic secret "
                       "keys of size %lu (min 16 bytes required)",
@@ -785,6 +1077,11 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
             goto testcase_cleanup;
         }
 
+        if (tsuite->rsa_modbits > 4096 && !rsa8k) {
+            testcase_skip("Tests with modbits='%lu' are not enabled",
+                           tsuite->rsa_modbits);
+            goto testcase_cleanup;
+        }
         if (!keysize_supported(slot_id1, tsuite->wrapped_key_gen_mech.mechanism,
                                tsuite->rsa_modbits)) {
             testcase_skip("Token in slot %lu cannot be used with modbits='%lu'",
@@ -836,7 +1133,8 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
             }
         }
 
-        rc = generate_RSA_PKCS_KeyPair(session1, tsuite->rsa_modbits,
+        rc = generate_RSA_PKCS_KeyPair(session1, CKM_RSA_PKCS_KEY_PAIR_GEN,
+                                       tsuite->rsa_modbits,
                                        tsuite->rsa_publ_exp,
                                        tsuite->rsa_publ_exp_len,
                                        &publ_key, &priv_key);
@@ -863,6 +1161,7 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
         break;
 
     case CKM_GENERIC_SECRET_KEY_GEN:
+    case CKM_SHA_1_KEY_GEN:
         rc = generate_SecretKey(session1, tsuite->sym_keylen,
                                 &tsuite->wrapped_key_gen_mech, &sym_key);
         break;
@@ -890,6 +1189,40 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
         }
         break;
 
+    case CKM_IBM_ML_DSA_KEY_PAIR_GEN:
+        rc = generate_IBM_ML_DSA_KeyPair(session1, tsuite->pqc_keyform,
+                                         &publ_key, &priv_key,
+                                         CK_TRUE); // must be extractable for Wrap/Unwrap
+
+        if (rc == CKR_KEY_SIZE_RANGE) {
+            testcase_skip("generate to be wrapped key %s with mech %s (%u) in "
+                          "slot %lu is not supported",
+                          tsuite->name,
+                          mech_to_str(tsuite->wrapped_key_gen_mech.mechanism),
+                          (unsigned int)tsuite->wrapped_key_gen_mech.mechanism,
+                          slot_id1);
+            rc = CKR_OK;
+            goto testcase_cleanup;
+        }
+        break;
+
+    case CKM_IBM_ML_KEM_KEY_PAIR_GEN:
+        rc = generate_IBM_ML_KEM_KeyPair(session1, tsuite->pqc_keyform,
+                                         &publ_key, &priv_key,
+                                         CK_TRUE); // must be extractable for Wrap/Unwrap
+
+        if (rc == CKR_KEY_SIZE_RANGE) {
+            testcase_skip("generate to be wrapped key %s with mech %s (%u) in "
+                          "slot %lu is not supported",
+                          tsuite->name,
+                          mech_to_str(tsuite->wrapped_key_gen_mech.mechanism),
+                          (unsigned int)tsuite->wrapped_key_gen_mech.mechanism,
+                          slot_id1);
+            rc = CKR_OK;
+            goto testcase_cleanup;
+        }
+        break;
+
     default:
         testcase_error("Testcase does not support %s (%u)",
                        mech_to_str(tsuite->wrapped_key_gen_mech.mechanism),
@@ -907,6 +1240,15 @@ CK_RV do_wrap_key_test(struct wrapped_mech_info *tsuite,
             rc = CKR_OK;
             goto testcase_cleanup;
         }
+        if (rc == CKR_KEY_SIZE_RANGE) {
+            testcase_skip("generate to be wrapped key of the key size with "
+                          "mech %s (%u) in slot %lu is not supported",
+                          mech_to_str(tsuite->wrapped_key_gen_mech.mechanism),
+                          (unsigned int)tsuite->wrapped_key_gen_mech.mechanism,
+                          slot_id1);
+            rc = CKR_OK;
+            goto testcase_cleanup;
+        }
 
         testcase_error("generate to be wrapped key with mech %s (%u) in slot "
                        "%lu failed, rc=%s",
@@ -1001,6 +1343,7 @@ do_wrap:
     case CKM_AES_KEY_GEN:
     case CKM_AES_XTS_KEY_GEN:
     case CKM_GENERIC_SECRET_KEY_GEN:
+    case CKM_SHA_1_KEY_GEN:
         unwrap_tmpl_num = 3;
         break;
    default:
@@ -1031,6 +1374,15 @@ do_wrap:
         break;
     }
 
+    if (tsuite->operation_mech.mechanism == CKM_IBM_ML_KEM ||
+        tsuite->operation_mech.mechanism == CKM_IBM_ML_KEM_WITH_ECDH) {
+        /* KEM needs CKA_DERIVE=TRUE in unwrap template */
+        unwrap_tmpl[unwrap_tmpl_num].type = CKA_DERIVE;
+        unwrap_tmpl[unwrap_tmpl_num].pValue = &ck_true;
+        unwrap_tmpl[unwrap_tmpl_num].ulValueLen = sizeof(ck_true);
+        unwrap_tmpl_num++;
+    }
+
     /* Unwrap the key on slot 2 */
     rc = funcs->C_UnwrapKey(session2, wrap_mech,
                             sym_wrap_key2 != CK_INVALID_HANDLE ?
@@ -1077,8 +1429,18 @@ do_wrap:
         if ((rc == CKR_MECHANISM_INVALID || rc == CKR_KEY_TYPE_INCONSISTENT) &&
             is_ep11_token(slot_id2) && wrap_mech->mechanism == CKM_DES3_CBC &&
             (key_type == CKK_EC || key_type == CKK_RSA ||
-             key_type == CKK_IBM_PQC_DILITHIUM)) {
-            testcase_skip("EP11 does not support unwrap of EC or RSA keys with DES3 CBC");
+             key_type == CKK_IBM_PQC_DILITHIUM || key_type == CKK_IBM_ML_DSA ||
+             key_type == CKK_IBM_ML_KEM)) {
+            testcase_skip("EP11 does not support unwrap of EC, RSA, or PQC keys with DES3 CBC");
+            rc = CKR_OK;
+            goto testcase_cleanup;
+        }
+        if (rc == CKR_ENCRYPTED_DATA_INVALID && is_ep11_token(slot_id2) &&
+            (key_type == CKK_IBM_PQC_DILITHIUM ||
+             key_type == CKK_IBM_PQC_KYBER ||
+             key_type == CKK_IBM_ML_DSA ||
+             key_type == CKK_IBM_ML_KEM)) {
+            testcase_skip("EP11 does not supported the PQC variant");
             rc = CKR_OK;
             goto testcase_cleanup;
         }
@@ -1091,7 +1453,7 @@ do_wrap:
     }
 
     if (key_type == CKK_AES || key_type == CKK_AES_XTS ||
-        key_type == CKK_GENERIC_SECRET) {
+        key_type == CKK_GENERIC_SECRET || key_type == CKK_SHA_1_HMAC) {
         /* Check if the unwrapped key has the desired key length */
         rc = funcs->C_GetAttributeValue(session2, unwrapped_key,
                                         getattr_tmpl, 1);
@@ -1160,7 +1522,7 @@ CK_RV do_wrapping_test(struct wrapping_mech_info *tsuite)
     CK_RV loc_rc, rc = CKR_OK;
     CK_ULONG i;
     char *s = NULL;
-    CK_BYTE modulus[512];
+    CK_BYTE modulus[OPENSSL_RSA_MAX_MODULUS_BITS / 8];
     CK_BYTE publ_exp[16];
     CK_BYTE key[64];
     CK_ULONG key_size = 0;
@@ -1243,6 +1605,11 @@ CK_RV do_wrapping_test(struct wrapping_mech_info *tsuite)
             goto testcase_cleanup;
         }
 
+        if (tsuite->rsa_modbits > 4096 && !rsa8k) {
+            testcase_skip("Tests with modbits='%lu' are not enabled",
+                           tsuite->rsa_modbits);
+            goto testcase_cleanup;
+        }
         if (!keysize_supported(slot_id1,
                                tsuite->wrapping_key_gen_mech.mechanism,
                                tsuite->rsa_modbits)) {
@@ -1295,7 +1662,8 @@ CK_RV do_wrapping_test(struct wrapping_mech_info *tsuite)
                 goto testcase_cleanup;
             }
         }
-        rc = generate_RSA_PKCS_KeyPair(session2, tsuite->rsa_modbits,
+        rc = generate_RSA_PKCS_KeyPair(session2, CKM_RSA_PKCS_KEY_PAIR_GEN,
+                                       tsuite->rsa_modbits,
                                        tsuite->rsa_publ_exp,
                                        tsuite->rsa_publ_exp_len,
                                        &publ_wrap_key2, &priv_wrap_key2);
@@ -1392,6 +1760,15 @@ CK_RV do_wrapping_test(struct wrapping_mech_info *tsuite)
             rc = CKR_OK;
             goto testcase_cleanup;
         }
+        if (rc == CKR_KEY_SIZE_RANGE) {
+            testcase_skip("generate wrapping key of the key size with mech %s "
+                          "(%u) in slot %lu is not supported",
+                          mech_to_str(tsuite->wrapping_key_gen_mech.mechanism),
+                          (unsigned int)tsuite->wrapping_key_gen_mech.mechanism,
+                          slot_id2);
+            rc = CKR_OK;
+            goto testcase_cleanup;
+        }
 
         testcase_error("generate wrapping key with mech %s (%u) in slot %lu "
                        "failed, rc=%s",
@@ -1485,6 +1862,12 @@ CK_RV do_wrapping_test(struct wrapping_mech_info *tsuite)
             rc = CKR_OK;
             goto testcase_cleanup;
         }
+        if (rc == CKR_KEY_SIZE_RANGE) {
+            testcase_skip("import wrapping key size is not supported in slot %lu",
+                          slot_id1);
+            rc = CKR_OK;
+            goto testcase_cleanup;
+        }
 
         testcase_error("import wrapping key in slot %lu failed, rc=%s",
                        slot_id1, p11_get_ckr(rc));
@@ -1579,18 +1962,18 @@ int main(int argc, char **argv)
                 return -1;
             }
             slot_id1 = atoi(argv[i]);
-        }
-        else if (strcmp(argv[i], "-slot2") == 0) {
+        } else if (strcmp(argv[i], "-slot2") == 0) {
             ++i;
             if (i >= argc) {
                 printf("Slot number missing\n");
                 return -1;
             }
             slot_id2 = atoi(argv[i]);
-        }
-
-        if (strcmp(argv[i], "-h") == 0) {
-            printf("usage:  %s [-slot1 <num>] [-slot2 <num>] [-h]\n\n",
+        } else if (strcmp(argv[i], "-rsa8k") == 0) {
+            ++i;
+            rsa8k = TRUE;
+        } else if (strcmp(argv[i], "-h") == 0) {
+            printf("usage:  %s [-slot1 <num>] [-slot2 <num>] [-rsa8k] [-h]\n\n",
                    argv[0]);
             printf("By default, Slot #1 and #2 are used\n\n");
             return -1;
