@@ -65,6 +65,7 @@
 #define MODE_MODIFY     (1 << 3)
 #define MODE_DERIVE     (1 << 4)
 #define MODE_UNWRAP     (1 << 5)
+#define MODE_UNWRAPPED  (1 << 6)
 
 // RSA block formatting types
 //
@@ -101,6 +102,9 @@ enum {
 #define AES_BLOCK_SIZE  16
 #define AES_INIT_VECTOR_SIZE AES_BLOCK_SIZE
 #define AES_COUNTER_SIZE        16
+#define AES_KEY_WRAP_BLOCK_SIZE 8
+#define AES_KEY_WRAP_IV_SIZE AES_KEY_WRAP_BLOCK_SIZE
+#define AES_KEY_WRAP_KWP_IV_SIZE 4
 
 #define DES_KEY_SIZE    8
 #define DES_BLOCK_SIZE  8
@@ -113,31 +117,22 @@ enum {
 
 #define SHA1_HASH_SIZE  20
 #define SHA1_BLOCK_SIZE 64
-#define SHA1_BLOCK_SIZE_MASK (SHA1_BLOCK_SIZE - 1)
 #define SHA224_HASH_SIZE  28
 #define SHA224_BLOCK_SIZE 64
-#define SHA224_BLOCK_SIZE_MASK (SHA224_BLOCK_SIZE - 1)
 #define SHA256_HASH_SIZE  32
 #define SHA256_BLOCK_SIZE 64
-#define SHA256_BLOCK_SIZE_MASK (SHA256_BLOCK_SIZE - 1)
 #define SHA384_HASH_SIZE  48
 #define SHA384_BLOCK_SIZE 128
-#define SHA384_BLOCK_SIZE_MASK (SHA384_BLOCK_SIZE - 1)
 #define SHA512_HASH_SIZE  64
 #define SHA512_BLOCK_SIZE 128
-#define SHA512_BLOCK_SIZE_MASK (SHA512_BLOCK_SIZE - 1)
 #define SHA3_224_HASH_SIZE SHA224_HASH_SIZE
 #define SHA3_224_BLOCK_SIZE 144
-#define SHA3_224_BLOCK_SIZE_MASK (SHA3_224_BLOCK_SIZE - 1)
 #define SHA3_256_HASH_SIZE SHA256_HASH_SIZE
 #define SHA3_256_BLOCK_SIZE 136
-#define SHA3_256_BLOCK_SIZE_MASK (SHA3_256_BLOCK_SIZE - 1)
 #define SHA3_384_HASH_SIZE SHA384_HASH_SIZE
 #define SHA3_384_BLOCK_SIZE 104
-#define SHA3_384_BLOCK_SIZE_MASK (SHA3_384_BLOCK_SIZE - 1)
 #define SHA3_512_HASH_SIZE SHA512_HASH_SIZE
 #define SHA3_512_BLOCK_SIZE 72
-#define SHA3_512_BLOCK_SIZE_MASK (SHA3_512_BLOCK_SIZE - 1)
 #define MAX_SHA_HASH_SIZE SHA512_HASH_SIZE
 #define MAX_SHA_BLOCK_SIZE SHA3_224_BLOCK_SIZE
 
