@@ -62,6 +62,10 @@ CK_RV digest_mgr_init(STDLL_TokData_t *tokdata,
     case CKM_SHA512:
     case CKM_SHA512_224:
     case CKM_SHA512_256:
+    case CKM_SHA3_224:
+    case CKM_SHA3_256:
+    case CKM_SHA3_384:
+    case CKM_SHA3_512:
     case CKM_IBM_SHA3_224:
     case CKM_IBM_SHA3_256:
     case CKM_IBM_SHA3_384:
@@ -150,7 +154,6 @@ CK_RV digest_mgr_cleanup(STDLL_TokData_t *tokdata, SESSION *sess,
     ctx->multi_init = FALSE;
     ctx->multi = FALSE;
     ctx->active = FALSE;
-    ctx->context_len = 0;
     ctx->state_unsaveable = FALSE;
     ctx->count_statistics = FALSE;
 
@@ -167,6 +170,7 @@ CK_RV digest_mgr_cleanup(STDLL_TokData_t *tokdata, SESSION *sess,
             free(ctx->context);
         ctx->context = NULL;
     }
+    ctx->context_len = 0;
     ctx->context_free_func = NULL;
 
     return CKR_OK;
@@ -227,6 +231,10 @@ CK_RV digest_mgr_digest(STDLL_TokData_t *tokdata,
     case CKM_SHA512:
     case CKM_SHA512_224:
     case CKM_SHA512_256:
+    case CKM_SHA3_224:
+    case CKM_SHA3_256:
+    case CKM_SHA3_384:
+    case CKM_SHA3_512:
     case CKM_IBM_SHA3_224:
     case CKM_IBM_SHA3_256:
     case CKM_IBM_SHA3_384:
@@ -304,6 +312,10 @@ CK_RV digest_mgr_digest_update(STDLL_TokData_t *tokdata,
     case CKM_SHA512:
     case CKM_SHA512_224:
     case CKM_SHA512_256:
+    case CKM_SHA3_224:
+    case CKM_SHA3_256:
+    case CKM_SHA3_384:
+    case CKM_SHA3_512:
     case CKM_IBM_SHA3_224:
     case CKM_IBM_SHA3_256:
     case CKM_IBM_SHA3_384:
@@ -356,7 +368,7 @@ CK_RV digest_mgr_digest_key(STDLL_TokData_t *tokdata,
      */
     if (token_specific.secure_key_token) {
         TRACE_ERROR("%s because its a secure key token\n",
-                    ock_err(CKR_KEY_INDIGESTIBLE));
+                    ock_err(ERR_KEY_INDIGESTIBLE));
         rc = CKR_KEY_INDIGESTIBLE;
         goto out;
     }
@@ -449,6 +461,10 @@ CK_RV digest_mgr_digest_final(STDLL_TokData_t *tokdata,
     case CKM_SHA512:
     case CKM_SHA512_224:
     case CKM_SHA512_256:
+    case CKM_SHA3_224:
+    case CKM_SHA3_256:
+    case CKM_SHA3_384:
+    case CKM_SHA3_512:
     case CKM_IBM_SHA3_224:
     case CKM_IBM_SHA3_256:
     case CKM_IBM_SHA3_384:
