@@ -237,6 +237,7 @@ CK_RV key_mgr_generate_key(STDLL_TokData_t *tokdata,
     OBJECT *key_obj = NULL;
     CK_ATTRIBUTE *new_attr = NULL;
     CK_ULONG keyclass, subclass = 0;
+    CK_KEY_TYPE keytype1, keytype2;
     CK_RV rc;
 
     if (!sess || !mech || !handle) {
@@ -333,6 +334,30 @@ CK_RV key_mgr_generate_key(STDLL_TokData_t *tokdata,
 
         subclass = CKK_GENERIC_SECRET;
         break;
+    case CKM_SHA_1_KEY_GEN:
+    case CKM_SHA224_KEY_GEN:
+    case CKM_SHA256_KEY_GEN:
+    case CKM_SHA384_KEY_GEN:
+    case CKM_SHA512_KEY_GEN:
+    case CKM_SHA512_224_KEY_GEN:
+    case CKM_SHA512_256_KEY_GEN:
+    case CKM_SHA3_224_KEY_GEN:
+    case CKM_SHA3_256_KEY_GEN:
+    case CKM_SHA3_384_KEY_GEN:
+    case CKM_SHA3_512_KEY_GEN:
+        rc = pkcsget_keytype_for_mech(mech->mechanism, &keytype1, &keytype2);
+        if (rc != CKR_OK) {
+            TRACE_ERROR("pkcsget_keytype_for_mech failed.\n");
+            return rc;
+        }
+
+        if (subclass != 0 && subclass != keytype1) {
+            TRACE_ERROR("%s\n", ock_err(ERR_TEMPLATE_INCONSISTENT));
+            return CKR_TEMPLATE_INCONSISTENT;
+        }
+
+        subclass = keytype1;
+        break;
     default:
         TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_INVALID));
         return CKR_MECHANISM_INVALID;
@@ -369,6 +394,17 @@ CK_RV key_mgr_generate_key(STDLL_TokData_t *tokdata,
         rc = ckm_aes_key_gen(tokdata, key_obj->template, TRUE);
         break;
     case CKM_GENERIC_SECRET_KEY_GEN:
+    case CKM_SHA_1_KEY_GEN:
+    case CKM_SHA224_KEY_GEN:
+    case CKM_SHA256_KEY_GEN:
+    case CKM_SHA384_KEY_GEN:
+    case CKM_SHA512_KEY_GEN:
+    case CKM_SHA512_224_KEY_GEN:
+    case CKM_SHA512_256_KEY_GEN:
+    case CKM_SHA3_224_KEY_GEN:
+    case CKM_SHA3_256_KEY_GEN:
+    case CKM_SHA3_384_KEY_GEN:
+    case CKM_SHA3_512_KEY_GEN:
         rc = ckm_generic_secret_key_gen(tokdata, key_obj->template);
         break;
     default:
@@ -575,6 +611,27 @@ CK_RV key_mgr_generate_key_pair(STDLL_TokData_t *tokdata,
         }
         subclass = CKK_IBM_DILITHIUM;
         break;
+    case CKM_IBM_ML_DSA_KEY_PAIR_GEN:
+        if (subclass != 0 && subclass != CKK_IBM_ML_DSA) {
+            TRACE_ERROR("%s\n", ock_err(ERR_TEMPLATE_INCONSISTENT));
+            return CKR_TEMPLATE_INCONSISTENT;
+        }
+        subclass = CKK_IBM_ML_DSA;
+        break;
+    case CKM_IBM_KYBER:
+        if (subclass != 0 && subclass != CKK_IBM_PQC_KYBER) {
+            TRACE_ERROR("%s\n", ock_err(ERR_TEMPLATE_INCONSISTENT));
+            return CKR_TEMPLATE_INCONSISTENT;
+        }
+        subclass = CKK_IBM_KYBER;
+        break;
+    case CKM_IBM_ML_KEM_KEY_PAIR_GEN:
+        if (subclass != 0 && subclass != CKK_IBM_ML_KEM) {
+            TRACE_ERROR("%s\n", ock_err(ERR_TEMPLATE_INCONSISTENT));
+            return CKR_TEMPLATE_INCONSISTENT;
+        }
+        subclass = CKK_IBM_ML_KEM;
+        break;
     default:
         TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_INVALID));
         return CKR_MECHANISM_INVALID;
@@ -627,8 +684,14 @@ CK_RV key_mgr_generate_key_pair(STDLL_TokData_t *tokdata,
 #endif
 /* End code contributed by Corrent corp. */
     case CKM_IBM_DILITHIUM:
-        rc = ckm_ibm_dilithium_key_pair_gen(tokdata, publ_key_obj->template,
-                                            priv_key_obj->template);
+    case CKM_IBM_ML_DSA_KEY_PAIR_GEN:
+        rc = ckm_ibm_ml_dsa_key_pair_gen(tokdata, mech, publ_key_obj->template,
+                                         priv_key_obj->template);
+        break;
+    case CKM_IBM_KYBER:
+    case CKM_IBM_ML_KEM_KEY_PAIR_GEN:
+        rc = ckm_ibm_ml_kem_key_pair_gen(tokdata, mech, publ_key_obj->template,
+                                         priv_key_obj->template);
         break;
     default:
         TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_INVALID));
@@ -1037,6 +1100,17 @@ CK_RV key_mgr_wrap_key(STDLL_TokData_t *tokdata,
         break;
 #endif
     case CKK_GENERIC_SECRET:
+    case CKK_SHA_1_HMAC:
+    case CKK_SHA224_HMAC:
+    case CKK_SHA256_HMAC:
+    case CKK_SHA384_HMAC:
+    case CKK_SHA512_HMAC:
+    case CKK_SHA3_224_HMAC:
+    case CKK_SHA3_256_HMAC:
+    case CKK_SHA3_384_HMAC:
+    case CKK_SHA3_512_HMAC:
+    case CKK_SHA512_224_HMAC:
+    case CKK_SHA512_256_HMAC:
         rc = generic_secret_wrap_get_data(key_obj->template, FALSE,
                                           &data, &data_len);
         if (rc != CKR_OK) {
@@ -1062,18 +1136,13 @@ CK_RV key_mgr_wrap_key(STDLL_TokData_t *tokdata,
         }
         break;
     case CKK_IBM_PQC_DILITHIUM:
-        rc = ibm_dilithium_priv_wrap_get_data(key_obj->template, FALSE,
-                                              &data, &data_len);
-        if (rc != CKR_OK) {
-            TRACE_DEVEL("ibm_dilithium_priv_wrap_get_data failed.\n");
-            goto done;
-        }
-        break;
     case CKK_IBM_PQC_KYBER:
-        rc = ibm_kyber_priv_wrap_get_data(key_obj->template, FALSE,
-                                          &data, &data_len);
+    case CKK_IBM_ML_DSA:
+    case CKK_IBM_ML_KEM:
+        rc = pqc_priv_wrap_get_data(key_obj->template, keytype, FALSE,
+                                    &data, &data_len);
         if (rc != CKR_OK) {
-            TRACE_DEVEL("ibm_kyber_priv_wrap_get_data failed.\n");
+            TRACE_DEVEL("pqc_priv_wrap_get_data failed.\n");
             goto done;
         }
         break;
@@ -1372,8 +1441,23 @@ CK_RV key_mgr_unwrap_key(STDLL_TokData_t *tokdata,
         case CKM_AES_XTS:
         case CKM_AES_GCM:
         case CKM_AES_KEY_WRAP:
-            if (keytype != CKK_AES && keytype != CKK_AES_XTS &&
-                keytype != CKK_GENERIC_SECRET) {
+            switch (keytype) {
+            case CKK_AES:
+            case CKK_AES_XTS:
+            case CKK_GENERIC_SECRET:
+            case CKK_SHA_1_HMAC:
+            case CKK_SHA224_HMAC:
+            case CKK_SHA256_HMAC:
+            case CKK_SHA384_HMAC:
+            case CKK_SHA512_HMAC:
+            case CKK_SHA3_224_HMAC:
+            case CKK_SHA3_256_HMAC:
+            case CKK_SHA3_384_HMAC:
+            case CKK_SHA3_512_HMAC:
+            case CKK_SHA512_224_HMAC:
+            case CKK_SHA512_256_HMAC:
+                break;
+            default:
                 TRACE_ERROR("The key type does not allow CKA_VALUE_LEN to be "
                             "specified in the unwrapping template.\n");
                 rc = CKR_TEMPLATE_INCONSISTENT;
@@ -1634,6 +1718,24 @@ CK_RV key_mgr_get_private_key_type(CK_BYTE *keydata,
             return CKR_OK;
         }
     }
+    // Check only the OBJECT IDENTIFIERs for ML-DSA
+    //
+    for (i = 0; ml_dsa_oids[i].oid != NULL; i++) {
+        if (alg_len == ml_dsa_oids[i].oid_len &&
+            memcmp(alg, ml_dsa_oids[i].oid, ml_dsa_oids[i].oid_len) == 0) {
+            *keytype = CKK_IBM_ML_DSA;
+            return CKR_OK;
+        }
+    }
+    // Check only the OBJECT IDENTIFIERs for ML-KEM
+    //
+    for (i = 0; ml_kem_oids[i].oid != NULL; i++) {
+        if (alg_len == ml_kem_oids[i].oid_len &&
+            memcmp(alg, ml_kem_oids[i].oid, ml_kem_oids[i].oid_len) == 0) {
+            *keytype = CKK_IBM_ML_KEM;
+            return CKR_OK;
+        }
+    }
 
     TRACE_ERROR("%s\n", ock_err(ERR_TEMPLATE_INCOMPLETE));
     return CKR_TEMPLATE_INCOMPLETE;
@@ -1755,6 +1857,8 @@ CK_RV key_mgr_derive_key(STDLL_TokData_t *tokdata,
     case CKM_SHA256_KEY_DERIVATION:
     case CKM_SHA384_KEY_DERIVATION:
     case CKM_SHA512_KEY_DERIVATION:
+    case CKM_SHA512_224_KEY_DERIVATION:
+    case CKM_SHA512_256_KEY_DERIVATION:
     case CKM_SHA3_224_KEY_DERIVATION:
     case CKM_SHA3_256_KEY_DERIVATION:
     case CKM_SHA3_384_KEY_DERIVATION:
@@ -1777,6 +1881,17 @@ CK_RV key_mgr_derive_key(STDLL_TokData_t *tokdata,
         rc = ckm_shake_derive(tokdata, sess, mech, base_key_obj, new_attrs,
                               new_attr_count, derived_key);
         break;
+    case CKM_IBM_KYBER:
+    case CKM_IBM_ML_KEM:
+    case CKM_IBM_ML_KEM_WITH_ECDH:
+        if (!derived_key) {
+            TRACE_ERROR("%s received bad argument(s)\n", __func__);
+            rc = CKR_FUNCTION_FAILED;
+            break;
+        }
+        rc = ibm_ml_kem_derive(tokdata, sess, mech, base_key_obj, new_attrs,
+                               new_attr_count, derived_key);
+        break;
     default:
         TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_INVALID));
         rc = CKR_MECHANISM_INVALID;
