@@ -13,12 +13,12 @@
  * Mechanisms for Elliptic Curve (EC)
  */
 
-#define _GNU_SOURCE
-#include <endian.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdint.h>
 #include <string.h>
+
+#include "platform.h"
 #include "pkcs11types.h"
 #include "defs.h"
 #include "host_defs.h"
@@ -28,6 +28,7 @@
 #include "trace.h"
 #include "tok_specific.h"
 #include "ec_defs.h"
+#include "p11util.h"
 
 #include "openssl/obj_mac.h"
 #include <openssl/ec.h>
@@ -275,24 +276,10 @@ CK_RV ec_hash_sign(STDLL_TokData_t *tokdata,
     memset(&digest_ctx, 0x0, sizeof(digest_ctx));
     memset(&sign_ctx, 0x0, sizeof(sign_ctx));
 
-    switch (ctx->mech.mechanism) {
-    case CKM_ECDSA_SHA1:
-        digest_mech.mechanism = CKM_SHA_1;
-        break;
-    case CKM_ECDSA_SHA224:
-        digest_mech.mechanism = CKM_SHA224;
-        break;
-    case CKM_ECDSA_SHA256:
-        digest_mech.mechanism = CKM_SHA256;
-        break;
-    case CKM_ECDSA_SHA384:
-        digest_mech.mechanism = CKM_SHA384;
-        break;
-    case CKM_ECDSA_SHA512:
-        digest_mech.mechanism = CKM_SHA512;
-        break;
-    default:
-        return CKR_MECHANISM_INVALID;
+    rc = get_digest_from_mech(ctx->mech.mechanism, &digest_mech.mechanism);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("%s get_digest_from_mech failed\n", __func__);
+        return rc;
     }
 
     digest_mech.ulParameterLen = 0;
@@ -323,7 +310,7 @@ CK_RV ec_hash_sign(STDLL_TokData_t *tokdata,
     sign_mech.pParameter = NULL;
 
     rc = sign_mgr_init(tokdata, sess, &sign_ctx, &sign_mech, FALSE, ctx->key,
-                       FALSE);
+                       FALSE, FALSE);
     if (rc != CKR_OK) {
         TRACE_DEVEL("Sign Mgr Init failed.\n");
         goto error;
@@ -356,24 +343,10 @@ CK_RV ec_hash_sign_update(STDLL_TokData_t *tokdata,
     context = (RSA_DIGEST_CONTEXT *) ctx->context;
 
     if (context->flag == FALSE) {
-        switch (ctx->mech.mechanism) {
-        case CKM_ECDSA_SHA1:
-            digest_mech.mechanism = CKM_SHA_1;
-            break;
-        case CKM_ECDSA_SHA224:
-            digest_mech.mechanism = CKM_SHA224;
-            break;
-        case CKM_ECDSA_SHA256:
-            digest_mech.mechanism = CKM_SHA256;
-            break;
-        case CKM_ECDSA_SHA384:
-            digest_mech.mechanism = CKM_SHA384;
-            break;
-        case CKM_ECDSA_SHA512:
-            digest_mech.mechanism = CKM_SHA512;
-            break;
-        default:
-            return CKR_MECHANISM_INVALID;
+        rc = get_digest_from_mech(ctx->mech.mechanism, &digest_mech.mechanism);
+        if (rc != CKR_OK) {
+            TRACE_ERROR("%s get_digest_from_mech failed\n", __func__);
+            return rc;
         }
 
         digest_mech.ulParameterLen = 0;
@@ -446,7 +419,7 @@ CK_RV ec_hash_sign_final(STDLL_TokData_t *tokdata,
     sign_mech.pParameter = NULL;
 
     rc = sign_mgr_init(tokdata, sess, &sign_ctx, &sign_mech, FALSE, ctx->key,
-                       FALSE);
+                       FALSE, FALSE);
     if (rc != CKR_OK) {
         TRACE_DEVEL("Sign Mgr Init failed.\n");
         goto done;
@@ -491,24 +464,10 @@ CK_RV ec_hash_verify(STDLL_TokData_t *tokdata,
     memset(&digest_ctx, 0x0, sizeof(digest_ctx));
     memset(&verify_ctx, 0x0, sizeof(verify_ctx));
 
-    switch (ctx->mech.mechanism) {
-    case CKM_ECDSA_SHA1:
-        digest_mech.mechanism = CKM_SHA_1;
-        break;
-    case CKM_ECDSA_SHA224:
-        digest_mech.mechanism = CKM_SHA224;
-        break;
-    case CKM_ECDSA_SHA256:
-        digest_mech.mechanism = CKM_SHA256;
-        break;
-    case CKM_ECDSA_SHA384:
-        digest_mech.mechanism = CKM_SHA384;
-        break;
-    case CKM_ECDSA_SHA512:
-        digest_mech.mechanism = CKM_SHA512;
-        break;
-    default:
-        return CKR_MECHANISM_INVALID;
+    rc = get_digest_from_mech(ctx->mech.mechanism, &digest_mech.mechanism);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("%s get_digest_from_mech failed\n", __func__);
+        return rc;
     }
 
     digest_mech.ulParameterLen = 0;
@@ -574,24 +533,10 @@ CK_RV ec_hash_verify_update(STDLL_TokData_t *tokdata,
     context = (RSA_DIGEST_CONTEXT *) ctx->context;
 
     if (context->flag == FALSE) {
-        switch (ctx->mech.mechanism) {
-        case CKM_ECDSA_SHA1:
-            digest_mech.mechanism = CKM_SHA_1;
-            break;
-        case CKM_ECDSA_SHA224:
-            digest_mech.mechanism = CKM_SHA224;
-            break;
-        case CKM_ECDSA_SHA256:
-            digest_mech.mechanism = CKM_SHA256;
-            break;
-        case CKM_ECDSA_SHA384:
-            digest_mech.mechanism = CKM_SHA384;
-            break;
-        case CKM_ECDSA_SHA512:
-            digest_mech.mechanism = CKM_SHA512;
-            break;
-        default:
-            return CKR_MECHANISM_INVALID;
+        rc = get_digest_from_mech(ctx->mech.mechanism, &digest_mech.mechanism);
+        if (rc != CKR_OK) {
+            TRACE_ERROR("%s get_digest_from_mech failed\n", __func__);
+            return rc;
         }
 
         digest_mech.ulParameterLen = 0;
@@ -677,8 +622,9 @@ done:
     return rc;
 }
 
-CK_RV ckm_kdf(STDLL_TokData_t *tokdata, SESSION *sess, CK_ULONG kdf,
-              CK_BYTE *data, CK_ULONG data_len, CK_BYTE *hash, CK_ULONG *h_len)
+static CK_RV ckm_kdf(STDLL_TokData_t *tokdata, SESSION *sess, CK_ULONG kdf,
+                     CK_BYTE *data, CK_ULONG data_len,
+                     CK_BYTE *hash, CK_ULONG *h_len)
 {
     CK_RV rc;
     DIGEST_CONTEXT ctx;
@@ -687,33 +633,15 @@ CK_RV ckm_kdf(STDLL_TokData_t *tokdata, SESSION *sess, CK_ULONG kdf,
     memset(&ctx, 0, sizeof(DIGEST_CONTEXT));
     memset(&digest_mech, 0, sizeof(CK_MECHANISM));
 
-    switch (kdf) {
-    case CKD_SHA1_KDF:
-        digest_mech.mechanism = CKM_SHA_1;
-        *h_len = SHA1_HASH_SIZE;
-        break;
-    case CKD_SHA224_KDF:
-        digest_mech.mechanism = CKM_SHA224;
-        *h_len = SHA224_HASH_SIZE;
-        break;
-    case CKD_SHA256_KDF:
-        digest_mech.mechanism = CKM_SHA256;
-        *h_len = SHA256_HASH_SIZE;
-        break;
-    case CKD_SHA384_KDF:
-        digest_mech.mechanism = CKM_SHA384;
-        *h_len = SHA384_HASH_SIZE;
-        break;
-    case CKD_SHA512_KDF:
-        digest_mech.mechanism = CKM_SHA512;
-        *h_len = SHA512_HASH_SIZE;
-        break;
-    case CKD_NULL:
-        memcpy(hash, data, data_len - 4);
-        *h_len = data_len - 4;  // data length minus counter length
-        return CKR_OK;
-    default:
-        TRACE_ERROR("%s\n", ock_err(ERR_FUNCTION_NOT_SUPPORTED));
+    rc = digest_from_kdf(kdf, &digest_mech.mechanism);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("digest_from_kdf failed\n");
+        return CKR_FUNCTION_NOT_SUPPORTED;
+    }
+
+    rc = get_sha_size(digest_mech.mechanism, h_len);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("get_sha_size failed\n");
         return CKR_FUNCTION_NOT_SUPPORTED;
     }
 
@@ -726,7 +654,7 @@ CK_RV ckm_kdf(STDLL_TokData_t *tokdata, SESSION *sess, CK_ULONG kdf,
     rc = digest_mgr_digest(tokdata, sess, FALSE, &ctx, data, data_len, hash,
                            h_len);
     if (rc != CKR_OK) {
-        TRACE_ERROR("digest_mgr_digest failed with rc = %s\n", ock_err(rc));
+        TRACE_ERROR("digest_mgr_digest failed with rc=0x%lx\n", rc);
         digest_mgr_cleanup(tokdata, sess, &ctx);
         return rc;
     }
@@ -739,13 +667,14 @@ CK_RV ckm_kdf_X9_63(STDLL_TokData_t *tokdata, SESSION *sess, CK_ULONG kdf,
                     const CK_BYTE *shared_data, CK_ULONG shared_data_len,
                     CK_BYTE *key, CK_ULONG key_len)
 {
-    CK_ULONG counter_length = 4;
+    CK_ULONG counter_length = sizeof(uint32_t);
     CK_BYTE *ctx = NULL;
     CK_ULONG ctx_len;
     CK_BYTE hash[MAX_SUPPORTED_HASH_LENGTH];
     CK_ULONG h_len;
     CK_RV rc;
-    unsigned int i, counter, counter_en;
+    uint32_t counter, counter_en;
+    unsigned int i;
 
     /* Check max keylen according to ANSI X9.63 */
     /* digest_len * 2^32 */
@@ -756,18 +685,15 @@ CK_RV ckm_kdf_X9_63(STDLL_TokData_t *tokdata, SESSION *sess, CK_ULONG kdf,
         return CKR_KEY_SIZE_RANGE;
     }
 
-    /* If no KDF to be used, just return the shared_data.
-     * Cannot concatenate hashes. */
-    if (kdf == CKD_NULL) {
-        memcpy(key, z, z_len);
-        return CKR_OK;
-    }
-
-    /* Allocate memory for hash context */
+    /*
+     * Allocate memory for hash context:
+     *  <secret> || <be32(counter 1..N)> || <sharedinfo>
+     */
     ctx_len = z_len + counter_length + shared_data_len;
     ctx = malloc(ctx_len);
     if (!ctx)
         return CKR_HOST_MEMORY;
+
     memcpy(ctx, z, z_len);
     if (shared_data_len > 0)
         memcpy(ctx + z_len + counter_length, shared_data, shared_data_len);
@@ -790,11 +716,70 @@ CK_RV ckm_kdf_X9_63(STDLL_TokData_t *tokdata, SESSION *sess, CK_ULONG kdf,
     return CKR_OK;
 }
 
+CK_RV ckm_kdf_sp800_56c(STDLL_TokData_t *tokdata, SESSION *sess, CK_ULONG kdf,
+                        CK_ULONG kdf_digest_len,
+                        const CK_BYTE *z, CK_ULONG z_len,
+                        const CK_BYTE *shared_data, CK_ULONG shared_data_len,
+                        uint32_t purpose, CK_BYTE *key, CK_ULONG key_len)
+{
+    CK_ULONG counter_length = sizeof(uint32_t);
+    CK_ULONG purpose_length = purpose > 0 ? sizeof(uint32_t) : 0;
+    CK_BYTE *ctx = NULL;
+    CK_ULONG ctx_len;
+    CK_BYTE hash[MAX_SUPPORTED_HASH_LENGTH];
+    CK_ULONG h_len;
+    CK_RV rc;
+    uint32_t counter, counter_en, purpose_en = htobe32(purpose);
+    unsigned int i;
+
+    /* Check max keylen according to FIPS SP800-56C */
+    /* digest_len * 2^32 */
+    CK_ULONG max_keybytes = kdf_digest_len * 0x100000000ul;
+    if (key_len >= max_keybytes) {
+        TRACE_ERROR("Desired key length %lu greater than max supported key "
+                    "length %lu.\n", key_len, max_keybytes);
+        return CKR_KEY_SIZE_RANGE;
+    }
+
+    /*
+     * Allocate memory for hash context:
+     *  be32(counter) || [be32(purpose)] || <secret> || <fixed.info>
+     */
+    ctx_len = counter_length + purpose_length + z_len + shared_data_len;
+    ctx = malloc(ctx_len);
+    if (!ctx)
+        return CKR_HOST_MEMORY;
+
+    if (purpose > 0)
+        memcpy(ctx + counter_length, &purpose_en, purpose_length);
+    memcpy(ctx + counter_length + purpose_length, z, z_len);
+    if (shared_data_len > 0)
+        memcpy(ctx + counter_length + purpose_length + z_len,
+               shared_data, shared_data_len);
+
+    /* Provide key bytes according to ANSI X9.63 */
+    counter = 1;
+    for (i = 0; i < key_len / kdf_digest_len; i++) {
+        counter_en = htobe32(counter);
+        memcpy(ctx, &counter_en, counter_length);
+        rc = ckm_kdf(tokdata, sess, kdf, ctx, ctx_len, hash, &h_len);
+        if (rc != 0) {
+            free(ctx);
+            return rc;
+        }
+        memcpy(key + i * kdf_digest_len, hash, kdf_digest_len);
+        counter++;
+    }
+
+    free(ctx);
+    return CKR_OK;
+}
+
 CK_RV ckm_ecdh_pkcs_derive(STDLL_TokData_t *tokdata, SESSION *sess,
                            CK_VOID_PTR other_pubkey, CK_ULONG other_pubkey_len,
                            OBJECT* base_key_obj,
                            CK_BYTE *secret_value, CK_ULONG *secret_value_len,
-                           CK_MECHANISM_PTR mech)
+                           CK_MECHANISM_PTR mech, CK_BBOOL count_statistic)
 {
     CK_RV rc;
     CK_ATTRIBUTE *attr;
@@ -851,7 +836,7 @@ CK_RV ckm_ecdh_pkcs_derive(STDLL_TokData_t *tokdata, SESSION *sess,
     }
 
 done:
-    if (rc == CKR_OK)
+    if (count_statistic == TRUE && rc == CKR_OK)
         INC_COUNTER(tokdata, sess, mech, base_key_obj, POLICY_STRENGTH_IDX_0);
 
     return rc;
@@ -925,7 +910,7 @@ CK_RV ecdh_get_derived_key_size(CK_ULONG prime_len, CK_BYTE *curve_oid,
      */
     if (*key_len == 0) {
         /* Determine digest length */
-        if (kdf != CKD_NULL) {
+        if (kdf != CKD_NULL && kdf != CKD_IBM_HYBRID_NULL) {
             rc = digest_from_kdf(kdf, &digest_mech);
             if (rc != CKR_OK) {
                 TRACE_ERROR("Cannot determine mech from kdf.\n");
@@ -965,7 +950,8 @@ CK_RV ecdh_get_derived_key_size(CK_ULONG prime_len, CK_BYTE *curve_oid,
     }
 
     /* If no KDF used, max possible key length is the prime_len */
-    if (kdf == CKD_NULL && *key_len > prime_len) {
+    if ((kdf == CKD_NULL || kdf == CKD_IBM_HYBRID_NULL) &&
+        *key_len > prime_len) {
         TRACE_ERROR("Can only provide %ld key bytes without a KDF, "
                     "but %ld bytes requested.\n", prime_len, *key_len);
         return CKR_ARGUMENTS_BAD;
@@ -977,7 +963,8 @@ CK_RV ecdh_get_derived_key_size(CK_ULONG prime_len, CK_BYTE *curve_oid,
 CK_RV ecdh_pkcs_derive(STDLL_TokData_t *tokdata, SESSION *sess,
                        CK_MECHANISM *mech, OBJECT *base_key_obj,
                        CK_ATTRIBUTE *pTemplate, CK_ULONG ulCount,
-                       CK_OBJECT_HANDLE *derived_key_obj)
+                       CK_OBJECT_HANDLE *derived_key_obj,
+                       CK_BBOOL count_statistic)
 {
     CK_RV rc;
     CK_ULONG class = 0, keytype = 0, key_len = 0;
@@ -1016,21 +1003,41 @@ CK_RV ecdh_pkcs_derive(STDLL_TokData_t *tokdata, SESSION *sess,
         return CKR_TEMPLATE_INCOMPLETE;
     }
 
-
     /* Optional shared data can only be provided together with a KDF */
-    if (pParms->kdf == CKD_NULL
+    if ((pParms->kdf == CKD_NULL || pParms->kdf == CKD_IBM_HYBRID_NULL)
         && (pParms->pSharedData != NULL || pParms->ulSharedDataLen != 0)) {
         TRACE_ERROR("No KDF specified, but shared data ptr is not NULL.\n");
         return CKR_MECHANISM_PARAM_INVALID;
     }
 
+    rc = object_mgr_create_skel(tokdata, sess, pTemplate, ulCount, MODE_DERIVE,
+                                class, keytype, &temp_obj);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("Object Mgr create skeleton failed, rc=0x%lx.\n", rc);
+        return rc;
+    }
+
+    if (token_specific.t_ecdh_pkcs_derive_kdf != NULL) {
+        rc = token_specific.t_ecdh_pkcs_derive_kdf(tokdata, sess, base_key_obj,
+                                                   pParms, temp_obj,
+                                                   class, keytype);
+        if (rc != CKR_OK) {
+            TRACE_ERROR("t_ecdh_pkcs_derive_kdf failed, rc=0x%lx.\n", rc);
+            goto end;
+        }
+
+        INC_COUNTER(tokdata, sess, mech, base_key_obj, POLICY_STRENGTH_IDX_0);
+
+        goto derive_done;
+    }
+
     /* Derive the shared secret */
     rc = ckm_ecdh_pkcs_derive(tokdata, sess, pParms->pPublicData,
                               pParms->ulPublicDataLen, base_key_obj, z_value,
-                              &z_len, mech);
+                              &z_len, mech, count_statistic);
     if (rc != CKR_OK) {
         TRACE_ERROR("Error deriving the shared secret.\n");
-        return rc;
+        goto end;
     }
 
     /* Determine derived key length */
@@ -1038,27 +1045,29 @@ CK_RV ecdh_pkcs_derive(STDLL_TokData_t *tokdata, SESSION *sess,
                                      &key_len);
     if (rc == CKR_ATTRIBUTE_VALUE_INVALID) {
         TRACE_ERROR("%s\n", ock_err(ERR_ATTRIBUTE_VALUE_INVALID));
-        return rc;
+        goto end;
     }
 
     rc = ecdh_get_derived_key_size(z_len, NULL, 0, pParms->kdf, keytype,
                                    key_len, &key_len);
     if (rc != CKR_OK) {
         TRACE_ERROR("Can not determine the derived key length\n");
-        return rc;
+        goto end;
     }
 
     /* Determine digest length */
-    if (pParms->kdf != CKD_NULL) {
+    if (pParms->kdf != CKD_NULL && pParms->kdf != CKD_IBM_HYBRID_NULL) {
         rc = digest_from_kdf(pParms->kdf, &digest_mech);
         if (rc != CKR_OK) {
             TRACE_ERROR("Cannot determine mech from kdf.\n");
-            return CKR_ARGUMENTS_BAD;
+            rc = CKR_MECHANISM_PARAM_INVALID;
+            goto end;
         }
         rc = get_sha_size(digest_mech, &kdf_digest_len);
         if (rc != CKR_OK) {
             TRACE_ERROR("Cannot determine SHA digest size.\n");
-            return CKR_ARGUMENTS_BAD;
+            rc = CKR_MECHANISM_PARAM_INVALID;
+            goto end;
         }
     } else {
         kdf_digest_len = z_len;
@@ -1070,34 +1079,91 @@ CK_RV ecdh_pkcs_derive(STDLL_TokData_t *tokdata, SESSION *sess,
     if (!derived_key) {
         TRACE_ERROR("Cannot allocate %lu bytes for derived key.\n",
                     derived_key_len);
-        return CKR_HOST_MEMORY;
+        rc = CKR_HOST_MEMORY;
+        goto end;
     }
 
     /* Apply KDF function to shared secret */
-    rc = ckm_kdf_X9_63(tokdata, sess, pParms->kdf, kdf_digest_len,
-                       z_value, z_len, pParms->pSharedData,
-                       pParms->ulSharedDataLen, derived_key, derived_key_len);
-    if (rc != CKR_OK)
+    switch (pParms->kdf) {
+    case CKD_NULL:
+    case CKD_IBM_HYBRID_NULL:
+        memcpy(derived_key, z_value, z_len);
+        break;
+
+    case CKD_SHA1_KDF:
+    case CKD_SHA224_KDF:
+    case CKD_SHA256_KDF:
+    case CKD_SHA384_KDF:
+    case CKD_SHA512_KDF:
+    case CKD_SHA3_224_KDF:
+    case CKD_SHA3_256_KDF:
+    case CKD_SHA3_384_KDF:
+    case CKD_SHA3_512_KDF:
+    case CKD_IBM_HYBRID_SHA1_KDF:
+    case CKD_IBM_HYBRID_SHA224_KDF:
+    case CKD_IBM_HYBRID_SHA256_KDF:
+    case CKD_IBM_HYBRID_SHA384_KDF:
+    case CKD_IBM_HYBRID_SHA512_KDF:
+        rc = ckm_kdf_X9_63(tokdata, sess, pParms->kdf, kdf_digest_len,
+                           z_value, z_len, pParms->pSharedData,
+                           pParms->ulSharedDataLen, derived_key,
+                           derived_key_len);
+        if (rc != CKR_OK)
+            goto end;
+        break;
+
+    case CKD_SHA1_KDF_SP800:
+    case CKD_SHA224_KDF_SP800:
+    case CKD_SHA256_KDF_SP800:
+    case CKD_SHA384_KDF_SP800:
+    case CKD_SHA512_KDF_SP800:
+    case CKD_SHA3_224_KDF_SP800:
+    case CKD_SHA3_256_KDF_SP800:
+    case CKD_SHA3_384_KDF_SP800:
+    case CKD_SHA3_512_KDF_SP800:
+        rc = ckm_kdf_sp800_56c(tokdata, sess, pParms->kdf, kdf_digest_len,
+                               z_value, z_len, pParms->pSharedData,
+                               pParms->ulSharedDataLen, 0, derived_key,
+                               derived_key_len);
+        if (rc != CKR_OK)
+            goto end;
+        break;
+
+    default:
+        TRACE_ERROR("Unsupported KDF: 0x%lx\n", pParms->kdf);
+        rc = CKR_MECHANISM_PARAM_INVALID;
         goto end;
+    }
 
     /* Return the hashed and truncated derived bytes as CKA_VALUE attribute */
     rc = build_attribute(CKA_VALUE, derived_key, key_len, &value_attr);
     if (rc != CKR_OK) {
-        TRACE_ERROR("Failed to build the attribute from CKA_VALUE, rc=%s.\n",
-                    ock_err(rc));
+        TRACE_ERROR("Failed to build the attribute from CKA_VALUE, rc=0x%lx.\n",
+                    rc);
         goto end;
     }
 
     switch (keytype) {
     case CKK_GENERIC_SECRET:
+    case CKK_SHA_1_HMAC:
+    case CKK_SHA256_HMAC:
+    case CKK_SHA384_HMAC:
+    case CKK_SHA512_HMAC:
+    case CKK_SHA224_HMAC:
+    case CKK_SHA3_224_HMAC:
+    case CKK_SHA3_256_HMAC:
+    case CKK_SHA3_384_HMAC:
+    case CKK_SHA3_512_HMAC:
+    case CKK_SHA512_224_HMAC:
+    case CKK_SHA512_256_HMAC:
     case CKK_AES:
     case CKK_AES_XTS:
         /* Supply CKA_VAUE_LEN since this is required for those key types */
         rc = build_attribute(CKA_VALUE_LEN, (CK_BYTE*)&key_len,
                              sizeof(key_len), &vallen_attr);
         if (rc != CKR_OK) {
-            TRACE_ERROR("Failed to build the attribute from CKA_VALUE_LEN, rc=%s.\n",
-                        ock_err(rc));
+            TRACE_ERROR("Failed to build the attribute from CKA_VALUE_LEN, "
+                        "rc=0x%lx.\n", rc);
             free(value_attr);
             goto end;
         }
@@ -1106,17 +1172,6 @@ CK_RV ecdh_pkcs_derive(STDLL_TokData_t *tokdata, SESSION *sess,
         break;
     }
 
-    /* Create the object that will be passed back as a handle. This will contain
-     * the new (computed) value of the attribute. */
-    rc = object_mgr_create_skel(tokdata, sess, pTemplate, ulCount, MODE_KEYGEN,
-                                class, keytype, &temp_obj);
-    if (rc != CKR_OK) {
-        TRACE_ERROR("Object Mgr create skeleton failed, rc=%s.\n", ock_err(rc));
-        free(value_attr);
-        free(vallen_attr);
-        goto end;
-    }
-
     /* Update the template in the object with the new attribute */
     rc = template_update_attribute(temp_obj->template, value_attr);
     if (rc != CKR_OK) {
@@ -1135,20 +1190,23 @@ CK_RV ecdh_pkcs_derive(STDLL_TokData_t *tokdata, SESSION *sess,
         }
     }
 
+derive_done:
     /* At this point, the derived key is fully constructed...assign an object
      * handle and store the key */
     rc = object_mgr_create_final(tokdata, sess, temp_obj, derived_key_obj);
     if (rc != CKR_OK) {
-        TRACE_ERROR("Object Mgr create final failed, rc=%s.\n", ock_err(rc));
-        object_free(temp_obj);
-        temp_obj = NULL;
+        TRACE_ERROR("Object Mgr create final failed, rc=0x%lx.\n", rc);
         goto end;
     }
+    temp_obj = NULL;
 
     rc = CKR_OK;
 
 end:
-    free(derived_key);
+    if (derived_key != NULL)
+      free(derived_key);
+    if (temp_obj != NULL)
+        object_free(temp_obj);
 
     return rc;
 }
@@ -1600,3 +1658,453 @@ int ec_point_uncompressed_from_public_data(const CK_BYTE *data,
 
     return rc;
 }
+
+CK_RV ecdh_aes_key_wrap(STDLL_TokData_t *tokdata, SESSION *sess,
+                        CK_BBOOL length_only, ENCR_DECR_CONTEXT *ctx,
+                        CK_BYTE *in_data, CK_ULONG in_data_len,
+                        CK_BYTE *out_data, CK_ULONG *out_data_len)
+{
+    CK_ECDH_AES_KEY_WRAP_PARAMS *params;
+    CK_ATTRIBUTE *ec_params = NULL, *ec_point = NULL;
+    CK_MECHANISM ec_keygen_mech = { CKM_EC_KEY_PAIR_GEN, NULL, 0 };
+    CK_ECDH1_DERIVE_PARAMS ecdh_params = { 0 };
+    CK_MECHANISM ecdh_mech = { CKM_ECDH1_DERIVE, &ecdh_params,
+                               sizeof(ecdh_params) };
+    CK_MECHANISM aeskw_kwm_mech = { CKM_AES_KEY_WRAP_KWP, NULL, 0 };
+    ENCR_DECR_CONTEXT aeskw_ctx = { 0 };
+    CK_OBJECT_HANDLE ec_publ_key_handle = CK_INVALID_HANDLE;
+    CK_OBJECT_HANDLE ec_priv_key_handle = CK_INVALID_HANDLE;
+    CK_OBJECT_HANDLE aes_key_handle = CK_INVALID_HANDLE;
+    CK_OBJECT_CLASS aes_key_class = CKO_SECRET_KEY;
+    CK_KEY_TYPE aes_key_type = CKK_AES;
+    CK_ULONG aes_key_size, field_len = 0, pub_ec_point_len = 0;
+    CK_ULONG wrapped_key_len = 0, total_len;
+    CK_BYTE *pub_ec_point = NULL;
+    CK_BBOOL ck_true = TRUE;
+    CK_BBOOL ck_false = TRUE;
+    OBJECT *key_obj = NULL, *pub_key_obj = NULL;
+    CK_RV rc, rc2;
+
+    CK_ATTRIBUTE ec_publ_key_tmpl[] = {
+        { CKA_EC_PARAMS, NULL, 0 },
+        { CKA_HIDDEN, &ck_true, sizeof(ck_true) },
+        { CKA_TOKEN, &ck_false, sizeof(ck_false) },
+        { CKA_PRIVATE, &ck_true, sizeof(ck_true) },
+        { CKA_WRAP, &ck_false, sizeof(ck_false) },
+        { CKA_ENCRYPT, &ck_false, sizeof(ck_false) },
+        { CKA_VERIFY, &ck_false, sizeof(ck_false) },
+        { CKA_VERIFY_RECOVER, &ck_false, sizeof(ck_false) },
+        { CKA_DERIVE, &ck_true, sizeof(ck_true) },
+    };
+    CK_ATTRIBUTE ec_priv_key_tmpl[] = {
+        { CKA_HIDDEN, &ck_true, sizeof(ck_true) },
+        { CKA_SENSITIVE, &ck_true, sizeof(ck_true) },
+        { CKA_TOKEN, &ck_false, sizeof(ck_false) },
+        { CKA_PRIVATE, &ck_true, sizeof(ck_true) },
+        { CKA_UNWRAP, &ck_false, sizeof(ck_false) },
+        { CKA_DECRYPT, &ck_false, sizeof(ck_false) },
+        { CKA_SIGN, &ck_false, sizeof(ck_false) },
+        { CKA_SIGN_RECOVER, &ck_false, sizeof(ck_false) },
+        { CKA_DERIVE, &ck_true, sizeof(ck_true) },
+    };
+    CK_ATTRIBUTE aes_key_tmpl[] = {
+        { CKA_CLASS, &aes_key_class, sizeof(aes_key_class) },
+        { CKA_KEY_TYPE, &aes_key_type, sizeof(aes_key_type) },
+        { CKA_VALUE_LEN, &aes_key_size, sizeof(aes_key_size) },
+        { CKA_HIDDEN, &ck_true, sizeof(ck_true) },
+        { CKA_SENSITIVE, &ck_true, sizeof(ck_true) },
+        { CKA_TOKEN, &ck_false, sizeof(ck_false) },
+        { CKA_PRIVATE, &ck_true, sizeof(ck_true) },
+        { CKA_WRAP, &ck_false, sizeof(ck_false) },
+        { CKA_UNWRAP, &ck_true, sizeof(ck_true) },
+        { CKA_ENCRYPT, &ck_false, sizeof(ck_false) },
+        { CKA_DECRYPT, &ck_false, sizeof(ck_false) },
+        { CKA_SIGN, &ck_false, sizeof(ck_false) },
+        { CKA_VERIFY, &ck_false, sizeof(ck_false) },
+        { CKA_DERIVE, &ck_false, sizeof(ck_false) },
+    };
+
+    params = (CK_ECDH_AES_KEY_WRAP_PARAMS *)ctx->mech.pParameter;
+
+    /* Get the EC parameters of the EC public key used with this mechanism */
+    rc = object_mgr_find_in_map1(tokdata, ctx->key, &key_obj, READ_LOCK);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("Failed to acquire key from specified handle.\n");
+        if (rc == CKR_OBJECT_HANDLE_INVALID)
+            return CKR_KEY_HANDLE_INVALID;
+        else
+            return rc;
+    }
+
+    rc = template_attribute_get_non_empty(key_obj->template, CKA_EC_PARAMS,
+                                          &ec_params);
+    if (rc != CKR_OK) {
+        TRACE_DEVEL("Failed to get CKA_EC_PARAMS.\n");
+        goto done;
+    }
+
+    /* Generate a temporary EC key pair using the same EC parameters */
+    ec_publ_key_tmpl[0] = *ec_params;
+    rc = key_mgr_generate_key_pair(tokdata, sess, &ec_keygen_mech,
+                                   ec_publ_key_tmpl, sizeof(ec_publ_key_tmpl) /
+                                                       sizeof(CK_ATTRIBUTE),
+                                   ec_priv_key_tmpl, sizeof(ec_priv_key_tmpl) /
+                                                       sizeof(CK_ATTRIBUTE),
+                                   &ec_publ_key_handle, &ec_priv_key_handle,
+                                   FALSE);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("Failed to generate temporary EC key pair: "
+                    "%s (0x%lx)\n", p11_get_ckr(rc), rc);
+        goto done;
+    }
+
+    /* Perform ECDH to derive a shared AES key */
+    ecdh_params.kdf = params->kdf;
+    ecdh_params.pSharedData = params->pSharedData;
+    ecdh_params.ulSharedDataLen = params->ulSharedDataLen;
+
+    rc = template_attribute_get_non_empty(key_obj->template, CKA_EC_POINT,
+                                          &ec_point);
+    if (rc != CKR_OK) {
+        TRACE_DEVEL("Failed to get CKA_EC_POINT.\n");
+        goto done;
+    }
+
+    rc = ber_decode_OCTET_STRING((CK_BYTE *)ec_point->pValue,
+                                  &ecdh_params.pPublicData,
+                                  &ecdh_params.ulPublicDataLen,
+                                  &field_len);
+    if (rc != CKR_OK || field_len != ec_point->ulValueLen) {
+        rc = CKR_FUNCTION_FAILED;
+        TRACE_DEVEL("Failed to decode CKA_EC_POINT.\n");
+        goto done;
+    }
+
+    aes_key_size = params->ulAESKeyBits / 8;
+    rc = key_mgr_derive_key(tokdata, sess, &ecdh_mech,
+                            ec_priv_key_handle, &aes_key_handle,
+                            aes_key_tmpl,
+                            sizeof(aes_key_tmpl) / sizeof(CK_ATTRIBUTE),
+                            FALSE);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("Failed to derive temporary AES key (%lu bits): "
+                    "%s (0x%lx)\n", params->ulAESKeyBits, p11_get_ckr(rc), rc);
+        goto done;
+    }
+
+    /*
+     * Encrypt (wrap) the to-be-wrapped key as the second part of the wrapped
+     * key data (length-only operation to get the size of the second part).
+     */
+    rc = encr_mgr_init(tokdata, sess, &aeskw_ctx, OP_WRAP,
+                       &aeskw_kwm_mech, aes_key_handle, TRUE);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("Failed to encrypt the to-be-wrapped key: %s (0x%lx)\n",
+                    p11_get_ckr(rc), rc);
+        goto done;
+    }
+
+    rc = encr_mgr_encrypt(tokdata, sess, TRUE, &aeskw_ctx,
+                          in_data, in_data_len, NULL, &wrapped_key_len);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("Failed to encrypt the to-be-wrapped key: %s (0x%lx)\n",
+                    p11_get_ckr(rc), rc);
+        goto done;
+    }
+
+    /* Calculate the final length of the wrapped key data */
+    total_len = ecdh_params.ulPublicDataLen + wrapped_key_len;
+
+    if (length_only) {
+        *out_data_len = total_len;
+        goto done;
+    }
+
+    if (*out_data_len < total_len) {
+        *out_data_len = total_len;
+        TRACE_ERROR("%s\n", ock_err(ERR_BUFFER_TOO_SMALL));
+        rc = CKR_BUFFER_TOO_SMALL;
+        goto done;
+    }
+
+    /*
+     * Copy the (raw) EC point of the public transport EC key as first part of
+     * the wrapped key data.
+     */
+    rc = object_mgr_find_in_map1(tokdata, ec_publ_key_handle,
+                                 &pub_key_obj, READ_LOCK);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("Failed to acquire key from EC public key handle.\n");
+        if (rc == CKR_OBJECT_HANDLE_INVALID)
+            return CKR_KEY_HANDLE_INVALID;
+        else
+            return rc;
+    }
+
+    rc = template_attribute_get_non_empty(pub_key_obj->template, CKA_EC_POINT,
+                                          &ec_point);
+    if (rc != CKR_OK) {
+        TRACE_DEVEL("Failed to get CKA_EC_POINT.\n");
+        goto done;
+    }
+
+    rc = ber_decode_OCTET_STRING((CK_BYTE *)ec_point->pValue,
+                                  &pub_ec_point, &pub_ec_point_len, &field_len);
+    if (rc != CKR_OK || field_len != ec_point->ulValueLen) {
+        rc = CKR_FUNCTION_FAILED;
+        TRACE_DEVEL("Failed to decode CKA_EC_POINT.\n");
+        goto done;
+    }
+
+    memcpy(out_data, pub_ec_point, pub_ec_point_len);
+
+    /*
+     * Encrypt (wrap) the to-be-wrapped key as the second part of the wrapped
+     * key data.
+     */
+    rc = encr_mgr_encrypt(tokdata, sess, FALSE, &aeskw_ctx,
+                          in_data, in_data_len,
+                          out_data + ecdh_params.ulPublicDataLen,
+                          &wrapped_key_len);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("Failed to encrypt the to-be-wrapped key: %s (0x%lx)\n",
+                    p11_get_ckr(rc), rc);
+        goto done;
+    }
+
+    *out_data_len = total_len;
+
+done:
+    if (ec_publ_key_handle != CK_INVALID_HANDLE) {
+        rc2 = object_mgr_destroy_object(tokdata, sess, ec_publ_key_handle);
+        if (rc2 != CKR_OK) {
+            TRACE_ERROR("Failed to destroy temporary EC public key: %s "
+                        "(0x%lx)\n", p11_get_ckr(rc2), rc2);
+            if (rc == CKR_OK)
+                rc = rc2;
+        }
+    }
+    if (ec_priv_key_handle != CK_INVALID_HANDLE) {
+        rc2 = object_mgr_destroy_object(tokdata, sess, ec_priv_key_handle);
+        if (rc2 != CKR_OK) {
+            TRACE_ERROR("Failed to destroy temporary EC private key: %s "
+                        "(0x%lx)\n", p11_get_ckr(rc2), rc2);
+            if (rc == CKR_OK)
+                rc = rc2;
+        }
+    }
+    if (aes_key_handle != CK_INVALID_HANDLE) {
+        rc2 = object_mgr_destroy_object(tokdata, sess, aes_key_handle);
+        if (rc2 != CKR_OK) {
+            TRACE_ERROR("Failed to destroy temporary AES key: %s (0x%lx)\n",
+                        p11_get_ckr(rc2), rc2);
+            if (rc == CKR_OK)
+                rc = rc2;
+        }
+    }
+
+    encr_mgr_cleanup(tokdata, sess, &aeskw_ctx);
+
+    object_put(tokdata, key_obj, TRUE);
+    key_obj = NULL;
+    object_put(tokdata, pub_key_obj, TRUE);
+    pub_key_obj = NULL;
+
+    return rc;
+}
+
+CK_RV ecdh_aes_key_unwrap(STDLL_TokData_t *tokdata, SESSION *sess,
+                          CK_BBOOL length_only, ENCR_DECR_CONTEXT *ctx,
+                          CK_BYTE *in_data, CK_ULONG in_data_len,
+                          CK_BYTE *out_data, CK_ULONG *out_data_len)
+{
+    CK_ECDH_AES_KEY_WRAP_PARAMS *params;
+    CK_ECDH1_DERIVE_PARAMS ecdh_params = { 0 };
+    CK_MECHANISM ecdh_mech = { CKM_ECDH1_DERIVE, &ecdh_params,
+                               sizeof(ecdh_params) };
+    CK_MECHANISM aeskw_kwm_mech = { CKM_AES_KEY_WRAP_KWP, NULL, 0 };
+    CK_OBJECT_HANDLE aes_key_handle = CK_INVALID_HANDLE;
+    ENCR_DECR_CONTEXT aeskw_ctx = { 0 };
+    CK_BBOOL ck_true = TRUE;
+    CK_BBOOL ck_false = TRUE;
+    OBJECT *key_obj = NULL;
+    CK_ULONG prime_len = 0, public_data_len = 0;
+    CK_OBJECT_CLASS aes_key_class = CKO_SECRET_KEY;
+    CK_KEY_TYPE aes_key_type = CKK_AES;
+    CK_ULONG aes_key_size;
+    CK_BYTE form;
+    CK_RV rc, rc2;
+
+    CK_ATTRIBUTE aes_key_tmpl[] = {
+        { CKA_CLASS, &aes_key_class, sizeof(aes_key_class) },
+        { CKA_KEY_TYPE, &aes_key_type, sizeof(aes_key_type) },
+        { CKA_VALUE_LEN, &aes_key_size, sizeof(aes_key_size) },
+        { CKA_HIDDEN, &ck_true, sizeof(ck_true) },
+        { CKA_SENSITIVE, &ck_true, sizeof(ck_true) },
+        { CKA_TOKEN, &ck_false, sizeof(ck_false) },
+        { CKA_PRIVATE, &ck_true, sizeof(ck_true) },
+        { CKA_WRAP, &ck_false, sizeof(ck_false) },
+        { CKA_UNWRAP, &ck_true, sizeof(ck_true) },
+        { CKA_ENCRYPT, &ck_false, sizeof(ck_false) },
+        { CKA_DECRYPT, &ck_false, sizeof(ck_false) },
+        { CKA_SIGN, &ck_false, sizeof(ck_false) },
+        { CKA_VERIFY, &ck_false, sizeof(ck_false) },
+        { CKA_DERIVE, &ck_false, sizeof(ck_false) },
+    };
+
+    params = (CK_ECDH_AES_KEY_WRAP_PARAMS *)ctx->mech.pParameter;
+
+    /* Get the EC parameters of the EC key used with this mechanism */
+    rc = object_mgr_find_in_map1(tokdata, ctx->key, &key_obj, READ_LOCK);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("Failed to acquire key from specified handle.\n");
+        if (rc == CKR_OBJECT_HANDLE_INVALID)
+            return CKR_KEY_HANDLE_INVALID;
+        else
+            return rc;
+    }
+
+    rc = get_ecsiglen(key_obj, &prime_len);
+    if (rc != CKR_OK) {
+        TRACE_DEVEL("get_ecsiglen failed.\n");
+        goto done;
+    }
+    prime_len /= 2; /* prime length is half the size of an EC signature */
+
+    /* Input must be at least a compressed EC point plus wrapped data */
+    if (in_data_len <= prime_len + 1 + 2 * AES_KEY_WRAP_BLOCK_SIZE) {
+        TRACE_ERROR("%s\n", ock_err(ERR_ENCRYPTED_DATA_LEN_RANGE));
+        rc = CKR_ENCRYPTED_DATA_LEN_RANGE;
+        goto done;
+    }
+
+    form  = in_data[0] & ~0x01;
+    switch (form) {
+    case POINT_CONVERSION_COMPRESSED:
+        public_data_len = prime_len + 1;
+        /* Length checked above already */
+        break;
+    case POINT_CONVERSION_UNCOMPRESSED:
+    case POINT_CONVERSION_HYBRID:
+        public_data_len = 2 * prime_len + 1;
+        if (in_data_len < public_data_len + 2 * AES_KEY_WRAP_BLOCK_SIZE) {
+            TRACE_ERROR("%s\n", ock_err(ERR_ENCRYPTED_DATA_LEN_RANGE));
+            rc = CKR_ENCRYPTED_DATA_LEN_RANGE;
+            goto done;
+        }
+        break;
+    default:
+        TRACE_ERROR("%s\n", ock_err(ERR_ENCRYPTED_DATA_INVALID));
+        rc = CKR_ENCRYPTED_DATA_INVALID;
+        goto done;
+    }
+
+    /* Perform ECDH to derive a shared AES key */
+    ecdh_params.kdf = params->kdf;
+    ecdh_params.pSharedData = params->pSharedData;
+    ecdh_params.ulSharedDataLen = params->ulSharedDataLen;
+    ecdh_params.pPublicData = in_data;
+    ecdh_params.ulPublicDataLen = public_data_len;
+
+    aes_key_size = params->ulAESKeyBits / 8;
+    rc = key_mgr_derive_key(tokdata, sess, &ecdh_mech,
+                            ctx->key, &aes_key_handle,
+                            aes_key_tmpl,
+                            sizeof(aes_key_tmpl) / sizeof(CK_ATTRIBUTE),
+                            FALSE);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("Failed to derive temporary AES key (%lu bits): "
+                    "%s (0x%lx)\n", params->ulAESKeyBits, p11_get_ckr(rc), rc);
+        goto done;
+    }
+
+    /*
+     * Decrypt (unwrap) the final key data from the the second part of the
+     * wrapped key data.
+     */
+    rc = decr_mgr_init(tokdata, sess, &aeskw_ctx, OP_UNWRAP,
+                       &aeskw_kwm_mech, aes_key_handle, TRUE, FALSE);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("Failed to decrypt the to-be-unwrapped key: %s (0x%lx)\n",
+                    p11_get_ckr(rc), rc);
+        goto done;
+    }
+
+    rc = decr_mgr_decrypt(tokdata, sess, length_only, &aeskw_ctx,
+                          in_data + public_data_len,
+                          in_data_len - public_data_len,
+                          out_data, out_data_len);
+    if (rc != CKR_OK) {
+        TRACE_ERROR("Failed to decrypt the to-be-unwrapped key: %s (0x%lx)\n",
+                    p11_get_ckr(rc), rc);
+        goto done;
+    }
+
+done:
+    if (aes_key_handle != CK_INVALID_HANDLE) {
+        rc2 = object_mgr_destroy_object(tokdata, sess, aes_key_handle);
+        if (rc2 != CKR_OK) {
+            TRACE_ERROR("Failed to destroy temporary AES key: %s (0x%lx)\n",
+                        p11_get_ckr(rc2), rc2);
+            if (rc == CKR_OK)
+                rc = rc2;
+        }
+    }
+
+    encr_mgr_cleanup(tokdata, sess, &aeskw_ctx);
+
+    object_put(tokdata, key_obj, TRUE);
+    key_obj = NULL;
+
+    return rc;
+}
+
+CK_RV ec_agg_dup_param(CK_IBM_ECDSA_OTHER_BLS_PARAMS *from,
+                       CK_IBM_ECDSA_OTHER_BLS_PARAMS *to)
+{
+    CK_ULONG i;
+
+    if (from == NULL || to == NULL)
+        return CKR_ARGUMENTS_BAD;
+
+    if (from->numElements != 0 && from->elementSize > 0) {
+        for (i = 0; i < CK_IBM_BLS_MAX_AGGREGATIONS; i++) {
+            if (i < from->numElements) {
+                to->pAggrElements[i] = (CK_BYTE*)malloc(from->elementSize);
+                if (to->pAggrElements[i] == NULL) {
+                    TRACE_ERROR("%s\n", ock_err(ERR_HOST_MEMORY));
+                    ec_agg_free_param(to);
+                    return CKR_HOST_MEMORY;
+                }
+                memcpy(to->pAggrElements[i], from->pAggrElements[i],
+                       from->elementSize);
+            } else {
+                to->pAggrElements[i] = NULL;
+            }
+        }
+        to->numElements = from->numElements;
+        to->elementSize = from->elementSize;
+    } else {
+        to->numElements = 0;
+        to->elementSize = 0;
+        memset(to->pAggrElements, 0, sizeof(to->pAggrElements));
+    }
+    return CKR_OK;
+}
+
+CK_RV ec_agg_free_param(CK_IBM_ECDSA_OTHER_BLS_PARAMS *params)
+{
+    CK_ULONG i;
+
+    if (params == NULL)
+        return CKR_ARGUMENTS_BAD;
+
+    for (i = 0; i < CK_IBM_BLS_MAX_AGGREGATIONS; i++) {
+        free(params->pAggrElements[i]);
+        params->pAggrElements[i] = NULL;
+    }
+
+    params->numElements = 0;
+    params->elementSize = 0;
+    return CKR_OK;
+}
