@@ -140,7 +140,7 @@ CK_RV object_create(STDLL_TokData_t * tokdata,
 //
 // The old_obj must hold the READ lock!
 //
-CK_RV object_copy(STDLL_TokData_t * tokdata,
+CK_RV object_copy(STDLL_TokData_t * tokdata, SESSION *sess,
                   CK_ATTRIBUTE * pTemplate,
                   CK_ULONG ulCount, OBJECT * old_obj, OBJECT ** new_obj)
 {
@@ -181,6 +181,10 @@ CK_RV object_copy(STDLL_TokData_t * tokdata,
     if (rc != CKR_OK)
         goto error;
 
+    rc = object_init_ex_data_lock(o);
+    if (rc != CKR_OK)
+        goto error;
+
     // copy the original object's attribute template
     //
     rc = template_copy(o->template, old_obj->template);
@@ -228,7 +232,7 @@ CK_RV object_copy(STDLL_TokData_t * tokdata,
      * changed.
      */
     if (token_specific.t_set_attribute_values != NULL) {
-        rc = token_specific.t_set_attribute_values(tokdata, o, new_tmpl);
+        rc = token_specific.t_set_attribute_values(tokdata, sess, o, new_tmpl);
         if (rc != CKR_OK) {
             TRACE_DEVEL("token_specific_set_attribute_values failed with %lu\n",
                         rc);
@@ -330,6 +334,13 @@ void object_free(OBJECT * obj)
 {
     /* refactorization here to do actual free - fix from coverity scan */
     if (obj) {
+        if (obj->ex_data != NULL) {
+            if (obj->ex_data_free != NULL)
+                obj->ex_data_free(obj, obj->ex_data, obj->ex_data_len);
+            else
+                free(obj->ex_data);
+        }
+        object_destroy_ex_data_lock(obj);
         if (obj->template)
             template_free(obj->template);
         object_destroy_lock(obj);
@@ -620,7 +631,7 @@ CK_RV object_get_attribute_values(OBJECT * obj,
 
 // object_set_attribute_values()
 //
-CK_RV object_set_attribute_values(STDLL_TokData_t * tokdata,
+CK_RV object_set_attribute_values(STDLL_TokData_t * tokdata, SESSION *sess,
                                   OBJECT * obj,
                                   CK_ATTRIBUTE * pTemplate, CK_ULONG ulCount)
 {
@@ -669,7 +680,8 @@ CK_RV object_set_attribute_values(STDLL_TokData_t * tokdata,
     }
 
     if (token_specific.t_set_attribute_values != NULL) {
-        rc = token_specific.t_set_attribute_values(tokdata, obj, new_tmpl);
+        rc = token_specific.t_set_attribute_values(tokdata, sess,
+                                                   obj, new_tmpl);
         if (rc != CKR_OK) {
             TRACE_DEVEL("token_specific_set_attribute_values failed with %lu\n",
                         rc);
@@ -785,6 +797,12 @@ CK_RV object_restore_withSize(struct policy *policy,
         if (rc != CKR_OK)
             goto error;
 
+        rc = object_init_ex_data_lock(obj);
+        if (rc != CKR_OK) {
+            object_destroy_lock(obj);
+            goto error;
+        }
+
         *new_obj = obj;
     } else {
         /* Reload of existing object only changes the template */
@@ -892,13 +910,22 @@ CK_RV object_create_skel(STDLL_TokData_t * tokdata,
     if (rc != CKR_OK)
         goto done;
 
+    rc = object_init_ex_data_lock(o);
+    if (rc != CKR_OK) {
+        object_destroy_lock(o);
+        goto done;
+    }
+
     *obj = o;
 
     return CKR_OK;
 
 done:
-    if (o)
+    if (o) {
+        if (o->template)
+            template_free(o->template);
         free(o);
+    }
     if (tmpl)
         template_free(tmpl);
     if (tmpl2)
@@ -963,3 +990,55 @@ CK_RV object_unlock(OBJECT *obj)
 
     return CKR_OK;
 }
+
+CK_RV object_init_ex_data_lock(OBJECT *obj)
+{
+    if (pthread_rwlock_init(&obj->ex_data_rwlock, NULL) != 0) {
+        TRACE_DEVEL("Ex_data Lock init failed.\n");
+        return CKR_CANT_LOCK;
+    }
+
+    return CKR_OK;
+}
+
+CK_RV object_destroy_ex_data_lock(OBJECT *obj)
+{
+    if (pthread_rwlock_destroy(&obj->ex_data_rwlock) != 0) {
+        TRACE_DEVEL("Ex_data Lock destroy failed.\n");
+        return CKR_CANT_LOCK;
+    }
+
+    return CKR_OK;
+}
+
+CK_RV object_ex_data_lock(OBJECT *obj, OBJ_LOCK_TYPE type)
+{
+    switch (type) {
+    case NO_LOCK:
+        break;
+    case READ_LOCK:
+        if (pthread_rwlock_rdlock(&obj->ex_data_rwlock) != 0) {
+            TRACE_DEVEL("Ex_data Read-Lock failed.\n");
+            return CKR_CANT_LOCK;
+        }
+        break;
+    case WRITE_LOCK:
+        if (pthread_rwlock_wrlock(&obj->ex_data_rwlock) != 0) {
+            TRACE_DEVEL("Ex_data Write-Lock failed.\n");
+            return CKR_CANT_LOCK;
+        }
+        break;
+    }
+
+    return CKR_OK;
+}
+
+CK_RV object_ex_data_unlock(OBJECT *obj)
+{
+    if (pthread_rwlock_unlock(&obj->ex_data_rwlock) != 0) {
+        TRACE_DEVEL("Ex_data Unlock failed.\n");
+        return CKR_CANT_LOCK;
+    }
+
+    return CKR_OK;
+}
