@@ -228,18 +228,22 @@ const char *p11_get_cka(CK_ATTRIBUTE_TYPE atype)
         _sym2str(CKA_IBM_OPAQUE_PKEY);
         _sym2str(CKA_IBM_DILITHIUM_KEYFORM);
         _sym2str(CKA_IBM_DILITHIUM_MODE);
-        _sym2str(CKA_IBM_DILITHIUM_RHO);
-        _sym2str(CKA_IBM_DILITHIUM_SEED);
-        _sym2str(CKA_IBM_DILITHIUM_TR);
-        _sym2str(CKA_IBM_DILITHIUM_S1);
-        _sym2str(CKA_IBM_DILITHIUM_S2);
-        _sym2str(CKA_IBM_DILITHIUM_T0);
-        _sym2str(CKA_IBM_DILITHIUM_T1);
+        _sym2str(CKA_IBM_ML_DSA_RHO);
+        _sym2str(CKA_IBM_ML_DSA_SEED);
+        _sym2str(CKA_IBM_ML_DSA_TR);
+        _sym2str(CKA_IBM_ML_DSA_S1);
+        _sym2str(CKA_IBM_ML_DSA_S2);
+        _sym2str(CKA_IBM_ML_DSA_T0);
+        _sym2str(CKA_IBM_ML_DSA_T1);
+        _sym2str(CKA_IBM_ML_DSA_PRIVATE_SEED);
         _sym2str(CKA_IBM_PQC_PARAMS);
         _sym2str(CKA_IBM_KYBER_KEYFORM);
         _sym2str(CKA_IBM_KYBER_MODE);
-        _sym2str(CKA_IBM_KYBER_PK);
-        _sym2str(CKA_IBM_KYBER_SK);
+        _sym2str(CKA_IBM_ML_KEM_PK);
+        _sym2str(CKA_IBM_ML_KEM_SK);
+        _sym2str(CKA_IBM_ML_KEM_PRIVATE_SEED);
+        _sym2str(CKA_IBM_CCA_AES_KEY_MODE);
+        _sym2str(CKA_IBM_PARAMETER_SET);
     default:
         sprintf(buf, "unknown attribute type 0x%08lx", atype);
         return buf;
