@@ -167,9 +167,11 @@ CK_RV template_add_default_attributes(TEMPLATE *tmpl, TEMPLATE *basetmpl,
         case CKK_DH:
             return dh_publ_set_default_attributes(tmpl, mode);
         case CKK_IBM_PQC_DILITHIUM:
-            return ibm_dilithium_publ_set_default_attributes(tmpl, mode);
+        case CKK_IBM_ML_DSA:
+            return ibm_ml_dsa_publ_set_default_attributes(tmpl, mode, subclass);
         case CKK_IBM_PQC_KYBER:
-            return ibm_kyber_publ_set_default_attributes(tmpl, mode);
+        case CKK_IBM_ML_KEM:
+            return ibm_ml_kem_publ_set_default_attributes(tmpl, mode, subclass);
         default:
             TRACE_ERROR("%s: %lx\n", ock_err(ERR_ATTRIBUTE_VALUE_INVALID),
                         subclass);
@@ -186,9 +188,11 @@ CK_RV template_add_default_attributes(TEMPLATE *tmpl, TEMPLATE *basetmpl,
         case CKK_DH:
             return dh_priv_set_default_attributes(tmpl, mode);
         case CKK_IBM_PQC_DILITHIUM:
-            return ibm_dilithium_priv_set_default_attributes(tmpl, mode);
+        case CKK_IBM_ML_DSA:
+            return ibm_ml_dsa_priv_set_default_attributes(tmpl, mode, subclass);
         case CKK_IBM_PQC_KYBER:
-            return ibm_kyber_priv_set_default_attributes(tmpl, mode);
+        case CKK_IBM_ML_KEM:
+            return ibm_ml_kem_priv_set_default_attributes(tmpl, mode, subclass);
         default:
             TRACE_ERROR("%s: %lx\n", ock_err(ERR_ATTRIBUTE_VALUE_INVALID),
                         subclass);
@@ -197,7 +201,18 @@ CK_RV template_add_default_attributes(TEMPLATE *tmpl, TEMPLATE *basetmpl,
     case CKO_SECRET_KEY:
         switch (subclass) {
         case CKK_GENERIC_SECRET:
-            return generic_secret_set_default_attributes(tmpl, mode);
+        case CKK_SHA_1_HMAC:
+        case CKK_SHA224_HMAC:
+        case CKK_SHA256_HMAC:
+        case CKK_SHA384_HMAC:
+        case CKK_SHA512_HMAC:
+        case CKK_SHA3_224_HMAC:
+        case CKK_SHA3_256_HMAC:
+        case CKK_SHA3_384_HMAC:
+        case CKK_SHA3_512_HMAC:
+        case CKK_SHA512_224_HMAC:
+        case CKK_SHA512_256_HMAC:
+            return generic_secret_set_default_attributes(tmpl, mode, subclass);
         case CKK_DES:
             return des_set_default_attributes(tmpl, mode);
         case CKK_DES2:
@@ -418,9 +433,17 @@ CK_RV template_check_required_attributes(TEMPLATE *tmpl, CK_ULONG class,
         case CKK_DH:
             return dh_publ_check_required_attributes(tmpl, mode);
         case CKK_IBM_PQC_DILITHIUM:
-            return ibm_dilithium_publ_check_required_attributes(tmpl, mode);
+            return ibm_ml_dsa_publ_check_required_attributes(tmpl, mode,
+                                                             CKM_IBM_DILITHIUM);
         case CKK_IBM_PQC_KYBER:
-            return ibm_kyber_publ_check_required_attributes(tmpl, mode);
+            return ibm_ml_kem_publ_check_required_attributes(tmpl, mode,
+                                                             CKM_IBM_KYBER);
+        case CKK_IBM_ML_DSA:
+            return ibm_ml_dsa_publ_check_required_attributes(tmpl, mode,
+                                                             CKM_IBM_ML_DSA);
+        case CKK_IBM_ML_KEM:
+            return ibm_ml_kem_publ_check_required_attributes(tmpl, mode,
+                                                             CKM_IBM_ML_KEM);
         default:
             TRACE_ERROR("%s: %lx\n", ock_err(ERR_ATTRIBUTE_VALUE_INVALID),
                         subclass);
@@ -437,9 +460,17 @@ CK_RV template_check_required_attributes(TEMPLATE *tmpl, CK_ULONG class,
         case CKK_DH:
             return dh_priv_check_required_attributes(tmpl, mode);
         case CKK_IBM_PQC_DILITHIUM:
-            return ibm_dilithium_priv_check_required_attributes(tmpl, mode);
+            return ibm_ml_dsa_priv_check_required_attributes(tmpl, mode,
+                                                             CKM_IBM_DILITHIUM);
         case CKK_IBM_PQC_KYBER:
-            return ibm_kyber_priv_check_required_attributes(tmpl, mode);
+            return ibm_ml_kem_priv_check_required_attributes(tmpl, mode,
+                                                             CKM_IBM_KYBER);
+        case CKK_IBM_ML_DSA:
+            return ibm_ml_dsa_priv_check_required_attributes(tmpl, mode,
+                                                             CKM_IBM_ML_DSA);
+        case CKK_IBM_ML_KEM:
+            return ibm_ml_kem_priv_check_required_attributes(tmpl, mode,
+                                                             CKM_IBM_ML_KEM);
         default:
             TRACE_ERROR("%s: %lx\n", ock_err(ERR_ATTRIBUTE_VALUE_INVALID),
                         subclass);
@@ -448,6 +479,17 @@ CK_RV template_check_required_attributes(TEMPLATE *tmpl, CK_ULONG class,
     } else if (class == CKO_SECRET_KEY) {
         switch (subclass) {
         case CKK_GENERIC_SECRET:
+        case CKK_SHA_1_HMAC:
+        case CKK_SHA224_HMAC:
+        case CKK_SHA256_HMAC:
+        case CKK_SHA384_HMAC:
+        case CKK_SHA512_HMAC:
+        case CKK_SHA3_224_HMAC:
+        case CKK_SHA3_256_HMAC:
+        case CKK_SHA3_384_HMAC:
+        case CKK_SHA3_512_HMAC:
+        case CKK_SHA512_224_HMAC:
+        case CKK_SHA512_256_HMAC:
             return generic_secret_check_required_attributes(tmpl, mode);
         case CKK_DES:
             return des_check_required_attributes(tmpl, mode);
@@ -1390,9 +1432,11 @@ CK_BBOOL template_check_exportability(TEMPLATE *tmpl, CK_ATTRIBUTE_TYPE type)
         case CKK_DH:
             return dh_priv_check_exportability(type);
         case CKK_IBM_DILITHIUM:
-            return ibm_dilithium_priv_check_exportability(type);
+        case CKK_IBM_ML_DSA:
+            return ibm_ml_dsa_priv_check_exportability(type);
         case CKK_IBM_KYBER:
-            return ibm_kyber_priv_check_exportability(type);
+        case CKK_IBM_ML_KEM:
+            return ibm_ml_kem_priv_check_exportability(type);
         default:
             TRACE_ERROR("%s: %lx\n", ock_err(ERR_ATTRIBUTE_VALUE_INVALID),
                         subclass);
@@ -1725,9 +1769,17 @@ CK_RV template_validate_attribute(STDLL_TokData_t *tokdata, TEMPLATE *tmpl,
         case CKK_DH:
             return dh_publ_validate_attribute(tokdata, tmpl, attr, mode);
         case CKK_IBM_PQC_DILITHIUM:
-            return ibm_dilithium_publ_validate_attribute(tokdata, tmpl, attr, mode);
+            return ibm_ml_dsa_publ_validate_attribute(tokdata, tmpl, attr, mode,
+                                                      CKM_IBM_DILITHIUM);
         case CKK_IBM_PQC_KYBER:
-            return ibm_kyber_publ_validate_attribute(tokdata, tmpl, attr, mode);
+            return ibm_ml_kem_publ_validate_attribute(tokdata, tmpl, attr, mode,
+                                                      CKM_IBM_KYBER);
+        case CKK_IBM_ML_DSA:
+            return ibm_ml_dsa_publ_validate_attribute(tokdata, tmpl, attr, mode,
+                                                      CKM_IBM_ML_DSA);
+        case CKK_IBM_ML_KEM:
+            return ibm_ml_kem_publ_validate_attribute(tokdata, tmpl, attr, mode,
+                                                      CKM_IBM_ML_KEM);
         default:
             TRACE_ERROR("%s\n", ock_err(ERR_ATTRIBUTE_VALUE_INVALID));
             return CKR_ATTRIBUTE_VALUE_INVALID; // unknown key type
@@ -1743,9 +1795,17 @@ CK_RV template_validate_attribute(STDLL_TokData_t *tokdata, TEMPLATE *tmpl,
         case CKK_DH:
             return dh_priv_validate_attribute(tokdata, tmpl, attr, mode);
         case CKK_IBM_PQC_DILITHIUM:
-            return ibm_dilithium_priv_validate_attribute(tokdata, tmpl, attr, mode);
+            return ibm_ml_dsa_priv_validate_attribute(tokdata, tmpl, attr, mode,
+                                                      CKM_IBM_DILITHIUM);
         case CKK_IBM_PQC_KYBER:
-            return ibm_kyber_priv_validate_attribute(tokdata, tmpl, attr, mode);
+            return ibm_ml_kem_priv_validate_attribute(tokdata, tmpl, attr, mode,
+                                                      CKM_IBM_KYBER);
+        case CKK_IBM_ML_DSA:
+            return ibm_ml_dsa_priv_validate_attribute(tokdata, tmpl, attr, mode,
+                                                      CKM_IBM_ML_DSA);
+        case CKK_IBM_ML_KEM:
+            return ibm_ml_kem_priv_validate_attribute(tokdata, tmpl, attr, mode,
+                                                      CKM_IBM_ML_KEM);
         default:
             TRACE_ERROR("%s\n", ock_err(ERR_ATTRIBUTE_VALUE_INVALID));
             return CKR_ATTRIBUTE_VALUE_INVALID; // unknown key type
@@ -1753,6 +1813,17 @@ CK_RV template_validate_attribute(STDLL_TokData_t *tokdata, TEMPLATE *tmpl,
     } else if (class == CKO_SECRET_KEY) {
         switch (subclass) {
         case CKK_GENERIC_SECRET:
+        case CKK_SHA_1_HMAC:
+        case CKK_SHA224_HMAC:
+        case CKK_SHA256_HMAC:
+        case CKK_SHA384_HMAC:
+        case CKK_SHA512_HMAC:
+        case CKK_SHA3_224_HMAC:
+        case CKK_SHA3_256_HMAC:
+        case CKK_SHA3_384_HMAC:
+        case CKK_SHA3_512_HMAC:
+        case CKK_SHA512_224_HMAC:
+        case CKK_SHA512_256_HMAC:
             return generic_secret_validate_attribute(tokdata, tmpl, attr, mode);
         case CKK_DES:
             return des_validate_attribute(tokdata, tmpl, attr, mode);
