@@ -68,6 +68,9 @@ struct token_specific_struct {
     CK_RV(*t_save_token_data) (STDLL_TokData_t *tokdata,
                                CK_SLOT_ID slot_id, FILE *fh);
 
+    CK_RV (*t_get_token_info) (STDLL_TokData_t *tokdata,
+                               CK_TOKEN_INFO_PTR pInfo);
+
     // Random Number Gen
     CK_RV(*t_rng) (STDLL_TokData_t *, CK_BYTE *, CK_ULONG);
 
@@ -279,19 +282,30 @@ struct token_specific_struct {
     CK_RV(*t_dsa_verify) (STDLL_TokData_t *, CK_BYTE *, CK_BYTE *, OBJECT *);
 
     // Token Specific PQC functions
-    CK_RV (*t_ibm_dilithium_generate_keypair)(STDLL_TokData_t *,
-                                              const struct pqc_oid *,
-                                              TEMPLATE *, TEMPLATE *);
+    CK_RV (*t_ibm_ml_dsa_generate_keypair)(STDLL_TokData_t *,
+                                           CK_MECHANISM *mech,
+                                           const struct pqc_oid *,
+                                           TEMPLATE *, TEMPLATE *);
 
-    CK_RV (*t_ibm_dilithium_sign)(STDLL_TokData_t *, SESSION *, CK_BBOOL,
-                                  const struct pqc_oid *,
-                                  CK_BYTE *, CK_ULONG,
-                                  CK_BYTE *, CK_ULONG *, OBJECT *);
+    CK_RV (*t_ibm_ml_dsa_sign)(STDLL_TokData_t *, SESSION *, CK_BBOOL,
+                               const struct pqc_oid *, CK_MECHANISM *,
+                               CK_BYTE *, CK_ULONG,
+                               CK_BYTE *, CK_ULONG *, OBJECT *);
 
-    CK_RV (*t_ibm_dilithium_verify)(STDLL_TokData_t *, SESSION *,
-                                    const struct pqc_oid *,
-                                    CK_BYTE *, CK_ULONG,
-                                    CK_BYTE *, CK_ULONG, OBJECT *);
+    CK_RV (*t_ibm_ml_dsa_verify)(STDLL_TokData_t *, SESSION *,
+                                 const struct pqc_oid *, CK_MECHANISM *,
+                                 CK_BYTE *, CK_ULONG,
+                                 CK_BYTE *, CK_ULONG, OBJECT *);
+
+    CK_RV (*t_ibm_ml_kem_generate_keypair)(STDLL_TokData_t *,
+                                           CK_MECHANISM *mech,
+                                           const struct pqc_oid *,
+                                           TEMPLATE *, TEMPLATE *);
+
+    CK_RV (*t_ibm_ml_kem_derive)(STDLL_TokData_t *, SESSION *,
+                                 const struct pqc_oid *, CK_MECHANISM *,
+                                 OBJECT *, CK_OBJECT_CLASS, CK_KEY_TYPE,
+                                 OBJECT *, CK_KEY_TYPE, CK_ULONG);
 
     CK_RV(*t_get_mechanism_list) (STDLL_TokData_t *, CK_MECHANISM_TYPE_PTR,
                                   CK_ULONG_PTR);
