@@ -40,6 +40,9 @@ CK_RV token_specific_load_token_data(STDLL_TokData_t *, CK_SLOT_ID slot_id,
 CK_RV token_specific_save_token_data(STDLL_TokData_t *, CK_SLOT_ID slot_id,
                                      FILE *fh);
 
+CK_RV token_specific_get_token_info(STDLL_TokData_t *tokdata,
+                                    CK_TOKEN_INFO_PTR pInfo);
+
 CK_RV token_specific_final(STDLL_TokData_t *, CK_BBOOL);
 CK_RV token_specific_init_token(STDLL_TokData_t *, CK_SLOT_ID, CK_CHAR_PTR,
                                 CK_ULONG, CK_CHAR_PTR);
@@ -319,19 +322,30 @@ CK_RV token_specific_dsa_sign(STDLL_TokData_t *, CK_BYTE *, CK_ULONG, CK_ULONG);
 CK_RV token_specific_dsa_verify(STDLL_TokData_t *,
                                 CK_BYTE *, CK_BYTE *, OBJECT *);
 
-CK_RV token_specific_ibm_dilithium_generate_keypair(STDLL_TokData_t *,
-                                                    const struct pqc_oid *,
-                                                    TEMPLATE *, TEMPLATE *);
+CK_RV token_specific_ibm_ml_dsa_generate_keypair(STDLL_TokData_t *,
+                                                 CK_MECHANISM *mech,
+                                                 const struct pqc_oid *,
+                                                 TEMPLATE *, TEMPLATE *);
 
-CK_RV token_specific_ibm_dilithium_sign(STDLL_TokData_t *, SESSION *, CK_BBOOL,
-                                        const struct pqc_oid *,
-                                        CK_BYTE *, CK_ULONG,
-                                        CK_BYTE *, CK_ULONG *, OBJECT *);
+CK_RV token_specific_ibm_ml_dsa_sign(STDLL_TokData_t *, SESSION *, CK_BBOOL,
+                                     const struct pqc_oid *, CK_MECHANISM *,
+                                     CK_BYTE *, CK_ULONG,
+                                     CK_BYTE *, CK_ULONG *, OBJECT *);
 
-CK_RV token_specific_ibm_dilithium_verify(STDLL_TokData_t *, SESSION *,
-                                          const struct pqc_oid *,
-                                          CK_BYTE *, CK_ULONG,
-                                          CK_BYTE *, CK_ULONG, OBJECT *);
+CK_RV token_specific_ibm_ml_dsa_verify(STDLL_TokData_t *, SESSION *,
+                                       const struct pqc_oid *, CK_MECHANISM *,
+                                       CK_BYTE *, CK_ULONG,
+                                       CK_BYTE *, CK_ULONG, OBJECT *);
+
+CK_RV token_specific_ibm_ml_kem_generate_keypair(STDLL_TokData_t *,
+                                                 CK_MECHANISM *mech,
+                                                 const struct pqc_oid *,
+                                                 TEMPLATE *, TEMPLATE *);
+
+CK_RV token_specific_ibm_ml_kem_derive(STDLL_TokData_t *, SESSION *,
+                                       const struct pqc_oid *, CK_MECHANISM *,
+                                       OBJECT *, CK_OBJECT_CLASS, CK_KEY_TYPE,
+                                       OBJECT *, CK_KEY_TYPE, CK_ULONG);
 
 CK_RV token_specific_get_mechanism_list(STDLL_TokData_t *,
                                         CK_MECHANISM_TYPE_PTR pMechanismList,
