@@ -40,7 +40,7 @@ CK_RV verify_mgr_init(STDLL_TokData_t *tokdata,
     OBJECT *key_obj = NULL;
     CK_ATTRIBUTE *attr = NULL;
     CK_BYTE *ptr = NULL;
-    CK_KEY_TYPE keytype;
+    CK_KEY_TYPE keytype, exp_keytype, alt_keytype;
     CK_OBJECT_CLASS class;
     CK_BBOOL flag;
     CK_RV rc;
@@ -173,6 +173,10 @@ CK_RV verify_mgr_init(STDLL_TokData_t *tokdata,
     case CKM_ECDSA_SHA256:
     case CKM_ECDSA_SHA384:
     case CKM_ECDSA_SHA512:
+    case CKM_ECDSA_SHA3_224:
+    case CKM_ECDSA_SHA3_256:
+    case CKM_ECDSA_SHA3_384:
+    case CKM_ECDSA_SHA3_512:
         if (mech->ulParameterLen != 0) {
             TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
             rc = CKR_MECHANISM_PARAM_INVALID;
@@ -229,6 +233,10 @@ CK_RV verify_mgr_init(STDLL_TokData_t *tokdata,
     case CKM_SHA256_RSA_PKCS:
     case CKM_SHA384_RSA_PKCS:
     case CKM_SHA512_RSA_PKCS:
+    case CKM_SHA3_224_RSA_PKCS:
+    case CKM_SHA3_256_RSA_PKCS:
+    case CKM_SHA3_384_RSA_PKCS:
+    case CKM_SHA3_512_RSA_PKCS:
         if (mech->ulParameterLen != 0) {
             TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
             rc = CKR_MECHANISM_PARAM_INVALID;
@@ -275,6 +283,10 @@ CK_RV verify_mgr_init(STDLL_TokData_t *tokdata,
     case CKM_SHA256_RSA_PKCS_PSS:
     case CKM_SHA384_RSA_PKCS_PSS:
     case CKM_SHA512_RSA_PKCS_PSS:
+    case CKM_SHA3_224_RSA_PKCS_PSS:
+    case CKM_SHA3_256_RSA_PKCS_PSS:
+    case CKM_SHA3_384_RSA_PKCS_PSS:
+    case CKM_SHA3_512_RSA_PKCS_PSS:
         rc = template_attribute_get_non_empty(key_obj->template, CKA_MODULUS,
                                               &attr);
         if (rc != CKR_OK) {
@@ -400,6 +412,10 @@ CK_RV verify_mgr_init(STDLL_TokData_t *tokdata,
     case CKM_SHA512_HMAC:
     case CKM_SHA512_224_HMAC:
     case CKM_SHA512_256_HMAC:
+    case CKM_SHA3_224_HMAC:
+    case CKM_SHA3_256_HMAC:
+    case CKM_SHA3_384_HMAC:
+    case CKM_SHA3_512_HMAC:
     case CKM_IBM_SHA3_224_HMAC:
     case CKM_IBM_SHA3_256_HMAC:
     case CKM_IBM_SHA3_384_HMAC:
@@ -417,7 +433,14 @@ CK_RV verify_mgr_init(STDLL_TokData_t *tokdata,
             goto done;
         }
 
-        if (keytype != CKK_GENERIC_SECRET) {
+        rc = pkcsget_keytype_for_mech(mech->mechanism, &exp_keytype,
+                                      &alt_keytype);
+        if (rc != CKR_OK) {
+            TRACE_ERROR("pkcsget_keytype_for_mech failed.\n");
+            goto done;
+        }
+
+        if (keytype != exp_keytype && keytype != alt_keytype) {
             TRACE_ERROR("%s\n", ock_err(ERR_KEY_TYPE_INCONSISTENT));
             rc = CKR_KEY_TYPE_INCONSISTENT;
             goto done;
@@ -490,9 +513,16 @@ CK_RV verify_mgr_init(STDLL_TokData_t *tokdata,
     case CKM_SHA512_HMAC_GENERAL:
     case CKM_SHA512_224_HMAC_GENERAL:
     case CKM_SHA512_256_HMAC_GENERAL:
+    case CKM_SHA3_224_HMAC_GENERAL:
+    case CKM_SHA3_256_HMAC_GENERAL:
+    case CKM_SHA3_384_HMAC_GENERAL:
+    case CKM_SHA3_512_HMAC_GENERAL:
         {
             CK_MAC_GENERAL_PARAMS *param =
                 (CK_MAC_GENERAL_PARAMS *) mech->pParameter;
+            CK_MECHANISM_TYPE digest_mech;
+            CK_BBOOL general;
+            CK_ULONG hsize;
 
             if (mech->ulParameterLen != sizeof(CK_MAC_GENERAL_PARAMS) ||
                 mech->pParameter == NULL) {
@@ -500,51 +530,20 @@ CK_RV verify_mgr_init(STDLL_TokData_t *tokdata,
                 rc= CKR_MECHANISM_PARAM_INVALID;
                 goto done;
             }
-#if !(NOMD2)
-            if ((mech->mechanism == CKM_MD2_HMAC_GENERAL) && (*param > 16)) {
-                TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
-                rc = CKR_MECHANISM_PARAM_INVALID;
+
+            rc = get_hmac_digest(mech->mechanism, &digest_mech, &general);
+            if (rc != CKR_OK) {
+                TRACE_ERROR("%s get_hmac_digest failed\n", __func__);
                 goto done;
             }
-#endif
-            if ((mech->mechanism == CKM_MD5_HMAC_GENERAL) && (*param > 16)) {
-                TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
-                rc = CKR_MECHANISM_PARAM_INVALID;
+
+            rc = get_sha_size(digest_mech, &hsize);
+            if (rc != CKR_OK) {
+                TRACE_ERROR("%s get_sha_size failed\n", __func__);
                 goto done;
             }
-            if ((mech->mechanism == CKM_SHA_1_HMAC_GENERAL) && (*param > 20)) {
-                TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
-                rc = CKR_MECHANISM_PARAM_INVALID;
-                goto done;
-            }
-            if ((mech->mechanism == CKM_SHA224_HMAC_GENERAL) && (*param > 28)) {
-                TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
-                rc = CKR_MECHANISM_PARAM_INVALID;
-                goto done;
-            }
-            if ((mech->mechanism == CKM_SHA256_HMAC_GENERAL) && (*param > 32)) {
-                TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
-                rc = CKR_MECHANISM_PARAM_INVALID;
-                goto done;
-            }
-            if ((mech->mechanism == CKM_SHA384_HMAC_GENERAL) && (*param > 48)) {
-                TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
-                rc = CKR_MECHANISM_PARAM_INVALID;
-                goto done;
-            }
-            if ((mech->mechanism == CKM_SHA512_HMAC_GENERAL) && (*param > 64)) {
-                TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
-                rc = CKR_MECHANISM_PARAM_INVALID;
-                goto done;
-            }
-            if ((mech->mechanism == CKM_SHA512_224_HMAC_GENERAL)
-                && (*param > 28)) {
-                TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
-                rc = CKR_MECHANISM_PARAM_INVALID;
-                goto done;
-            }
-            if ((mech->mechanism == CKM_SHA512_256_HMAC_GENERAL)
-                && (*param > 32)) {
+
+            if (*param > hsize) {
                 TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
                 rc = CKR_MECHANISM_PARAM_INVALID;
                 goto done;
@@ -557,7 +556,14 @@ CK_RV verify_mgr_init(STDLL_TokData_t *tokdata,
                 goto done;
             }
 
-            if (keytype != CKK_GENERIC_SECRET) {
+            rc = pkcsget_keytype_for_mech(mech->mechanism, &exp_keytype,
+                                          &alt_keytype);
+            if (rc != CKR_OK) {
+                TRACE_ERROR("pkcsget_keytype_for_mech failed.\n");
+                goto done;
+            }
+
+            if (keytype != exp_keytype && keytype != alt_keytype) {
                 TRACE_ERROR("%s\n", ock_err(ERR_KEY_TYPE_INCONSISTENT));
                 rc = CKR_KEY_TYPE_INCONSISTENT;
                 goto done;
@@ -781,6 +787,85 @@ CK_RV verify_mgr_init(STDLL_TokData_t *tokdata,
         }
         memset(ctx->context, 0x0, sizeof(AES_CMAC_CONTEXT));
         break;
+    case CKM_IBM_DILITHIUM:
+        if (mech->ulParameterLen != 0) {
+            TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
+            rc = CKR_MECHANISM_PARAM_INVALID;
+            goto done;
+        }
+
+        rc = template_attribute_get_ulong(key_obj->template, CKA_KEY_TYPE,
+                                          &keytype);
+        if (rc != CKR_OK) {
+            TRACE_ERROR("Could not find CKA_KEY_TYPE for the key.\n");
+            goto done;
+        }
+
+        if (keytype != CKK_IBM_PQC_DILITHIUM) {
+            TRACE_ERROR("%s\n", ock_err(ERR_KEY_TYPE_INCONSISTENT));
+            rc = CKR_KEY_TYPE_INCONSISTENT;
+            goto done;
+        }
+
+        rc = template_attribute_get_ulong(key_obj->template, CKA_CLASS,
+                                          &class);
+        if (rc != CKR_OK) {
+            TRACE_ERROR("Could not find CKA_CLASS for the key.\n");
+            goto done;
+        }
+
+        if (class != CKO_PUBLIC_KEY) {
+            TRACE_ERROR("This operation requires a public key.\n");
+            rc = CKR_KEY_FUNCTION_NOT_PERMITTED;
+            goto done;
+        }
+
+        ctx->context_len = 0;
+        ctx->context = NULL;
+        break;
+    case CKM_IBM_ML_DSA:
+        if (mech->ulParameterLen != 0 &&
+            mech->ulParameterLen != sizeof(CK_IBM_SIGN_ADDITIONAL_CONTEXT)) {
+            TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_PARAM_INVALID));
+            rc = CKR_MECHANISM_PARAM_INVALID;
+            goto done;
+        }
+        if (mech->ulParameterLen == sizeof(CK_IBM_SIGN_ADDITIONAL_CONTEXT) &&
+            mech->pParameter == NULL) {
+            rc = CKR_MECHANISM_PARAM_INVALID;
+            goto done;
+        }
+
+        rc = template_attribute_get_ulong(key_obj->template, CKA_KEY_TYPE,
+                                          &keytype);
+        if (rc != CKR_OK) {
+            TRACE_ERROR("Could not find CKA_KEY_TYPE for the key.\n");
+            goto done;
+        }
+
+        if (keytype != CKK_IBM_ML_DSA) {
+            TRACE_ERROR("%s\n", ock_err(ERR_KEY_TYPE_INCONSISTENT));
+            rc = CKR_KEY_TYPE_INCONSISTENT;
+            goto done;
+        }
+
+        rc = template_attribute_get_ulong(key_obj->template, CKA_CLASS,
+                                          &class);
+        if (rc != CKR_OK) {
+            TRACE_ERROR("Could not find CKA_CLASS for the key.\n");
+            goto done;
+        }
+
+        if (class != CKO_PUBLIC_KEY) {
+            TRACE_ERROR("This operation requires a public key.\n");
+            rc = CKR_KEY_FUNCTION_NOT_PERMITTED;
+            goto done;
+        }
+
+        ctx->context_len = 0;
+        ctx->context = NULL;
+        break;
+
     default:
         TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_INVALID));
         rc = CKR_MECHANISM_INVALID;
@@ -796,6 +881,21 @@ CK_RV verify_mgr_init(STDLL_TokData_t *tokdata,
             goto done;
         }
         memcpy(ptr, mech->pParameter, mech->ulParameterLen);
+
+        /* Deep copy mechanism parameter, if required */
+        switch (mech->mechanism)
+        {
+        case CKM_IBM_ML_DSA:
+            rc = ibm_ml_dsa_dup_param(mech->pParameter, ptr,
+                                      mech->ulParameterLen);
+            if (rc != CKR_OK) {
+                TRACE_ERROR("ibm_ml_dsa_dup_param failed\n");
+                free(ptr);
+            }
+            break;
+        default:
+            break;
+        }
     }
 
     ctx->key = key;
@@ -831,22 +931,32 @@ CK_RV verify_mgr_cleanup(STDLL_TokData_t *tokdata, SESSION *sess,
         return CKR_FUNCTION_FAILED;
     }
     ctx->key = 0;
-    ctx->mech.ulParameterLen = 0;
-    ctx->mech.mechanism = 0;
     ctx->multi_init = FALSE;
     ctx->multi = FALSE;
     ctx->active = FALSE;
     ctx->init_pending = FALSE;
     ctx->recover = FALSE;
-    ctx->context_len = 0;
     ctx->pkey_active = FALSE;
     ctx->state_unsaveable = FALSE;
     ctx->count_statistics = FALSE;
 
     if (ctx->mech.pParameter) {
+        /* Deep free mechanism parameter, if required */
+        switch (ctx->mech.mechanism)
+        {
+        case CKM_IBM_ML_DSA:
+            ibm_ml_dsa_free_param(ctx->mech.pParameter,
+                                  ctx->mech.ulParameterLen);
+            break;
+        default:
+            break;
+        }
+
         free(ctx->mech.pParameter);
         ctx->mech.pParameter = NULL;
     }
+    ctx->mech.ulParameterLen = 0;
+    ctx->mech.mechanism = 0;
 
     if (ctx->context) {
         if (ctx->context_free_func != NULL)
@@ -856,6 +966,7 @@ CK_RV verify_mgr_cleanup(STDLL_TokData_t *tokdata, SESSION *sess,
             free(ctx->context);
         ctx->context = NULL;
     }
+    ctx->context_len = 0;
     ctx->context_free_func = NULL;
 
     return CKR_OK;
@@ -919,6 +1030,10 @@ CK_RV verify_mgr_verify(STDLL_TokData_t *tokdata,
     case CKM_SHA256_RSA_PKCS:
     case CKM_SHA384_RSA_PKCS:
     case CKM_SHA512_RSA_PKCS:
+    case CKM_SHA3_224_RSA_PKCS:
+    case CKM_SHA3_256_RSA_PKCS:
+    case CKM_SHA3_384_RSA_PKCS:
+    case CKM_SHA3_512_RSA_PKCS:
         return rsa_hash_pkcs_verify(tokdata, sess, ctx,
                                     in_data, in_data_len, signature, sig_len);
     case CKM_SHA1_RSA_PKCS_PSS:
@@ -926,6 +1041,10 @@ CK_RV verify_mgr_verify(STDLL_TokData_t *tokdata,
     case CKM_SHA256_RSA_PKCS_PSS:
     case CKM_SHA384_RSA_PKCS_PSS:
     case CKM_SHA512_RSA_PKCS_PSS:
+    case CKM_SHA3_224_RSA_PKCS_PSS:
+    case CKM_SHA3_256_RSA_PKCS_PSS:
+    case CKM_SHA3_384_RSA_PKCS_PSS:
+    case CKM_SHA3_512_RSA_PKCS_PSS:
         return rsa_hash_pss_verify(tokdata, sess, ctx, in_data, in_data_len,
                                    signature, sig_len);
 #if !(NODSA)
@@ -957,6 +1076,14 @@ CK_RV verify_mgr_verify(STDLL_TokData_t *tokdata,
     case CKM_SHA512_224_HMAC_GENERAL:
     case CKM_SHA512_256_HMAC:
     case CKM_SHA512_256_HMAC_GENERAL:
+    case CKM_SHA3_224_HMAC:
+    case CKM_SHA3_224_HMAC_GENERAL:
+    case CKM_SHA3_256_HMAC:
+    case CKM_SHA3_256_HMAC_GENERAL:
+    case CKM_SHA3_384_HMAC:
+    case CKM_SHA3_384_HMAC_GENERAL:
+    case CKM_SHA3_512_HMAC:
+    case CKM_SHA3_512_HMAC_GENERAL:
     case CKM_IBM_SHA3_224_HMAC:
     case CKM_IBM_SHA3_256_HMAC:
     case CKM_IBM_SHA3_384_HMAC:
@@ -988,11 +1115,19 @@ CK_RV verify_mgr_verify(STDLL_TokData_t *tokdata,
     case CKM_ECDSA_SHA256:
     case CKM_ECDSA_SHA384:
     case CKM_ECDSA_SHA512:
+    case CKM_ECDSA_SHA3_224:
+    case CKM_ECDSA_SHA3_256:
+    case CKM_ECDSA_SHA3_384:
+    case CKM_ECDSA_SHA3_512:
         return ec_hash_verify(tokdata, sess, ctx,
                               in_data, in_data_len, signature, sig_len);
     case CKM_ECDSA:
         return ec_verify(tokdata, sess, ctx,
                          in_data, in_data_len, signature, sig_len);
+    case CKM_IBM_DILITHIUM:
+    case CKM_IBM_ML_DSA:
+        return ibm_ml_dsa_verify(tokdata, sess, ctx,
+                                 in_data, in_data_len, signature, sig_len);
     default:
         TRACE_ERROR("%s\n", ock_err(ERR_MECHANISM_INVALID));
         return CKR_MECHANISM_INVALID;
@@ -1042,6 +1177,10 @@ CK_RV verify_mgr_verify_update(STDLL_TokData_t *tokdata,
     case CKM_SHA256_RSA_PKCS:
     case CKM_SHA384_RSA_PKCS:
     case CKM_SHA512_RSA_PKCS:
+    case CKM_SHA3_224_RSA_PKCS:
+    case CKM_SHA3_256_RSA_PKCS:
+    case CKM_SHA3_384_RSA_PKCS:
+    case CKM_SHA3_512_RSA_PKCS:
         return rsa_hash_pkcs_verify_update(tokdata, sess, ctx, in_data,
                                            in_data_len);
     case CKM_SHA1_RSA_PKCS_PSS:
@@ -1049,6 +1188,10 @@ CK_RV verify_mgr_verify_update(STDLL_TokData_t *tokdata,
     case CKM_SHA256_RSA_PKCS_PSS:
     case CKM_SHA384_RSA_PKCS_PSS:
     case CKM_SHA512_RSA_PKCS_PSS:
+    case CKM_SHA3_224_RSA_PKCS_PSS:
+    case CKM_SHA3_256_RSA_PKCS_PSS:
+    case CKM_SHA3_384_RSA_PKCS_PSS:
+    case CKM_SHA3_512_RSA_PKCS_PSS:
         return rsa_hash_pss_update(tokdata, sess, ctx, in_data, in_data_len);
     case CKM_SSL3_MD5_MAC:
     case CKM_SSL3_SHA1_MAC:
@@ -1071,6 +1214,10 @@ CK_RV verify_mgr_verify_update(STDLL_TokData_t *tokdata,
     case CKM_ECDSA_SHA256:
     case CKM_ECDSA_SHA384:
     case CKM_ECDSA_SHA512:
+    case CKM_ECDSA_SHA3_224:
+    case CKM_ECDSA_SHA3_256:
+    case CKM_ECDSA_SHA3_384:
+    case CKM_ECDSA_SHA3_512:
         return ec_hash_verify_update(tokdata, sess, ctx, in_data, in_data_len);
     case CKM_SHA_1_HMAC:
     case CKM_SHA224_HMAC:
@@ -1086,6 +1233,14 @@ CK_RV verify_mgr_verify_update(STDLL_TokData_t *tokdata,
     case CKM_SHA512_HMAC_GENERAL:
     case CKM_SHA512_224_HMAC_GENERAL:
     case CKM_SHA512_256_HMAC_GENERAL:
+    case CKM_SHA3_224_HMAC:
+    case CKM_SHA3_224_HMAC_GENERAL:
+    case CKM_SHA3_256_HMAC:
+    case CKM_SHA3_256_HMAC_GENERAL:
+    case CKM_SHA3_384_HMAC:
+    case CKM_SHA3_384_HMAC_GENERAL:
+    case CKM_SHA3_512_HMAC:
+    case CKM_SHA3_512_HMAC_GENERAL:
     case CKM_IBM_SHA3_224_HMAC:
     case CKM_IBM_SHA3_256_HMAC:
     case CKM_IBM_SHA3_384_HMAC:
@@ -1135,6 +1290,10 @@ CK_RV verify_mgr_verify_final(STDLL_TokData_t *tokdata,
     case CKM_SHA256_RSA_PKCS:
     case CKM_SHA384_RSA_PKCS:
     case CKM_SHA512_RSA_PKCS:
+    case CKM_SHA3_224_RSA_PKCS:
+    case CKM_SHA3_256_RSA_PKCS:
+    case CKM_SHA3_384_RSA_PKCS:
+    case CKM_SHA3_512_RSA_PKCS:
         return rsa_hash_pkcs_verify_final(tokdata, sess, ctx, signature,
                                           sig_len);
     case CKM_SHA1_RSA_PKCS_PSS:
@@ -1142,6 +1301,10 @@ CK_RV verify_mgr_verify_final(STDLL_TokData_t *tokdata,
     case CKM_SHA256_RSA_PKCS_PSS:
     case CKM_SHA384_RSA_PKCS_PSS:
     case CKM_SHA512_RSA_PKCS_PSS:
+    case CKM_SHA3_224_RSA_PKCS_PSS:
+    case CKM_SHA3_256_RSA_PKCS_PSS:
+    case CKM_SHA3_384_RSA_PKCS_PSS:
+    case CKM_SHA3_512_RSA_PKCS_PSS:
         return rsa_hash_pss_verify_final(tokdata, sess, ctx, signature,
                                          sig_len);
     case CKM_SSL3_MD5_MAC:
@@ -1164,6 +1327,10 @@ CK_RV verify_mgr_verify_final(STDLL_TokData_t *tokdata,
     case CKM_ECDSA_SHA256:
     case CKM_ECDSA_SHA384:
     case CKM_ECDSA_SHA512:
+    case CKM_ECDSA_SHA3_224:
+    case CKM_ECDSA_SHA3_256:
+    case CKM_ECDSA_SHA3_384:
+    case CKM_ECDSA_SHA3_512:
         return ec_hash_verify_final(tokdata, sess, ctx, signature, sig_len);
     case CKM_SHA_1_HMAC:
     case CKM_SHA224_HMAC:
@@ -1179,6 +1346,14 @@ CK_RV verify_mgr_verify_final(STDLL_TokData_t *tokdata,
     case CKM_SHA512_HMAC_GENERAL:
     case CKM_SHA512_224_HMAC_GENERAL:
     case CKM_SHA512_256_HMAC_GENERAL:
+    case CKM_SHA3_224_HMAC:
+    case CKM_SHA3_224_HMAC_GENERAL:
+    case CKM_SHA3_256_HMAC:
+    case CKM_SHA3_256_HMAC_GENERAL:
+    case CKM_SHA3_384_HMAC:
+    case CKM_SHA3_384_HMAC_GENERAL:
+    case CKM_SHA3_512_HMAC:
+    case CKM_SHA3_512_HMAC_GENERAL:
     case CKM_IBM_SHA3_224_HMAC:
     case CKM_IBM_SHA3_256_HMAC:
     case CKM_IBM_SHA3_384_HMAC:
