@@ -42,6 +42,7 @@
 
 #include <openssl/rand.h>
 #include <openssl/dh.h>
+#include <openssl/rsa.h>
 #include <openssl/evp.h>
 #include <openssl/crypto.h>
 #include <openssl/bn.h>
@@ -60,48 +61,74 @@ const char descr[] = "IBM Soft token";
 const char label[] = "softtok";
 
 static const MECH_LIST_ELEMENT soft_mech_list[] = {
-    {CKM_RSA_PKCS_KEY_PAIR_GEN, {512, 4096, CKF_GENERATE_KEY_PAIR}},
+    {CKM_RSA_PKCS_KEY_PAIR_GEN,
+            {512, OPENSSL_RSA_MAX_MODULUS_BITS, CKF_GENERATE_KEY_PAIR}},
 #if !(NODSA)
     {CKM_DSA_KEY_PAIR_GEN, {512, 1024, CKF_GENERATE_KEY_PAIR}},
 #endif
     {CKM_DES_KEY_GEN, {8, 8, CKF_GENERATE}},
     {CKM_DES3_KEY_GEN, {24, 24, CKF_GENERATE}},
     {CKM_RSA_PKCS,
-     {512, 4096, CKF_ENCRYPT | CKF_DECRYPT | CKF_WRAP | CKF_UNWRAP | CKF_SIGN |
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS,
+      CKF_ENCRYPT | CKF_DECRYPT | CKF_WRAP | CKF_UNWRAP | CKF_SIGN |
       CKF_VERIFY | CKF_SIGN_RECOVER | CKF_VERIFY_RECOVER}},
-    {CKM_SHA1_RSA_PKCS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
-    {CKM_SHA224_RSA_PKCS, {512, 4096, CKF_SIGN|CKF_VERIFY}},
-    {CKM_SHA256_RSA_PKCS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
-    {CKM_SHA384_RSA_PKCS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
-    {CKM_SHA512_RSA_PKCS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
-    {CKM_SHA3_224_RSA_PKCS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
-    {CKM_SHA3_256_RSA_PKCS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
-    {CKM_SHA3_384_RSA_PKCS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
-    {CKM_SHA3_512_RSA_PKCS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
-    {CKM_RSA_PKCS_PSS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
-    {CKM_SHA1_RSA_PKCS_PSS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
-    {CKM_SHA224_RSA_PKCS_PSS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
-    {CKM_SHA256_RSA_PKCS_PSS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
-    {CKM_SHA384_RSA_PKCS_PSS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
-    {CKM_SHA512_RSA_PKCS_PSS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
-    {CKM_SHA3_224_RSA_PKCS_PSS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
-    {CKM_SHA3_256_RSA_PKCS_PSS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
-    {CKM_SHA3_384_RSA_PKCS_PSS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
-    {CKM_SHA3_512_RSA_PKCS_PSS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA1_RSA_PKCS,
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA224_RSA_PKCS,
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS, CKF_SIGN|CKF_VERIFY}},
+    {CKM_SHA256_RSA_PKCS,
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA384_RSA_PKCS,
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA512_RSA_PKCS,
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_224_RSA_PKCS,
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_256_RSA_PKCS,
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_384_RSA_PKCS,
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_512_RSA_PKCS,
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS, CKF_SIGN | CKF_VERIFY}},
+    {CKM_RSA_PKCS_PSS,
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA1_RSA_PKCS_PSS,
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA224_RSA_PKCS_PSS,
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA256_RSA_PKCS_PSS,
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA384_RSA_PKCS_PSS,
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA512_RSA_PKCS_PSS,
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_224_RSA_PKCS_PSS,
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_256_RSA_PKCS_PSS,
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_384_RSA_PKCS_PSS,
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA3_512_RSA_PKCS_PSS,
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS, CKF_SIGN | CKF_VERIFY}},
 #if !(NOX509)
     {CKM_RSA_X_509,
-     {512, 4096, CKF_ENCRYPT | CKF_DECRYPT | CKF_WRAP | CKF_UNWRAP | CKF_SIGN |
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS,
+      CKF_ENCRYPT | CKF_DECRYPT | CKF_WRAP | CKF_UNWRAP | CKF_SIGN |
       CKF_VERIFY | CKF_SIGN_RECOVER | CKF_VERIFY_RECOVER}},
 #endif
     {CKM_RSA_PKCS_OAEP,
-     {512, 4096, CKF_ENCRYPT | CKF_DECRYPT | CKF_WRAP | CKF_UNWRAP}},
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS,
+      CKF_ENCRYPT | CKF_DECRYPT | CKF_WRAP | CKF_UNWRAP}},
 #if !(NOMD2)
-    {CKM_MD2_RSA_PKCS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
+    {CKM_MD2_RSA_PKCS,
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS, CKF_SIGN | CKF_VERIFY}},
 #endif
 #if !(NOMD5)
-    {CKM_MD5_RSA_PKCS, {512, 4096, CKF_SIGN | CKF_VERIFY}},
+    {CKM_MD5_RSA_PKCS,
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS, CKF_SIGN | CKF_VERIFY}},
 #endif
-    {CKM_RSA_AES_KEY_WRAP, {512, 4096, CKF_WRAP | CKF_UNWRAP}},
+    {CKM_RSA_AES_KEY_WRAP,
+     {512, OPENSSL_RSA_MAX_MODULUS_BITS, CKF_WRAP | CKF_UNWRAP}},
 #if !(NODSA)
     {CKM_DSA, {512, 1024, CKF_SIGN | CKF_VERIFY}},
 #endif
@@ -152,11 +179,13 @@ static const MECH_LIST_ELEMENT soft_mech_list[] = {
     {CKM_SHA512_224, {0, 0, CKF_DIGEST}},
     {CKM_SHA512_224_HMAC, {112, 2048, CKF_SIGN | CKF_VERIFY}},
     {CKM_SHA512_224_HMAC_GENERAL, {112, 2048, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA512_224_KEY_DERIVATION, {8, 224, CKF_DERIVE}},
 #endif
 #ifdef NID_sha512_256WithRSAEncryption
     {CKM_SHA512_256, {0, 0, CKF_DIGEST}},
     {CKM_SHA512_256_HMAC, {128, 2048, CKF_SIGN | CKF_VERIFY}},
     {CKM_SHA512_256_HMAC_GENERAL, {128, 2048, CKF_SIGN | CKF_VERIFY}},
+    {CKM_SHA512_256_KEY_DERIVATION, {8, 256, CKF_DERIVE}},
 #endif
 #ifdef NID_sha3_224
     {CKM_SHA3_224, {0, 0, CKF_DIGEST}},
@@ -239,36 +268,55 @@ static const MECH_LIST_ELEMENT soft_mech_list[] = {
     {CKM_AES_KEY_WRAP_PKCS7,
      {32, 64, CKF_ENCRYPT | CKF_DECRYPT | CKF_WRAP | CKF_UNWRAP}},
     {CKM_GENERIC_SECRET_KEY_GEN, {80, 2048, CKF_GENERATE}},
+    {CKM_SHA_1_KEY_GEN, {80, 2048, CKF_GENERATE}},
+    {CKM_SHA224_KEY_GEN, {80, 2048, CKF_GENERATE}},
+    {CKM_SHA256_KEY_GEN, {80, 2048, CKF_GENERATE}},
+    {CKM_SHA384_KEY_GEN, {80, 2048, CKF_GENERATE}},
+    {CKM_SHA512_KEY_GEN, {80, 2048, CKF_GENERATE}},
+    {CKM_SHA512_224_KEY_GEN, {80, 2048, CKF_GENERATE}},
+    {CKM_SHA512_256_KEY_GEN, {80, 2048, CKF_GENERATE}},
+    {CKM_SHA3_224_KEY_GEN, {80, 2048, CKF_GENERATE}},
+    {CKM_SHA3_256_KEY_GEN, {80, 2048, CKF_GENERATE}},
+    {CKM_SHA3_384_KEY_GEN, {80, 2048, CKF_GENERATE}},
+    {CKM_SHA3_512_KEY_GEN, {80, 2048, CKF_GENERATE}},
 #if !(NO_EC)
     {CKM_EC_KEY_PAIR_GEN, {160, 521, CKF_GENERATE_KEY_PAIR |
-                           CKF_EC_NAMEDCURVE | CKF_EC_F_P}},
-    {CKM_ECDSA, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_NAMEDCURVE |
-                 CKF_EC_F_P}},
-    {CKM_ECDSA_SHA1, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_NAMEDCURVE |
-                      CKF_EC_F_P}},
-    {CKM_ECDSA_SHA224, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_NAMEDCURVE |
-                        CKF_EC_F_P}},
-    {CKM_ECDSA_SHA256, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_NAMEDCURVE |
-                        CKF_EC_F_P}},
-    {CKM_ECDSA_SHA384, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_NAMEDCURVE |
-                        CKF_EC_F_P}},
-    {CKM_ECDSA_SHA512, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_NAMEDCURVE |
-                        CKF_EC_F_P}},
-    {CKM_ECDSA_SHA3_224, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_NAMEDCURVE |
-                          CKF_EC_F_P}},
-    {CKM_ECDSA_SHA3_256, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_NAMEDCURVE |
-                          CKF_EC_F_P}},
-    {CKM_ECDSA_SHA3_384, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_NAMEDCURVE |
-                          CKF_EC_F_P}},
-    {CKM_ECDSA_SHA3_512, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_NAMEDCURVE |
-                          CKF_EC_F_P}},
-    {CKM_ECDH1_DERIVE, {160, 521, CKF_DERIVE | CKF_EC_NAMEDCURVE | CKF_EC_F_P}},
+                           CKF_EC_OID | CKF_EC_F_P | CKF_EC_UNCOMPRESS |
+                           CKF_EC_COMPRESS}},
+    {CKM_ECDSA, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_OID |
+                 CKF_EC_F_P | CKF_EC_UNCOMPRESS | CKF_EC_COMPRESS}},
+    {CKM_ECDSA_SHA1, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_OID |
+                      CKF_EC_F_P | CKF_EC_UNCOMPRESS | CKF_EC_COMPRESS}},
+    {CKM_ECDSA_SHA224, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_OID |
+                        CKF_EC_F_P | CKF_EC_UNCOMPRESS | CKF_EC_COMPRESS}},
+    {CKM_ECDSA_SHA256, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_OID |
+                        CKF_EC_F_P | CKF_EC_UNCOMPRESS | CKF_EC_COMPRESS}},
+    {CKM_ECDSA_SHA384, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_OID |
+                        CKF_EC_F_P | CKF_EC_UNCOMPRESS | CKF_EC_COMPRESS}},
+    {CKM_ECDSA_SHA512, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_OID |
+                        CKF_EC_F_P | CKF_EC_UNCOMPRESS | CKF_EC_COMPRESS}},
+    {CKM_ECDSA_SHA3_224, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_OID |
+                          CKF_EC_F_P | CKF_EC_UNCOMPRESS | CKF_EC_COMPRESS}},
+    {CKM_ECDSA_SHA3_256, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_OID |
+                          CKF_EC_F_P | CKF_EC_UNCOMPRESS | CKF_EC_COMPRESS}},
+    {CKM_ECDSA_SHA3_384, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_OID |
+                          CKF_EC_F_P | CKF_EC_UNCOMPRESS | CKF_EC_COMPRESS}},
+    {CKM_ECDSA_SHA3_512, {160, 521, CKF_SIGN | CKF_VERIFY | CKF_EC_OID |
+                          CKF_EC_F_P | CKF_EC_UNCOMPRESS | CKF_EC_COMPRESS}},
+    {CKM_ECDH1_DERIVE, {160, 521, CKF_DERIVE | CKF_EC_OID | CKF_EC_F_P |
+                        CKF_EC_UNCOMPRESS | CKF_EC_COMPRESS}},
     {CKM_ECDH_AES_KEY_WRAP, {160, 521, CKF_WRAP | CKF_UNWRAP |
-                             CKF_EC_NAMEDCURVE | CKF_EC_F_P}},
+                             CKF_EC_OID | CKF_EC_F_P | CKF_EC_UNCOMPRESS |
+                             CKF_EC_COMPRESS}},
 #endif
 #if OPENSSL_VERSION_PREREQ(3, 0)
     {CKM_IBM_DILITHIUM, {256, 256, CKF_GENERATE_KEY_PAIR |
                                    CKF_SIGN | CKF_VERIFY}},
+    {CKM_IBM_ML_DSA_KEY_PAIR_GEN, {1312, 2592, CKF_GENERATE_KEY_PAIR}},
+    {CKM_IBM_ML_DSA, {1312, 2592, CKF_SIGN | CKF_VERIFY}},
+    {CKM_IBM_ML_KEM_KEY_PAIR_GEN, {800, 1568, CKF_GENERATE_KEY_PAIR}},
+    {CKM_IBM_ML_KEM, {800, 1568, CKF_DERIVE}},
+    {CKM_IBM_ML_KEM_WITH_ECDH, {800, 1568, CKF_DERIVE}},
 #endif
 };
 
@@ -278,6 +326,9 @@ static const CK_ULONG soft_mech_list_len =
 struct soft_private_data {
 #if OPENSSL_VERSION_PREREQ(3, 0)
     OSSL_PROVIDER *oqs_provider;
+    CK_BBOOL supports_dilithium;
+    CK_BBOOL supports_ml_dsa;
+    CK_BBOOL supports_ml_kem;
 #else
     void *dummy;
 #endif
@@ -317,27 +368,28 @@ CK_RV token_specific_init(STDLL_TokData_t *tokdata, CK_SLOT_ID SlotNumber,
      * Try to load the 'oqsprovider'. This optional provider must be installed
      * and configured separately, it does not come with OpenSSL 3.x by default.
      * If loading the 'oqsprovider' fails, this is not an error, it just means
-     * that the soft token does not support any quantum safe mechanisms.
-     * Also check if the 'oqsprovider' does support the Dilithium R3_44 variant.
-     * If not don't consider it loaded, because it can't support any of the
-     * Dilithium variants supported by the Soft token.
+     * that the soft token may not support any quantum safe mechanisms, if not
+     * also OpenSSL >= 3.5 is use, which has built-in support for ML-DSA.
      */
     soft_private->oqs_provider = OSSL_PROVIDER_load(NULL, "oqsprovider");
     if (soft_private->oqs_provider == NULL) {
         TRACE_DEVEL("OSSL_PROVIDER_load for 'oqsprovider' failed, no quantum "
                     "safe mechanisms are supported.\n");
         ERR_pop_to_mark();
-    } else {
-        oid = find_pqc_by_keyform(dilithium_oids,
-                                  CK_IBM_DILITHIUM_KEYFORM_ROUND3_44);
-        if (oid == NULL || openssl_get_pqc_oid_name(oid) == NULL) {
-            OSSL_PROVIDER_unload(soft_private->oqs_provider);
-            soft_private->oqs_provider = NULL;
-            TRACE_DEVEL("The 'oqsprovider' does not support Dilithium R3_44, "
-                        "no quantum safe mechanisms are supported.\n");
-            ERR_pop_to_mark();
-        }
-    }
+    };
+
+    oid = find_pqc_by_keyform(dilithium_oids,
+                              CK_IBM_DILITHIUM_KEYFORM_ROUND3_44);
+    if (oid != NULL && openssl_get_pqc_oid_name(oid)!= NULL)
+        soft_private->supports_dilithium = CK_TRUE;
+
+    oid = find_pqc_by_keyform(ml_dsa_oids, CKP_IBM_ML_DSA_44);
+    if (oid != NULL && openssl_get_pqc_oid_name(oid)!= NULL)
+        soft_private->supports_ml_dsa = CK_TRUE;
+
+    oid = find_pqc_by_keyform(ml_dsa_oids, CKP_IBM_ML_KEM_512);
+    if (oid != NULL && openssl_get_pqc_oid_name(oid)!= NULL)
+        soft_private->supports_ml_kem = CK_TRUE;
 #endif
 
     tokdata->private_data = soft_private;
@@ -375,7 +427,8 @@ CK_RV token_specific_final(STDLL_TokData_t *tokdata,
 }
 
 static CK_BBOOL token_specific_filter_mechanism(STDLL_TokData_t *tokdata,
-                                                CK_MECHANISM_TYPE mechanism)
+                                                CK_MECHANISM_TYPE mechanism,
+                                                CK_MECHANISM_INFO *info)
 {
 #if OPENSSL_VERSION_PREREQ(3, 0)
     struct soft_private_data *soft_private = tokdata->private_data;
@@ -383,10 +436,19 @@ static CK_BBOOL token_specific_filter_mechanism(STDLL_TokData_t *tokdata,
     UNUSED(tokdata);
 #endif
 
+    UNUSED(info);
+
     switch(mechanism) {
 #if OPENSSL_VERSION_PREREQ(3, 0)
     case CKM_IBM_DILITHIUM:
-        return soft_private->oqs_provider != NULL ? CK_TRUE : CK_FALSE;
+        return soft_private->supports_dilithium;
+    case CKM_IBM_ML_DSA_KEY_PAIR_GEN:
+    case CKM_IBM_ML_DSA:
+        return soft_private->supports_ml_dsa;
+    case CKM_IBM_ML_KEM_KEY_PAIR_GEN:
+    case CKM_IBM_ML_KEM:
+    case CKM_IBM_ML_KEM_WITH_ECDH:
+        return soft_private->supports_ml_kem;
 #endif
     default:
         return CK_TRUE;
@@ -1463,36 +1525,55 @@ CK_RV token_specific_ecdh_pkcs_derive(STDLL_TokData_t *tokdata,
 #endif
 
 #if OPENSSL_VERSION_PREREQ(3, 0)
-CK_RV import_ibm_dilithium_key(STDLL_TokData_t *tokdata, OBJECT *obj)
+static CK_RV import_pqc_key(STDLL_TokData_t *tokdata, OBJECT *obj,
+                            CK_KEY_TYPE keytype)
 {
-    struct soft_private_data *soft_private = tokdata->private_data;
     const struct pqc_oid *oid = NULL;
+    const char *alg_name;
     CK_ATTRIBUTE *attr = NULL;
     CK_OBJECT_CLASS class;
     CK_BYTE *data = NULL;
     CK_ULONG data_len;
+    CK_MECHANISM_TYPE mech;
+    EVP_PKEY *pkey = NULL;
+    size_t pub_len = 0;
+    CK_BYTE *pub_key = NULL;
     CK_RV rc;
 
-    if (soft_private->oqs_provider == NULL) {
-        TRACE_ERROR("The oqsprovider is not loaded\n");
-        return CKR_MECHANISM_INVALID;
+    switch (keytype) {
+    case CKK_IBM_DILITHIUM:
+        mech = CKM_IBM_DILITHIUM;
+        break;
+    case CKK_IBM_ML_DSA:
+        mech = CKM_IBM_ML_DSA;
+        break;
+    case CKK_IBM_ML_KEM:
+        mech = CKM_IBM_ML_KEM;
+        break;
+    default:
+        return CKR_KEY_TYPE_INCONSISTENT;
     }
 
+    if (!token_specific_filter_mechanism(tokdata, mech, NULL))
+        return CKR_MECHANISM_INVALID;
+
     rc = template_attribute_get_ulong(obj->template, CKA_CLASS, &class);
     if (rc != CKR_OK)
         return CKR_OK;
 
     /* A clear IBM Dilithium key must either have a CKA_VALUE containing
      * the SPKI or PKCS#8 encoded private key, or must have a keyform/mode
-     * value and the individual attributes
+     * value and the individual attributes.
+     * A clear ML-DSA or ML-KEM key must have a CKA_VALUE containing the SPKI
+     * or PKCS#8 encoded private key. Individual key attributes are not used.
      */
     if (template_attribute_find(obj->template, CKA_VALUE, &attr) == TRUE &&
         attr->ulValueLen > 0 && attr->pValue != NULL) {
         switch (class) {
         case CKO_PRIVATE_KEY:
             /* Private key in PKCS#8 form is present in CKA_VALUE */
-            rc = ibm_dilithium_priv_unwrap(obj->template, attr->pValue,
-                                           attr->ulValueLen, FALSE);
+            rc = pqc_priv_unwrap(obj->template, keytype, 
+                                 attr->pValue, attr->ulValueLen, FALSE);
             if (rc != CKR_OK) {
                 TRACE_ERROR("Failed to decode private key from "
                             "CKA_VALUE.\n");
@@ -1501,10 +1582,9 @@ CK_RV import_ibm_dilithium_key(STDLL_TokData_t *tokdata, OBJECT *obj)
             break;
         case CKO_PUBLIC_KEY:
             /* Public key in SPKI form is present in CKA_VALUE */
-            rc = ibm_dilithium_priv_unwrap_get_data(obj->template,
-                                                    attr->pValue,
-                                                    attr->ulValueLen,
-                                                    FALSE);
+            rc = pqc_priv_unwrap_get_data(obj->template, keytype,
+                                          attr->pValue, attr->ulValueLen,
+                                          FALSE);
             if (rc != CKR_OK) {
                 TRACE_ERROR("Failed to decode public key from "
                             "CKA_VALUE.\n");
@@ -1515,19 +1595,19 @@ CK_RV import_ibm_dilithium_key(STDLL_TokData_t *tokdata, OBJECT *obj)
             return CKR_TEMPLATE_INCONSISTENT;
         }
     } else {
-        /* Add CKA_VALUE withPKCS#8 or SPKI */
+        /* Add CKA_VALUE with PKCS#8 or SPKI */
         switch (class) {
          case CKO_PRIVATE_KEY:
-             rc = ibm_dilithium_priv_wrap_get_data(obj->template,
-                                                   FALSE, &data, &data_len);
+             rc = pqc_priv_wrap_get_data(obj->template, keytype,
+                                         FALSE, &data, &data_len);
              if (rc != CKR_OK) {
                  TRACE_ERROR("Failed to encode private key.\n");
                  return rc;
              }
              break;
          case CKO_PUBLIC_KEY:
-             rc = ibm_dilithium_publ_get_spki(obj->template,
-                                              FALSE, &data, &data_len);
+             rc = pqc_publ_get_spki(obj->template, keytype,
+                                    FALSE, &data, &data_len);
              if (rc != CKR_OK) {
                  TRACE_ERROR("Failed to encode public key.\n");
                  return rc;
@@ -1552,20 +1632,43 @@ CK_RV import_ibm_dilithium_key(STDLL_TokData_t *tokdata, OBJECT *obj)
         }
     }
 
-    oid = ibm_pqc_get_keyform_mode(obj->template, CKM_IBM_DILITHIUM);
+    oid = pqc_get_keyform_mode(obj->template, mech);
     if (oid == NULL) {
-        TRACE_ERROR("%s Failed to determine dilithium OID\n", __func__);
+        TRACE_ERROR("%s Failed to determine PQC OID\n", __func__);
         return CKR_TEMPLATE_INCOMPLETE;
     }
 
-    if (openssl_get_pqc_oid_name(oid) == NULL) {
-        TRACE_ERROR("Dilithium key form is not supported by oqsprovider\n");
+    alg_name = openssl_get_pqc_oid_name(oid);
+    if (alg_name == NULL) {
+        TRACE_ERROR("PQC key form is not supported by oqsprovider or "
+                    "OpenSSL\n");
         return CKR_KEY_SIZE_RANGE;
     }
 
-    rc = ibm_pqc_add_keyform_mode(obj->template, oid, CKM_IBM_DILITHIUM);
+    if (class == CKO_PRIVATE_KEY &&
+        (keytype == CKK_IBM_ML_DSA || keytype == CKK_IBM_ML_KEM)) {
+        /* Try tp add public key attributes if ML-DSA private key */
+        rc = openssl_make_pqc_key_from_template(obj->template, oid, mech,
+                                                TRUE, alg_name, &pkey);
+        if (rc != CKR_OK)
+            return rc;
+
+        rc = openssl_get_key_from_pkey(pkey, OSSL_PKEY_PARAM_PUB_KEY,
+                                       &pub_key, &pub_len, FALSE);
+        EVP_PKEY_free(pkey);
+        if (rc == CKR_OK) {
+            rc = pqc_unpack_pub_key(pub_key, pub_len, oid, mech, obj->template);
+            free(pub_key);
+            if (rc != CKR_OK) {
+                TRACE_ERROR("pqc_unpack_pub_key failed for pub key\n");
+                return rc;
+            }
+        }
+    }
+
+    rc = pqc_add_keyform_mode(obj->template, oid, mech);
     if (rc != CKR_OK) {
-        TRACE_ERROR("ibm_pqc_add_keyform_mode failed\n");
+        TRACE_ERROR("pqc_add_keyform_mode failed\n");
         return rc;
     }
 
@@ -1619,7 +1722,9 @@ CK_RV token_specific_object_add(STDLL_TokData_t * tokdata, SESSION * sess,
 
 #if OPENSSL_VERSION_PREREQ(3, 0)
     case CKK_IBM_DILITHIUM:
-        return import_ibm_dilithium_key(tokdata, obj);
+    case CKK_IBM_ML_DSA:
+    case CKK_IBM_ML_KEM:
+        return import_pqc_key(tokdata, obj, keytype);
 #endif
 
     default:
@@ -1635,9 +1740,9 @@ CK_RV token_specific_set_attrs_for_new_object(STDLL_TokData_t *tokdata,
     EVP_PKEY *pkey = NULL;
     CK_RV rc;
 #if OPENSSL_VERSION_PREREQ(3, 0)
-    struct soft_private_data *soft_private = tokdata->private_data;
     const struct pqc_oid *oid = NULL;
     const char *alg_name;
+    CK_MECHANISM_TYPE mech;
 #else
 
     UNUSED(tokdata);
@@ -1657,6 +1762,17 @@ CK_RV token_specific_set_attrs_for_new_object(STDLL_TokData_t *tokdata,
     case CKK_AES:
     case CKK_AES_XTS:
     case CKK_GENERIC_SECRET:
+    case CKK_SHA_1_HMAC:
+    case CKK_SHA224_HMAC:
+    case CKK_SHA256_HMAC:
+    case CKK_SHA384_HMAC:
+    case CKK_SHA512_HMAC:
+    case CKK_SHA3_224_HMAC:
+    case CKK_SHA3_256_HMAC:
+    case CKK_SHA3_384_HMAC:
+    case CKK_SHA3_512_HMAC:
+    case CKK_SHA512_224_HMAC:
+    case CKK_SHA512_256_HMAC:
 #if !(NODH)
     case CKK_DH:
 #endif
@@ -1677,12 +1793,24 @@ CK_RV token_specific_set_attrs_for_new_object(STDLL_TokData_t *tokdata,
 
 #if OPENSSL_VERSION_PREREQ(3, 0)
     case CKK_IBM_DILITHIUM:
-        if (soft_private->oqs_provider == NULL) {
-            TRACE_ERROR("The oqsprovider is not loaded\n");
-            return CKR_MECHANISM_INVALID;
+    case CKK_IBM_ML_DSA:
+    case CKK_IBM_ML_KEM:
+        switch (keytype) {
+        case CKK_IBM_DILITHIUM:
+            mech = CKM_IBM_DILITHIUM;
+            break;
+        case CKK_IBM_ML_DSA:
+            mech = CKM_IBM_ML_DSA;
+            break;
+        case CKK_IBM_ML_KEM:
+            mech = CKM_IBM_ML_KEM;
+            break;
         }
 
-        oid = ibm_pqc_get_keyform_mode(tmpl, CKM_IBM_DILITHIUM);
+        if (!token_specific_filter_mechanism(tokdata, mech, NULL))
+            return CKR_MECHANISM_INVALID;
+
+        oid = pqc_get_keyform_mode(tmpl, mech);
         if (oid == NULL) {
             TRACE_ERROR("%s Failed to determine dilithium OID\n", __func__);
             return CKR_TEMPLATE_INCOMPLETE;
@@ -1694,11 +1822,10 @@ CK_RV token_specific_set_attrs_for_new_object(STDLL_TokData_t *tokdata,
             return CKR_KEY_SIZE_RANGE;
         }
 
-        /* Check if the oqsprovider supports the variant */
-        rc = openssl_make_ibm_dilithium_key_from_template(tmpl, oid,
-                                                          class ==
-                                                                CKO_PRIVATE_KEY,
-                                                           alg_name, &pkey);
+        /* Check if the oqsprovider or OpenSSL supports the variant */
+        rc = openssl_make_pqc_key_from_template(tmpl, oid, mech,
+                                                class == CKO_PRIVATE_KEY,
+                                                alg_name, &pkey);
         if (pkey != NULL)
                 EVP_PKEY_free(pkey);
 
@@ -1712,66 +1839,103 @@ CK_RV token_specific_set_attrs_for_new_object(STDLL_TokData_t *tokdata,
 
 #if OPENSSL_VERSION_PREREQ(3, 0)
 
-CK_RV token_specific_ibm_dilithium_generate_keypair(STDLL_TokData_t *tokdata,
-                                                    const struct pqc_oid *oid,
-                                                    TEMPLATE *publ_tmpl,
-                                                    TEMPLATE *priv_tmpl)
+CK_RV token_specific_ibm_ml_dsa_generate_keypair(STDLL_TokData_t *tokdata,
+                                                 CK_MECHANISM *mech,
+                                                 const struct pqc_oid *oid,
+                                                 TEMPLATE *publ_tmpl,
+                                                 TEMPLATE *priv_tmpl)
 {
-    struct soft_private_data *soft_private = tokdata->private_data;
-
-    if (soft_private->oqs_provider == NULL) {
-        TRACE_ERROR("The oqsprovider is not loaded\n");
+    if (!token_specific_filter_mechanism(tokdata, mech->mechanism, NULL))
         return CKR_MECHANISM_INVALID;
-    }
 
-    return openssl_specific_ibm_dilithium_generate_keypair(tokdata, oid,
-                                                           publ_tmpl,
-                                                           priv_tmpl);
+    return openssl_specific_pqc_generate_keypair(tokdata, oid, mech,
+                                                 publ_tmpl, priv_tmpl);
 }
 
-CK_RV token_specific_ibm_dilithium_sign(STDLL_TokData_t *tokdata,
-                                        SESSION *sess,
-                                        CK_BBOOL length_only,
-                                        const struct pqc_oid *oid,
-                                        CK_BYTE *in_data,
-                                        CK_ULONG in_data_len,
-                                        CK_BYTE *signature,
-                                        CK_ULONG *signature_len,
-                                        OBJECT *key_obj)
+CK_RV token_specific_ibm_ml_dsa_sign(STDLL_TokData_t *tokdata,
+                                     SESSION *sess,
+                                     CK_BBOOL length_only,
+                                     const struct pqc_oid *oid,
+                                     CK_MECHANISM *mech,
+                                     CK_BYTE *in_data,
+                                     CK_ULONG in_data_len,
+                                     CK_BYTE *signature,
+                                     CK_ULONG *signature_len,
+                                     OBJECT *key_obj)
 {
-    struct soft_private_data *soft_private = tokdata->private_data;
-
-    if (soft_private->oqs_provider == NULL) {
-        TRACE_ERROR("The oqsprovider is not loaded\n");
+    if (!token_specific_filter_mechanism(tokdata, mech->mechanism, NULL))
         return CKR_MECHANISM_INVALID;
-    }
 
-    return openssl_specific_ibm_dilithium_sign(tokdata, sess, length_only, oid,
-                                               in_data, in_data_len,
-                                               signature, signature_len,
-                                               key_obj);
+    return openssl_specific_pqc_sign(tokdata, sess, length_only, oid,
+                                     mech, in_data, in_data_len,
+                                     signature, signature_len,
+                                     key_obj);
 }
 
-CK_RV token_specific_ibm_dilithium_verify(STDLL_TokData_t *tokdata,
-                                          SESSION *sess,
-                                          const struct pqc_oid *oid,
-                                          CK_BYTE *in_data,
-                                          CK_ULONG in_data_len,
-                                          CK_BYTE *signature,
-                                          CK_ULONG signature_len,
-                                          OBJECT *key_obj)
+CK_RV token_specific_ibm_ml_dsa_verify(STDLL_TokData_t *tokdata,
+                                       SESSION *sess,
+                                       const struct pqc_oid *oid,
+                                       CK_MECHANISM *mech,
+                                       CK_BYTE *in_data,
+                                       CK_ULONG in_data_len,
+                                       CK_BYTE *signature,
+                                       CK_ULONG signature_len,
+                                       OBJECT *key_obj)
 {
-    struct soft_private_data *soft_private = tokdata->private_data;
-
-    if (soft_private->oqs_provider == NULL) {
-        TRACE_ERROR("The oqsprovider is not loaded\n");
+    if (!token_specific_filter_mechanism(tokdata, mech->mechanism, NULL))
         return CKR_MECHANISM_INVALID;
-    }
 
-    return openssl_specific_ibm_dilithium_verify(tokdata, sess, oid,
-                                                 in_data, in_data_len,
-                                                 signature, signature_len,
-                                                 key_obj);
+    return openssl_specific_pqc_verify(tokdata, sess, oid, mech,
+                                       in_data, in_data_len,
+                                       signature, signature_len,
+                                       key_obj);
+}
+
+CK_RV token_specific_ibm_ml_kem_generate_keypair(STDLL_TokData_t *tokdata,
+                                                 CK_MECHANISM *mech,
+                                                 const struct pqc_oid *oid,
+                                                 TEMPLATE *publ_tmpl,
+                                                 TEMPLATE *priv_tmpl)
+{
+    if (!token_specific_filter_mechanism(tokdata, mech->mechanism, NULL))
+        return CKR_MECHANISM_INVALID;
+
+    return openssl_specific_pqc_generate_keypair(tokdata, oid, mech,
+                                                 publ_tmpl, priv_tmpl);
+}
+
+CK_RV token_specific_ibm_ml_kem_derive(STDLL_TokData_t *tokdata, SESSION *sess,
+                                       const struct pqc_oid *oid,
+                                       CK_MECHANISM *mech,
+                                       OBJECT *base_object,
+                                       CK_OBJECT_CLASS base_key_class,
+                                       CK_KEY_TYPE base_key_type,
+                                       OBJECT *derived_object,
+                                       CK_KEY_TYPE derived_key_type,
+                                       CK_ULONG derived_keylen)
+{
+    if (!token_specific_filter_mechanism(tokdata, mech->mechanism, NULL))
+        return CKR_MECHANISM_INVALID;
+
+    return openssl_specific_pqc_kem_derive(tokdata, sess, oid, mech,
+                                           base_object, base_key_class,
+                                           base_key_type,
+                                           derived_object, derived_key_type,
+                                           derived_keylen);
 }
 
 #endif
+
+CK_RV token_specific_get_token_info(STDLL_TokData_t *tokdata,
+                                    CK_TOKEN_INFO_PTR pInfo)
+{
+    UNUSED(tokdata);
+
+    /* OpenSSL_version_num returns the version as 0xMNN00PP0 */
+    pInfo->firmwareVersion.major = (OpenSSL_version_num() & 0xF0000000) >> 28;
+    pInfo->firmwareVersion.minor = (OpenSSL_version_num() & 0x0FF00000) >> 20;
+    pInfo->hardwareVersion.major = pInfo->firmwareVersion.major;
+    pInfo->hardwareVersion.minor = pInfo->firmwareVersion.minor;
+
+    return CKR_OK;
+}
