@@ -66,6 +66,7 @@ token_spec_t token_specific = {
     NULL,                       // init_token_data
     NULL,                       // load_token_data
     NULL,                       // save_token_data
+    &token_specific_get_token_info,
     NULL,                       // random number generator
     &token_specific_final,
     NULL,                       // init_token
@@ -106,11 +107,13 @@ token_spec_t token_specific = {
     &token_specific_ec_verify,
     &token_specific_ec_generate_keypair,
     &token_specific_ecdh_pkcs_derive,
+    NULL,                       // ecdh_derive_kdf
 #else
     NULL,                       // ec_sign
     NULL,                       // ec_verify
     NULL,                       // ec_generate_keypair
     NULL,                       // ecdh_derive
+    NULL,                       // ecdh_derive_kdf
 #endif
 /* Begin code contributed by Corrent corp. */
     // DH
@@ -126,6 +129,8 @@ token_spec_t token_specific = {
     &token_specific_sha,
     &token_specific_sha_update,
     &token_specific_sha_final,
+    // SHAKE derive
+    &token_specific_shake_key_derive,
     // HMAC
     &token_specific_hmac_sign_init,
     &token_specific_hmac_sign,
@@ -151,10 +156,25 @@ token_spec_t token_specific = {
     &token_specific_aes_mac,
     &token_specific_aes_cmac,
     &token_specific_aes_xts,
+    &token_specific_aes_key_wrap,
     // DSA
     NULL,                       // dsa_generate_keypair
     NULL,                       // dsa_sign
     NULL,                       // dsa_verify
+    // PQC
+#if OPENSSL_VERSION_PREREQ(3, 0)
+    &token_specific_ibm_ml_dsa_generate_keypair,
+    &token_specific_ibm_ml_dsa_sign,
+    &token_specific_ibm_ml_dsa_verify,
+    &token_specific_ibm_ml_kem_generate_keypair,
+    &token_specific_ibm_ml_kem_derive,
+#else
+    NULL,                       // ibm_ml_dsa_generate_keypair
+    NULL,                       // ibm_ml_dsa_sign
+    NULL,                       // ibm_ml_dsa_verify
+    NULL,                       // ibm_ml_kem_generate_keypair
+    NULL,                       // ibm_ml_kem_derive
+#endif
     &token_specific_get_mechanism_list,
     &token_specific_get_mechanism_info,
     &token_specific_object_add,
@@ -162,7 +182,7 @@ token_spec_t token_specific = {
     NULL,                       // key_unwrap
     NULL,                       // reencrypt_single
     NULL,                       // set_attribute_values
-    NULL,                       // set_attrs_for_new_object
+    &token_specific_set_attrs_for_new_object,
     NULL,                       // handle_event
     NULL,                       // check_obj_access
 };
