@@ -39,6 +39,7 @@ struct token_specific_struct token_specific = {
     &token_specific_init_token_data,
     NULL,                       // load_token_data
     NULL,                       // save_token_data
+    NULL,                       // get_token_info
     &token_specific_rng,
     &token_specific_final,
     &token_specific_init_token,
@@ -119,9 +120,11 @@ struct token_specific_struct token_specific = {
     NULL,                       // dsa_sign
     NULL,                       // dsa_verify
     // PQC
-    NULL,                       // ibm_dilithium_generate_keypair
-    NULL,                       // ibm_dilithium_sign
-    NULL,                       // ibm_dilithium_verify
+    NULL,                       // ibm_ml_dsa_generate_keypair
+    NULL,                       // ibm_ml_dsa_sign
+    NULL,                       // ibm_ml_dsa_verify
+    NULL,                       // ibm_ml_kem_generate_keypair
+    NULL,                       // ibm_ml_kem_derive
     &token_specific_get_mechanism_list,
     &token_specific_get_mechanism_info,
     NULL,                       // object_add
