@@ -77,6 +77,28 @@ const struct p11tool_enum_value p11tool_ibm_kyber_versions[] = {
     { .value = NULL, },
 };
 
+const struct p11tool_enum_value p11tool_ibm_ml_dsa_versions[] = {
+    { .value = "44", .args = NULL,
+      .private = { .num = CKP_IBM_ML_DSA_44 }, },
+    { .value = "65", .args = NULL,
+      .private = { .num = CKP_IBM_ML_DSA_65 }, },
+    { .value = "87", .args = NULL,
+      .private = { .num = CKP_IBM_ML_DSA_87 }, },
+    { .value = NULL, },
+};
+
+const struct p11tool_enum_value p11tool_ibm_ml_kem_versions[] = {
+    { .value = "512", .args = NULL,
+      .private = { .num = CKP_IBM_ML_KEM_512 }, },
+    { .value = "768", .args = NULL,
+      .private = { .num = CKP_IBM_ML_KEM_768 }, },
+    { .value = "1024", .args = NULL,
+      .private = { .num = CKP_IBM_ML_KEM_1024 }, },
+    { .value = NULL, },
+};
+
+static bool p11tool_argument_is_set(const struct p11tool_arg *arg);
+
 const struct p11tool_cmd *p11tool_find_command(const struct p11tool_cmd *cmds,
                                                const char *cmd)
 {
@@ -155,8 +177,76 @@ static CK_RV p11tool_build_opts(const struct p11tool_opt *opts,
     return CKR_OK;
 }
 
+static void p11tool_count_arg_opts(const struct p11tool_arg *args,
+                                   unsigned int *optstring_len,
+                                   unsigned int *longopts_count)
+{
+    const struct p11tool_arg *arg;
+    const struct p11tool_enum_value *val;
+
+    for (arg = args; arg->name != NULL; arg++) {
+        if (!p11tool_argument_is_set(arg))
+            continue;
+
+        if (arg->type != ARG_TYPE_ENUM)
+            continue;
+
+        for (val = arg->enum_values; val->value != NULL; val++) {
+            if (val->opts == NULL)
+                continue;
+
+            if (*arg->value.enum_value != val)
+                continue;
+
+            p11tool_count_opts(val->opts, optstring_len, longopts_count);
+
+            if (val->args != NULL)
+                p11tool_count_arg_opts(val->args, optstring_len,
+                                      longopts_count);
+        }
+    }
+}
+
+static CK_RV p11tool_build_arg_opts(const struct p11tool_arg *args,
+                                    char *optstring,
+                                    struct option *longopts)
+{
+    const struct p11tool_arg *arg;
+    const struct p11tool_enum_value *val;
+    CK_RV rc;
+
+    for (arg = args; arg->name != NULL; arg++) {
+        if (!p11tool_argument_is_set(arg))
+            continue;
+
+        if (arg->type != ARG_TYPE_ENUM)
+            continue;
+
+        for (val = arg->enum_values; val->value != NULL; val++) {
+            if (val->opts == NULL)
+                continue;
+
+            if (*arg->value.enum_value != val)
+                continue;
+
+            rc = p11tool_build_opts(val->opts, optstring, longopts);
+            if (rc != CKR_OK)
+                return rc;
+
+            if (val->args != NULL) {
+                rc = p11tool_build_arg_opts(val->args, optstring, longopts);
+                if (rc != CKR_OK)
+                    return rc;
+            }
+        }
+    }
+
+    return CKR_OK;
+}
+
 static CK_RV p11tool_build_cmd_opts(const struct p11tool_opt *cmd_opts,
                                     const struct p11tool_opt *generic_opts,
+                                    const struct p11tool_arg *cmd_args,
                                     char **optstring, struct option **longopts)
 {
     unsigned int optstring_len = 0, longopts_count = 0;
@@ -165,6 +255,8 @@ static CK_RV p11tool_build_cmd_opts(const struct p11tool_opt *cmd_opts,
     p11tool_count_opts(generic_opts, &optstring_len, &longopts_count);
     if (cmd_opts != NULL)
         p11tool_count_opts(cmd_opts, &optstring_len, &longopts_count);
+    if (cmd_args != NULL)
+        p11tool_count_arg_opts(cmd_args, &optstring_len, &longopts_count);
 
     *optstring = calloc(1 + optstring_len + 1, 1);
     *longopts = calloc(longopts_count + 1, sizeof(struct option));
@@ -185,6 +277,12 @@ static CK_RV p11tool_build_cmd_opts(const struct p11tool_opt *cmd_opts,
             goto error;
     }
 
+    if (cmd_args != NULL) {
+        rc = p11tool_build_arg_opts(cmd_args, *optstring, *longopts);
+        if (rc != CKR_OK)
+            goto error;
+    }
+
     return CKR_OK;
 
 error:
@@ -337,7 +435,44 @@ static CK_RV p11tool_process_option(const struct p11tool_opt *opts,
     return CKR_ARGUMENTS_BAD;
 }
 
+static CK_RV p11tool_process_arg_option(const struct p11tool_arg *args,
+                                        int ch, char *value)
+{
+    const struct p11tool_arg *arg;
+    const struct p11tool_enum_value *val;
+    CK_RV rc;
+
+    for (arg = args; arg->name != NULL; arg++) {
+        if (!p11tool_argument_is_set(arg))
+            continue;
+
+        if (arg->type != ARG_TYPE_ENUM)
+            continue;
+
+        for (val = arg->enum_values; val->value != NULL; val++) {
+            if (val->opts == NULL)
+                continue;
+
+            if (*arg->value.enum_value != val)
+                continue;
+
+            rc = p11tool_process_option(val->opts, ch, value);
+            if (rc != CKR_OK)
+                return rc;
+
+            if (val->args != NULL) {
+                rc = p11tool_process_arg_option(val->args, ch, value);
+                if (rc != CKR_OK)
+                    return rc;
+            }
+        }
+    }
+
+    return CKR_OK;
+}
+
 static CK_RV p11tool_process_cmd_option(const struct p11tool_opt *cmd_opts,
+                                        const struct p11tool_arg *cmd_args,
                                         const struct p11tool_opt *generic_opts,
                                         int opt, char *arg)
 {
@@ -353,6 +488,12 @@ static CK_RV p11tool_process_cmd_option(const struct p11tool_opt *cmd_opts,
             return CKR_OK;
     }
 
+    if (cmd_args != NULL) {
+        rc = p11tool_process_arg_option(cmd_args, opt, arg);
+        if (rc == CKR_OK)
+            return CKR_OK;
+    }
+
     return rc;
 }
 
@@ -373,22 +514,64 @@ static CK_RV p11tool_check_required_opts(const struct p11tool_opt *opts)
     return rc;
 }
 
+static CK_RV p11tool_check_required_arg_opts(const struct p11tool_arg *args)
+{
+    const struct p11tool_arg *arg;
+    const struct p11tool_enum_value *val;
+    CK_RV rc;
+
+    for (arg = args; arg->name != NULL; arg++) {
+        if (!p11tool_argument_is_set(arg))
+            continue;
+
+        if (arg->type != ARG_TYPE_ENUM)
+            continue;
+
+        for (val = arg->enum_values; val->value != NULL; val++) {
+            if (val->opts == NULL)
+                continue;
+
+            if (*arg->value.enum_value != val)
+                continue;
+
+            rc = p11tool_check_required_opts(val->opts);
+            if (rc != CKR_OK)
+                return rc;
+
+            if (val->args != NULL) {
+                rc = p11tool_check_required_arg_opts(val->args);
+                if (rc != CKR_OK)
+                    return rc;
+            }
+        }
+    }
+
+    return CKR_OK;
+}
+
 CK_RV p11tool_check_required_cmd_opts(const struct p11tool_opt *cmd_opts,
+                                      const struct p11tool_arg *cmd_args,
                                       const struct p11tool_opt *generic_opts)
 {
-    CK_RV rc;
+    CK_RV rc, rc2;
 
     rc = p11tool_check_required_opts(generic_opts);
     if (rc != CKR_OK)
         return rc;
 
     if (cmd_opts != NULL) {
-        rc = p11tool_check_required_opts(cmd_opts);
-        if (rc != CKR_OK)
-            return rc;
+        rc2 = p11tool_check_required_opts(cmd_opts);
+        if (rc == CKR_OK)
+            rc = rc2;
     }
 
-    return CKR_OK;
+    if (cmd_args != NULL) {
+        rc2 = p11tool_check_required_arg_opts(cmd_args);
+        if (rc == CKR_OK)
+            rc = rc2;
+    }
+
+    return rc;
 }
 
 CK_RV p11tool_parse_cmd_options(const struct p11tool_cmd *cmd,
@@ -401,6 +584,7 @@ CK_RV p11tool_parse_cmd_options(const struct p11tool_cmd *cmd,
     int c;
 
     rc = p11tool_build_cmd_opts(cmd != NULL ? cmd->opts : NULL, generic_opts,
+                                cmd != NULL ? cmd->args : NULL,
                                 &optstring, &longopts);
     if (rc != CKR_OK)
         goto done;
@@ -427,6 +611,7 @@ CK_RV p11tool_parse_cmd_options(const struct p11tool_cmd *cmd,
 
         default:
             rc = p11tool_process_cmd_option(cmd != NULL ? cmd->opts : NULL,
+                                            cmd != NULL ? cmd->args : NULL,
                                             generic_opts, c, optarg);
             if (rc != CKR_OK)
                 goto done;
@@ -554,6 +739,7 @@ void p11tool_print_indented(const char *str, int indent)
 static void p11tool_print_options_help(const struct p11tool_opt *opts,
                                        int indent_pos)
 {
+    const struct p11tool_enum_value *val;
     const struct p11tool_opt *opt;
     char tmp[200];
     int len;
@@ -567,6 +753,11 @@ static void p11tool_print_options_help(const struct p11tool_opt *opts,
         else
             len = snprintf(tmp, sizeof(tmp),"-%c", opt->short_opt);
 
+        if (len >= (int)sizeof(tmp) || len < 0) {
+            warnx("Error formatting option string. Skipping.\n");
+            continue;
+        }
+
         if (opt->arg.type != ARG_TYPE_PLAIN) {
             if (opt->arg.required)
                 snprintf(&tmp[len], sizeof(tmp) - len, " %s", opt->arg.name);
@@ -578,6 +769,12 @@ static void p11tool_print_options_help(const struct p11tool_opt *opts,
 
         printf("    %-*.*s ", indent_pos - 5, indent_pos - 5, tmp);
         p11tool_print_indented(opt->description, indent_pos);
+
+        if (opt->arg.type == ARG_TYPE_ENUM) {
+            for (val = opt->arg.enum_values; val->value != NULL; val++) {
+                printf("%*s    %s\n", indent_pos, "", val->value);
+            }
+        }
     }
 }
 
@@ -621,13 +818,21 @@ static void p11tool_print_arguments_help(const struct p11tool_cmd *cmd,
 
             newline = true;
 
-            printf("%*s        %s", indent, "", val->value);
+            width = indent_pos - 9 - indent;
+            if (width < (int)strlen(val->value))
+                width = (int)strlen(val->value);
+
+            printf("%*s        %-*.*s ", indent, "", width, width, val->value);
+            if (val->description) {
+                p11tool_print_indented(val->description, indent_pos);
+                newline = false;
+            }
 
             if (val->args != NULL) {
                 p11tool_print_arguments_help(cmd, val->args, indent + 8,
                                              indent_pos);
                 newline = false;
-            } else {
+            } else if (val->description == NULL) {
                 printf("\n");
             }
         }
@@ -637,6 +842,34 @@ static void p11tool_print_arguments_help(const struct p11tool_cmd *cmd,
         printf("\n");
 }
 
+static void p11tool_print_argument_options_help(const struct p11tool_arg *args,
+                                                int indent_pos)
+{
+    const struct p11tool_arg *arg;
+    const struct p11tool_enum_value *val;
+
+    for (arg = args; arg->name != NULL; arg++) {
+        if (arg->type != ARG_TYPE_ENUM)
+            continue;
+
+        for (val = arg->enum_values; val->value != NULL; val++) {
+            if (val->opts == NULL)
+                continue;
+
+            if (p11tool_argument_is_set(arg) && *arg->value.enum_value != val)
+                continue;
+
+            printf("\nOPTIONS FOR ARGUMENT '%s' VALUE '%s':\n",
+                   arg->name, val->value);
+
+            p11tool_print_options_help(val->opts, indent_pos);
+
+            if (val->args != NULL)
+                p11tool_print_argument_options_help(val->args, indent_pos);
+        }
+    }
+}
+
 void p11tool_print_help(const char *name,
                         const struct p11tool_cmd *commands,
                         const struct p11tool_opt *generic_opts,
@@ -673,6 +906,7 @@ void p11tool_print_command_help(const char *name,
     printf("OPTIONS:\n");
     p11tool_print_options_help(cmd->opts, indent_pos);
     p11tool_print_options_help(generic_opts, indent_pos);
+    p11tool_print_argument_options_help(cmd->args, indent_pos);
     printf("\n");
     if (cmd->help != NULL)
         cmd->help();
@@ -1105,6 +1339,62 @@ void p11tool_print_ibm_kyber_keyform_attr(const char *attr,
     }
 }
 
+void p11tool_print_ibm_ml_dsa_parameter_set_attr(const char *attr,
+                                                 const CK_ATTRIBUTE *val,
+                                                 int indent, bool sensitive)
+{
+    const struct p11tool_enum_value *eval;
+    const char *name = "[unknown]";
+
+    if ((val->ulValueLen == CK_UNAVAILABLE_INFORMATION ||
+         val->ulValueLen != sizeof (CK_ULONG)) &&
+        !sensitive)
+        return;
+
+    if (sensitive) {
+        printf("%*s%s: [sensitive]\n", indent, "", attr);
+    } else if (val->ulValueLen == 0) {
+        printf("%*s%s: [no value]\n", indent, "", attr);
+    } else {
+        for (eval = p11tool_ibm_ml_dsa_versions; eval->value != NULL; eval++) {
+            if (eval->private.num == *(CK_ULONG *)(val->pValue)) {
+                name = eval->value;
+                break;
+            }
+        }
+        printf("%*s%s: ML-DSA-%s (0x%lX)\n", indent, "", attr, name,
+               *(CK_ULONG *)(val->pValue));
+    }
+}
+
+void p11tool_print_ibm_ml_kem_parameter_set_attr(const char *attr,
+                                                 const CK_ATTRIBUTE *val,
+                                                 int indent, bool sensitive)
+{
+    const struct p11tool_enum_value *eval;
+    const char *name = "[unknown]";
+
+    if ((val->ulValueLen == CK_UNAVAILABLE_INFORMATION ||
+         val->ulValueLen != sizeof (CK_ULONG)) &&
+        !sensitive)
+        return;
+
+    if (sensitive) {
+        printf("%*s%s: [sensitive]\n", indent, "", attr);
+    } else if (val->ulValueLen == 0) {
+        printf("%*s%s: [no value]\n", indent, "", attr);
+    } else {
+        for (eval = p11tool_ibm_ml_kem_versions; eval->value != NULL; eval++) {
+            if (eval->private.num == *(CK_ULONG *)(val->pValue)) {
+                name = eval->value;
+                break;
+            }
+        }
+        printf("%*s%s: ML-KEM-%s (0x%lX)\n", indent, "", attr, name,
+               *(CK_ULONG *)(val->pValue));
+    }
+}
+
 void p11tool_print_class_attr(const char *attr, const CK_ATTRIBUTE *val,
                               int indent, bool sensitive)
 {
@@ -1380,13 +1670,14 @@ CK_RV p11tool_add_attributes(const struct p11tool_objtype *objtype,
     return CKR_OK;
 }
 
-CK_RV p11tool_parse_id(const char *id_string, CK_ATTRIBUTE **attrs,
-                       CK_ULONG *num_attrs)
+CK_RV p11tool_parse_hex(const char *id_string, CK_BYTE **buf, CK_ULONG *buflen)
 {
-    unsigned char *buf = NULL;
+    CK_RV rc = CKR_OK;
     BIGNUM *b = NULL;
     int len;
-    CK_RV rc = CKR_OK;
+
+    *buflen = 0;
+    *buf = NULL;
 
     len = BN_hex2bn(&b, id_string);
     if (len < (int)strlen(id_string) || len == 0 || b == NULL) {
@@ -1396,19 +1687,43 @@ CK_RV p11tool_parse_id(const char *id_string, CK_ATTRIBUTE **attrs,
     }
 
     len = len / 2 + (len % 2 > 0 ? 1 : 0);
-    buf = calloc(1, len);
-    if (buf == NULL) {
-        warnx("Failed to allocate memory for CKA_ID attribute");
+    *buf = calloc(1, len);
+    if (*buf == NULL) {
+        warnx("Failed to allocate memory for buffer");
         rc = CKR_HOST_MEMORY;
         goto done;
     }
 
-    if (BN_bn2binpad(b, buf, len) != len) {
+    if (BN_bn2binpad(b, *buf, len) != len) {
         warnx("Failed to prepare the value for CKA_ID attribute");
         rc = CKR_FUNCTION_FAILED;
         goto done;
     }
 
+    *buflen = len;
+
+done:
+    if (rc != CKR_OK && *buf != NULL) {
+        free(*buf);
+        *buf = NULL;
+    }
+    if (b != NULL)
+        BN_free(b);
+
+    return rc;
+}
+
+CK_RV p11tool_parse_id(const char *id_string, CK_ATTRIBUTE **attrs,
+                       CK_ULONG *num_attrs)
+{
+    CK_BYTE *buf = NULL;
+    CK_ULONG len;
+    CK_RV rc;
+
+    rc = p11tool_parse_hex(id_string, &buf, &len);
+    if (rc != CKR_OK)
+        return rc;
+
     rc = p11tool_add_attribute(CKA_ID, buf, len, attrs, num_attrs);
     if (rc != CKR_OK) {
         warnx("Failed to add attribute CKA_ID: 0x%lX: %s", rc, p11_get_ckr(rc));
@@ -1418,8 +1733,6 @@ CK_RV p11tool_parse_id(const char *id_string, CK_ATTRIBUTE **attrs,
 done:
     if (buf != NULL)
         free(buf);
-    if (b != NULL)
-        BN_free(b);
 
     return rc;
 }
@@ -2047,7 +2360,7 @@ static CK_RV p11tool_open_pkcs11_session(CK_SLOT_ID slot, CK_FLAGS flags,
     if (pin != NULL) {
         rc = p11tool_pkcs11_funcs->C_Login(p11tool_pkcs11_session, user_type,
                                            (CK_CHAR *)pin, strlen(pin));
-        if (rc != CKR_OK) {
+        if (rc != CKR_OK && rc != CKR_USER_ALREADY_LOGGED_IN) {
             warnx("Login failed: C_Login: 0x%lX: %s", rc, p11_get_ckr(rc));
             return rc;
         }
@@ -2121,7 +2434,7 @@ CK_RV p11tool_init_pkcs11(const struct p11tool_cmd *command, bool no_login,
         goto done;
 
     rc = p11tool_pkcs11_funcs->C_Initialize(NULL);
-    if (rc != CKR_OK) {
+    if (rc != CKR_OK && rc != CKR_CRYPTOKI_ALREADY_INITIALIZED) {
         warnx("C_Initialize failed: 0x%lX: %s", rc, p11_get_ckr(rc));
         goto done;
     }
@@ -2185,6 +2498,35 @@ bool p11tool_is_rejected_by_policy(CK_RV ret_code, CK_SESSION_HANDLE session)
     return (info.ulDeviceError == CKR_POLICY_VIOLATION);
 }
 
+CK_RV p11tool_check_wrap_mech_supported(CK_SLOT_ID slot,
+                                        CK_MECHANISM_TYPE mechanism,
+                                        CK_BBOOL wrap, CK_BBOOL unwrap)
+{
+    CK_MECHANISM_INFO mech_info;
+    CK_RV rc;
+
+    rc = p11tool_pkcs11_funcs->C_GetMechanismInfo(slot, mechanism, &mech_info);
+    if (rc != CKR_OK) {
+        warnx("Token in slot %lu does not support mechanism %s", slot,
+              p11_get_ckm(&mechtable_funcs, mechanism));
+        return rc;
+    }
+
+    if (wrap && (mech_info.flags & CKF_WRAP) == 0) {
+        warnx("Mechanism %s does not support to wrap keys",
+              p11_get_ckm(&mechtable_funcs, mechanism));
+        return CKR_MECHANISM_INVALID;
+    }
+
+    if (unwrap && (mech_info.flags & CKF_UNWRAP) == 0) {
+        warnx("Mechanism %s does not support to unwrap keys",
+              p11_get_ckm(&mechtable_funcs, mechanism));
+        return CKR_MECHANISM_INVALID;
+    }
+
+    return CKR_OK;
+}
+
 CK_RV p11tool_check_keygen_mech_supported(CK_SLOT_ID slot,
                                           CK_MECHANISM_TYPE mechanism,
                                           bool is_asymmetric, CK_ULONG keysize)
@@ -2263,7 +2605,9 @@ int p11tool_pem_password_cb(char *buf, int size, int rwflag, void *userdata)
         pem_password = getenv(data->env_var_name);
 
     if (data->force_prompt || pem_password == NULL) {
-        if (asprintf(&msg, "Please enter PEM password for '%s': ",
+        if (asprintf(&msg, "Please enter %s for '%s': ",
+                     strncmp(data->pem_file_name, "pkcs11:", 7) == 0 ?
+                                     "PKCS#11 user PIN" : "PEM password",
                      data->pem_file_name) <= 0) {
             warnx("Failed to allocate memory for message");
             return -1;
@@ -2310,10 +2654,45 @@ CK_RV p11tool_ASN1_TIME2date(const ASN1_TIME *asn1time, CK_DATE *date)
 }
 
 #if OPENSSL_VERSION_PREREQ(3, 0)
+CK_BBOOL p11tool_check_settable_fromdata_params(EVP_PKEY_CTX *ctx,
+                                                const char *name)
+{
+    const OSSL_PARAM *settable, *p;
+
+    settable = EVP_PKEY_fromdata_settable(ctx, EVP_PKEY_KEYPAIR);
+
+    for (p = settable; p != NULL && p->key != NULL; p++) {
+        if (strcmp(p->key, name) == 0)
+            return TRUE;
+    }
+
+    return CK_FALSE;
+}
+
 CK_RV p11tool_get_octet_string_param_from_pkey(EVP_PKEY *pkey,
                                                const char *param,
-                                               CK_BYTE **key, size_t *key_len)
+                                               CK_BYTE **key, size_t *key_len,
+                                               CK_BBOOL check)
 {
+    const OSSL_PARAM *settable, *p;
+    CK_BBOOL found = FALSE;
+
+    if (check) {
+        settable = EVP_PKEY_gettable_params(pkey);
+        for (p = settable; p != NULL && p->key != NULL; p++) {
+            if (strcmp(p->key, param) == 0) {
+                found = TRUE;
+                break;
+            }
+        }
+
+        if (!found) {
+            *key = NULL;
+            *key_len = 0;
+            return CKR_OK;
+        }
+    }
+
     if (EVP_PKEY_get_octet_string_param(pkey, param, NULL, 0, key_len) != 1 ||
         *key_len == OSSL_PARAM_UNMODIFIED) {
         warnx("EVP_PKEY_get_octet_string_param failed for '%s'\n", param);
@@ -2381,3 +2760,80 @@ CK_RV p11tool_prepare_uri(CK_OBJECT_HANDLE key, CK_OBJECT_CLASS *class,
 
     return CKR_OK;
 }
+
+CK_RV p11tool_bio_readall(BIO *bio, CK_BYTE **buffer, CK_ULONG *read_len)
+{
+    CK_ULONG ofs = 0, buf_len = 0;
+    CK_BYTE *buf = NULL, *tmp;
+    size_t read, left = 0;
+    int ret;
+
+    *buffer = NULL;
+    *read_len = 0;
+
+    do {
+        if (left == 0) {
+            buf_len += 1024;
+
+            tmp = OPENSSL_realloc(buf, buf_len);
+            if (tmp == NULL) {
+                warnx("Failed to allocate a buffer for reading a file");
+                free(buf);
+                return CKR_HOST_MEMORY;
+            }
+            buf = tmp;
+            left = 1024;
+        }
+
+        ret = BIO_read_ex(bio, buf + ofs, left, &read);
+        if (ret != 1)
+            break;
+
+        ofs += read;
+        left -= read;
+    } while(1);
+
+    if (ofs == 0) {
+        free(buf);
+        return CKR_FUNCTION_FAILED;
+    }
+
+    *buffer = buf;
+    *read_len = ofs;
+
+    return CKR_OK;
+}
+
+CK_RV p11tool_split_by_delim(char *str, char *delim, char ***list)
+{
+    char **l = NULL, **tmp;
+    CK_ULONG num = 0, left = 0, ofs = 0;
+    char *tok;
+    char *save = NULL;
+
+    do {
+        tok = strtok_r(save == NULL ? str : NULL, delim, &save);
+
+        if (left == 0) {
+            num += 16;
+
+            tmp = realloc(l, num * sizeof(char *));
+            if (tmp == NULL) {
+                warnx("Failed to allocate a buffer for reading a file");
+                free(l);
+                return CKR_HOST_MEMORY;
+            }
+
+            l = tmp;
+            left = 16;
+        }
+
+        l[ofs] = tok;
+        ofs++;
+        left--;
+    } while(tok != NULL);
+
+    *list = l;
+
+    return CKR_OK;
+}
